
#include "stdafx.h"
#include "tavr.h"
#include "tavr_loader.h"
#include "tavr_project_loader.h"

#define FOR(CONT_TYPE, ITER, CONT) \
	for (CONT_TYPE::iterator ITER = (CONT).begin(), ITER##_end = (CONT).end(); ITER != ITER##_end; ++ITER)

using namespace ecg;
using namespace std;

namespace tavr {

/*************************************************************************/
/* Swap()                                                                */
/*                                                                       */
/*************************************************************************/
template <class T>
void Swap(T & a, T & b)
{
	T tmp = a;
	a = b;
	b = tmp;
}

/************************************************************************/
/* tavr::PROJECT::Load
/*
/************************************************************************/
bool PROJECT::Load(std::istream & source)
{
	TAVR_PROJECT_LOADER loader(this);
	loader.Init(source, GetError());
	loader.Compile();
	if (loader.HasErrors()) {
		return false;
	}

	const PROJECT_SETTINGS * projSettings = GetSettings("project");
	if (projSettings == NULL) {
		GetError() << "ERROR: section 'project' not found" << endl;
		return false;
	}

	if (GetSettings("project/name") == NULL) {
		GetError() << "ERROR: property 'project/name' not found" << endl;
		return false;
	}

	// load modules
	for (int i=0; i<(int)projSettings->subSettings.size(); ++i) {
		PROJECT_SETTINGS * module = projSettings->subSettings[i];
		if (module->name != "module") {
			continue;
		}

		if (module->subSettings.empty()) {
			AddModule(module->value, module->value);
		} else {
			// TODO: support line numbers and error handling
			AddModule(module->GetValue("name"), module->GetValue("source"));
		}
	}

	return true;
}

/************************************************************************/
/* tavr::PROJECT_SETTINGS::GetValue
/*
/************************************************************************/
std::string PROJECT_SETTINGS::GetValue(const char * path, const std::string & def) const
{
	const PROJECT_SETTINGS * settings = GetSettings(path);
	if (settings != NULL) {
		return settings->value;
	}
	return def;
}

/************************************************************************/
/* tavr::PROJECT_SETTINGS::GetSettings
/*
/************************************************************************/
const PROJECT_SETTINGS * PROJECT_SETTINGS::GetSettings(const char * path) const
{
	const char * value = path;

	std::string name;
	const PROJECT_SETTINGS * cur = this;
	while (cur != NULL) {
		const char * slashPos = strchr(value, '/');
		if (slashPos != NULL) {
			name = string(value, slashPos);
		} else {
			name = value;
		}

		const PROJECT_SETTINGS * next = NULL;
		for (int i=0; i<(int)cur->subSettings.size(); ++i) {
			if (cur->subSettings[i]->name == name) {
				next = cur->subSettings[i];
				break;
			}
		}

		if (slashPos == NULL) {
			return next;
		}

		if (next == NULL) {
			return NULL;
		}

		value = slashPos + 1;
		cur = next;
	}

	return NULL;
}

/************************************************************************/
/* tavr::PROJECT::GetSettings
/*
/************************************************************************/
const PROJECT_SETTINGS *  PROJECT::GetSettings(const char * value) const
{
	return settings->GetSettings(value);
}

/************************************************************************/
/* tavr::PROJECT::AddModule
/*
/************************************************************************/
bool PROJECT::AddModule(const std::string & name, const std::string & source)
{
	if (modules.find(name) != modules.end()) {
		return false;
	}

	MODULE * module = alloc.Alloc<MODULE>();
	module->name	= name;
	module->source	= source;

	modules.insert(make_pair(name, module));
	return true;
}

/************************************************************************/
/* tavr::PROCEDURE::Remove
/*
/************************************************************************/
void METHOD::Remove(void)
{
	if (type != METHOD::TYPE_PROCEDURE) {
		return ;
	}

	assert(owner.names != NULL);
	for (int i=(int)owner.names->procs.size() - 1; i>=0; --i) {
		if (owner.names->procs[i] == this) {
			owner.names->procs.erase(owner.names->procs.begin() + i);
			return ;
		}
	}
}

/************************************************************************/
/* tavr::PROJECT::Compile
/*
/************************************************************************/
bool PROJECT::Compile(void)
{
	int	successCount = 0;
	int errorCount = 0;

	FOR (MODULE_MAP, moduleIter, modules) {
		// TODO: support file versions
		//if (module->second.state != MODULE::STATE_NOT_LOADED) {
		//	continue;
		//}

		MODULE * module = moduleIter->second;
		msg["info"]		= module->name;

		ifstream source(module->source.c_str());
		if (!source) {
			msg["name"] = module->source;
			PRINT_MSG(TVR1009, -1);
			++errorCount;
			continue;
		}

		// TODO: clear all references to this module
		for (MODULE_MEMBER * mem = module->members; mem != NULL; mem = mem->next) {
			switch (mem->type) {
				case MODULE_MEMBER::TYPE_METHOD:
					mem->info.proc->Remove();
					break;

				default:
					assert(!"Unknown module member type");
					break;
			}
		}

		module->treeAlloc.Clear();

		TAVR_LOADER ldr(this);
		ldr.SetModule(module);
		ldr.Init(source, GetError());

		GetError() << module->name << endl;

		ldr.Compile();

		if (ldr.HasErrors()) {
			errorCount += ldr.GetErrorCount();

			// free unused data
			module->treeAlloc.Clear();
			continue;
		}

#ifdef _DEBUG
		VerifyTree(module->tree);
#endif

		module->state = MODULE::STATE_TREE_WAS_LOADED;

		// basic analysis: collect type names (first pass)
		LOOKUP_CONTEXT lookupContext;
		lookupContext.module = module;
		lookupContext.curNamespace = globalName;
		int res = AnalyzeTypes(lookupContext, module->tree);
		if (res == 0) {
			module->state = MODULE::STATE_TYPES_WERE_PARSED;
		} else {
			errorCount += res;
		}
	}

	// collect method declarations (second pass)
	FOR (MODULE_MAP, moduleIter, modules) {
		LOOKUP_CONTEXT lookupContext;
		lookupContext.module = moduleIter->second;
		lookupContext.curNamespace = globalName;
		msg["info"]		= lookupContext.module->name;

		if (lookupContext.module->state != MODULE::STATE_TYPES_WERE_PARSED) {
			continue;
		}

		int res = AnalyzeMethods(lookupContext, lookupContext.module->tree);
		if (res == 0) {
			lookupContext.module->state = MODULE::STATE_METHODS_WERE_PARSED;
		} else {
			errorCount += res;
		}
	}

	// "MAIN" FUNCTION CONSTRAINTS
	if (errorCount == 0) {
		METHOD method;
		method.type = METHOD::TYPE_PROCEDURE;
		method.name = toSTR_ID("main");
		method.owner.names = globalName;
		method.result.SetType(defType[DEF_TYPE_INT]);

		METHOD::ARG args;
		args.type.SetType(defType[DEF_TYPE_STRING]);
		++args.type.array;
		method.args.push_back(args);

		LOOKUP_CONTEXT context;
		context.curNamespace = globalName;

		bool ambiguity;
		METHOD * mainProc = MethodLookup(&method, context, ambiguity);
		if (mainProc == NULL) {
			assert(!ambiguity);
			msg["name"] = Type2Str(&method);
			PRINT_MSG(TVR1043, -1);
			++errorCount;
		} else if (!method.result.IsEqual(mainProc->result)) {
			msg["name"] = Type2Str(mainProc);
			PRINT_MSG(TVR1044, -1);
			++errorCount;
		} else {
			assert(!ambiguity);
		}
	}

	if (target.PreGenerate()) {
		std::vector<int> moduleErrors(modules.size());
		GENERATE_INFO_MAP generateInfo;

		int index = 0;

		// generate code and check method implementations (third pass)
		FOR (MODULE_MAP, moduleIter, modules) {
			MODULE * module = moduleIter->second;
			msg["info"]		= module->name;

			if (module->state != MODULE::STATE_METHODS_WERE_PARSED) {
				++index;
				continue;
			}

			int errors = Generate(generateInfo, module);
			if (errors != 0) {
				errorCount += errors;
				moduleErrors[index] = errors;
			}

			++index;
		}

		// second pass (calculate sizes and generate targets)
		index = 0;
		FOR (MODULE_MAP, moduleIter, modules) {
			MODULE * module = moduleIter->second;
			msg["info"]		= module->name;

			if (module->state != MODULE::STATE_METHODS_WERE_PARSED) {
				++index;
				continue;
			}

			if (moduleErrors[index] != 0) {
				++index;
				continue;
			}

			int errors = PostGenerate(generateInfo, module);
			if (errors != 0) {
				errorCount += errors;
			}

			++index;
		}

		// target generate
		if (errorCount == 0) {
			index = 0;
			FOR (MODULE_MAP, moduleIter, modules) {
				MODULE * module = moduleIter->second;
				msg["info"]		= module->name;

				if (module->state != MODULE::STATE_METHODS_WERE_PARSED) {
					++index;
					continue;
				}

				if (moduleErrors[index] != 0) {
					++index;
					continue;
				}

				std::string outputName;
				if (target.Generate(module, outputName)) {
					++successCount;
					module->state = MODULE::STATE_COMPLETED;
				} else {
					msg["name"] = outputName;
					PRINT_MSG(TVR1010, -1);
					++errorCount;
				}

				++index;
			}
		}

		std::string outputName;
		if (!target.GenerateTypeinfo(outputName)) {
			msg["name"] = outputName;
			PRINT_MSG(TVR1010, -1);
			++errorCount;
		}

		if (errorCount == 0) {
			errorCount += !target.PostGenerate();
		}
	} else {
		++errorCount;
	}

	cout << "--------------------------" << endl;
	cout << successCount << " success, " << errorCount << " errors" << endl;

	return true;
}

/************************************************************************/
/* tavr::MODULE::NewMethod
/*
/************************************************************************/
METHOD * MODULE::NewMethod(void)
{
	METHOD * proc = memAlloc.Alloc<METHOD>();

	MODULE_MEMBER * mem = memAlloc.Alloc<MODULE_MEMBER>();
	mem->type		= MODULE_MEMBER::TYPE_METHOD;
	mem->info.proc	= proc;
	mem->next		= members;
	members			= mem;

	return proc;
}

/************************************************************************/
/* tavr::MODULE::NewType
/*
/************************************************************************/
TYPE * MODULE::NewType(void)
{
	TYPE * type = memAlloc.Alloc<TYPE>();

	MODULE_MEMBER * mem = memAlloc.Alloc<MODULE_MEMBER>();
	mem->type		= MODULE_MEMBER::TYPE_TYPE;
	mem->info.type	= type;
	mem->next		= members;
	members			= mem;

	return type;
}

/************************************************************************/
/* tavr::PROJECT::AnalyzeAttribs
/*
/************************************************************************/
int PROJECT::AnalyzeAttribs(const NODE * tree, ATTRIBS_F & attribs)
{
	ATTRIBS nextAttrib = ATTRIBS();
	for (; tree != NULL; tree = tree->next) {
		bool isAccessAttrib = false;
		switch (tree->type) {
			case NODE::TYPE_ATTRIB_EXTERN:		nextAttrib = ATTRIB_EXTERN; break;
			case NODE::TYPE_ATTRIB_STATIC:		nextAttrib = ATTRIB_STATIC; break;
			case NODE::TYPE_ATTRIB_PUBLIC:		nextAttrib = ATTRIB_PUBLIC;		isAccessAttrib = true; break;
			case NODE::TYPE_ATTRIB_PRIVATE:		nextAttrib = ATTRIB_PRIVATE;	isAccessAttrib = true; break;
			case NODE::TYPE_ATTRIB_PROTECTED:	nextAttrib = ATTRIB_PROTECTED;	isAccessAttrib = true; break;
			default:							assert(!"Unknown attrib type");	break;
		}

		if (attribs.is(nextAttrib)
		|| isAccessAttrib && attribs.is(ATTRIB_ACCESS_MASK))
		{
			msg["name"] = tree->name.str();
			PRINT_MSG(TVR1049, tree->line);
			return 1;
		}

		attribs.set(nextAttrib);
	}

	return 0;
}

/************************************************************************/
/* tavr::PROJECT::AnalyzeTypes
/*
/************************************************************************/
int PROJECT::AnalyzeTypes(const LOOKUP_CONTEXT & lookupContext, const NODE * tree)
{
	int res = 0;

	MODULE * module = lookupContext.module;
	assert(module != NULL);

	std::vector<const NODE *> decls;
	decls.reserve(64);
	for (const NODE * declTree = tree; declTree != NULL; declTree = declTree->next) {
		decls.push_back(declTree);
	}

	for (int declIndex = (int)decls.size() - 1; declIndex >= 0; --declIndex) {
		tree = decls[declIndex];

		switch (tree->type) {
			case NODE::TYPE_AUTO:
			case NODE::TYPE_STRUCT:
			case NODE::TYPE_EVENT:
				{
					const bool isAuto = (tree->type == NODE::TYPE_AUTO);
					const bool isEvent = (tree->type == NODE::TYPE_EVENT);

					TYPE * type				= module->NewType();

					type->subtype			= isAuto
						? TYPE::SUBTYPE_AUTO : isEvent
						? TYPE::SUBTYPE_EVENT
						: TYPE::SUBTYPE_STRUCT;

					type->name				= tree->name;
					type->declInfo.module	= lookupContext.module;
					type->declInfo.line		= tree->line;

					type->parentNamespace	= lookupContext.curNamespace;

					res += AnalyzeAttribs(tree->subnode->subnode, type->attribs);

					NAMESPACE * curNames	= lookupContext.curNamespace;
					for (int i=(int)curNames->types.size() - 1; i>=0; --i) {
						if (tree->name == curNames->types[i]->name) {
							msg["name"] = tree->name.str();
							PRINT_MSG(TVR1016, tree->line);
							PRINT_MSG_INFO(TVR_SEEDECL_INFO, curNames->types[i]->declInfo);
							++res;
							break;
						}
					}

					if (isEvent) {
						type->eventId = eventIdCntr++;
					}

					if (res == 0) {
						curNames->types.push_back(type);
					}
				}
				break;
		}
	}

	return res;
}

/************************************************************************/
/* tavr::PROJECT::AnalyzeMethodsProc
/*
/************************************************************************/
int PROJECT::AnalyzeMethodsProc(LOOKUP_CONTEXT & context, const NODE * procDecl, METHOD::TYPE procType, METHOD::OWNER owner, METHOD *& proc)
{
	MODULE * module = context.module;

	proc = module->NewMethod();

	const NODE * attribListArg	= procDecl->subnode;
	const NODE * procHeader		= attribListArg->next->subnode;
	const NODE * procImplArg	= attribListArg->next->next;

	assert(procHeader->subnode != NULL);
	const NODE * procArgList = procHeader->subnode->next;

	proc->name = procHeader->name;
	proc->declInfo.module	= context.module;
	proc->declInfo.line		= procHeader->line;

	proc->type				= procType;
	proc->owner				= owner;

	// VERIFY ATTRIBS
	int res = 0;
	
	// CHECK REDEFINITION
	EVALUATE eval;

	// make sure that there is no element with such name in current namespace
	eval.context.module			= context.module;
	eval.context.curType		= context.curType;
	eval.context.curNamespace	= context.GetParentNamespace();
	PLACEMENT_INFO placementInfo;
	if (NameLookup(proc->name, eval.context, eval.typeinfo, placementInfo)) {
		bool isError = false;
		if (context.curType != NULL) {
			switch (eval.typeinfo.type) {
				case TYPEINFO::TYPE_PROC_NAME:
					// do not allow to declare items with same name as out value
					switch (eval.typeinfo.info.procName->type) {
						case METHOD::TYPE_OUT:
						case METHOD::TYPE_AUTO_PROC:
							isError = true;
							break;
					}
					break;

				case TYPEINFO::TYPE_TYPE:
					if (proc->type != METHOD::TYPE_CONSTRUCTOR && context.curType == eval.typeinfo.info.type) {
						isError = true;
					}
					break;

				default: isError = true; break;
			}
		} else {
			isError = (eval.typeinfo.type != TYPEINFO::TYPE_PROC_NAME);
		}
		if (isError) {
			msg["name"] = proc->name.str();
			PRINT_MSG(TVR1016, proc->declInfo.line);
			PRINT_MSG_INFO(TVR_SEEDECL_INFO, placementInfo.declInfo);
			return 1;
		}
	}

	// RESULT TYPE
	eval.context			= context;

	// hack for constructors (constructors do not have result types)
	int evalErrors = 0;
	if (procHeader->subnode->type != NODE::TYPE_ARGUMENT) {
		if (evalErrors = Evaluate(procHeader->subnode, eval)) {
			return evalErrors;
		}

		if (eval.typeinfo.type != TYPEINFO::TYPE_TYPE) {
			msg["name"]		= proc->name.str();
			PRINT_MSG(TVR1012, proc->declInfo.line);
			return 1;
		}
	} else {
		eval.typeinfo.SetType(defType[DEF_TYPE_VOID]);
	}

	proc->result = eval.typeinfo;

	// ARGUMENTS
	for (const NODE * procArg = procArgList; procArg != NULL; procArg = procArg->next) {
		METHOD::ARG arg;
		arg.name = procArg->name;

		// change context to avoid local methods lookup
		TYPE * curType = NULL;
		NAMESPACE * curNamespace = NULL;
		if (context.curType != NULL) {
			curNamespace = eval.context.GetParentNamespace();
			Swap(curType,		eval.context.curType);
			Swap(curNamespace,	eval.context.curNamespace);
		}

		// CHECK FOR 'inout/out' PROPERTY
		const NODE * argType = procArg;
		
		switch (argType->subnode->type) {
			case NODE::TYPE_EXPR_INOUT:
				arg.subtype = METHOD::ARG::SUBTYPE_INOUT;
				argType = argType->subnode;
				break;
			case NODE::TYPE_EXPR_OUT:
				arg.subtype = METHOD::ARG::SUBTYPE_OUT;
				argType = argType->subnode;
				break;
			default:
				arg.subtype = METHOD::ARG::SUBTYPE_DEFAULT;
				break;
		}

		int evalErrors = Evaluate(argType->subnode, eval);

		if (arg.subtype != METHOD::ARG::SUBTYPE_DEFAULT) {
			eval.typeinfo.isRef = true;
		}

		if (curType != NULL) {
			Swap(curType,		eval.context.curType);
			Swap(curNamespace,	eval.context.curNamespace);
		}

		if (evalErrors != 0) {
			return res + evalErrors;
		}

		if (eval.typeinfo.type != TYPEINFO::TYPE_TYPE) {
			// calc argument count
			int index = -1;
			int count = 0;
			for (const NODE * arg = procArgList; arg != NULL; arg = arg->next) {
				if (arg == procArg) {
					index = count;
				}
				++count;
			}
			assert(index != -1);

			msg["number"]	= ToString(count - index);
			PRINT_MSG(TVR1011, proc->declInfo.line);
			return res + 1;
		}
		arg.type = eval.typeinfo;

		// check that name redefinition
		if (arg.name != emptyString) {
			for (int i=(int)proc->args.size() - 1; i>=0; --i) {
				if (proc->args[i].name == arg.name) {
					msg["name"] = arg.name.str();
					PRINT_MSG(TVR1016, procArg->line);
					return 1;
				}
			}
		}

		proc->args.push_back(arg);
	}

	// arguments need reverse
	std::reverse(proc->args.begin(), proc->args.end());

	// check whether there is no method with same signature
	bool ambiguity;

	METHOD * alreadyExists = MethodLookup(proc, context, ambiguity, LOOKUP_TYPE_LOCALS_ONLY);
	if (alreadyExists != NULL) {
		// check whether difference only inf 'ref' arg types
		bool onlyRefs = false;
		for (int i=(int)proc->args.size() - 1; i >= 0; --i) {
			if (proc->args[i].subtype != alreadyExists->args[i].subtype) {
				onlyRefs = true;
				break;
			}
		}

		msg["name"] = Type2Str(proc);

		if (onlyRefs) {
			PRINT_MSG(TVR1036, proc->declInfo.line);
		} else if (proc->result < alreadyExists->result || alreadyExists->result < proc->result) {
			msg["name2"] = Type2Str(proc->result);
			PRINT_MSG(TVR1037, proc->declInfo.line);
		} else {
			PRINT_MSG(TVR1016, proc->declInfo.line);
		}

		msg["name"] = Type2Str(alreadyExists);
		PRINT_MSG_INFO(TVR_SEEDECL_INFO, alreadyExists->declInfo);
		return res + 1;
	}

	assert(!ambiguity);

	if (res != 0) {
		return res;
	}

	// PROC IMPL
	proc->impl				= NULL;
	if (procImplArg != NULL) {
		proc->impl			= procImplArg->subnode;
	}

	// VERIFY ATTRIBS
	if (proc->IsMethod()) {
		proc->attribs.set(owner.type->attribs.value() & (ATTRIB_EXTERN | ATTRIB_STATIC));
	}

	res += AnalyzeAttribs(attribListArg->subnode, proc->attribs);

	if (proc->attribs.is(ATTRIB_EXTERN) && proc->impl != NULL) {
		msg["name"] = Type2Str(proc);
		PRINT_MSG(TVR1047, proc->declInfo.line);
		++res;
	} else if (!proc->attribs.is(ATTRIB_EXTERN) && proc->impl == NULL) {
		msg["name"] = Type2Str(proc);
		PRINT_MSG(TVR1048, proc->declInfo.line);
		++res;
	}

	return res;
}

/************************************************************************/
/* tavr::PROJECT::AnalyzeMemberVar
/*
/************************************************************************/
int PROJECT::AnalyzeMemberVar(TYPE * type, const NODE * structDecl, const LOOKUP_CONTEXT & lookupContext, bool isAuto)
{
	int res = 0;

	EVALUATE eval;
	eval.context = lookupContext;

	NODE * attribList	= structDecl->subnode;
	NODE * varType		= attribList->next;
	NODE * nameList		= varType->next;

	// TODO: process attrib list

	res += Evaluate(varType->subnode, eval);

	if (res == 0 && (eval.typeinfo.type != TYPEINFO::TYPE_TYPE || eval.typeinfo.IsNull())) {
		if (eval.typeinfo.IsValue()) {
			PRINT_MSG(TVR1007, varType->subnode->line);
		} else {
			msg["name"] = Type2Str(eval.typeinfo);
			PRINT_MSG(TVR1015, varType->subnode->line);
		}
		++res;
	}

	if (res != 0) {
		return res;
	}

	TYPE::FIELD field;
	field.type = eval.typeinfo;

	eval.context.curNamespace = NULL;
	eval.context.curType = type;
	assert(eval.context.curMethod == NULL);

	PLACEMENT_INFO placementInfo;
	for (const NODE * fieldNode = nameList->subnode; fieldNode != NULL; fieldNode = fieldNode->next) {
		field.name	= fieldNode->name;
		field.value	= fieldNode->subnode;
		field.line	= fieldNode->line;

		bool redefError = false;
		if (field.name == type->name) {
			placementInfo.declInfo = type->declInfo;
			redefError = true;
		} else {
			redefError = NameLookup(field.name, eval.context, eval.typeinfo, placementInfo, LOOKUP_TYPE_LOCALS_ONLY);
		}

		if (redefError) {
			msg["name"] = field.name.str();
			PRINT_MSG(TVR1016, fieldNode->line);
			PRINT_MSG_INFO(TVR_SEEDECL_INFO, placementInfo.declInfo);
			return res + 1;
		}

		type->fields.push_back(field);
	}

	return res;
}

/************************************************************************/
/* tavr::TYPE::IsContainsTypesByRef
/*
/************************************************************************/
bool TYPE::DoesContainTypesByRef(std::set<const TYPE *> & processingInfo) const
{
	if (processingInfo.find(this) != processingInfo.end()) {
		return false;
	}

	processingInfo.insert(this);

	for (int i=(int)fields.size() - 1; i >= 0; --i) {
		const TYPEINFO & typeinfo = fields[i].type;
		if (typeinfo.IsTypeByRef()) {
			return true;
		}
		if (typeinfo.IsTypeByValue()) {
			if (typeinfo.info.type->DoesContainTypesByRef(processingInfo)) {
				return true;
			}
		}
	}

	processingInfo.erase(this);

	return false;
}

/************************************************************************/
/* tavr::PROJECT::AnalyzeMethodsOperator
/*
/************************************************************************/
int PROJECT::AnalyzeMethodsOperator(LOOKUP_CONTEXT & context, const NODE * procDecl, METHOD *& proc)
{
	assert(context.curType != NULL);
	METHOD::OWNER owner;
	owner.type = context.curType;

	int res = AnalyzeMethodsProc(context, procDecl, METHOD::TYPE_OPERATOR, owner, proc);
	if (res != 0) {
		return res;
	}

	int desiredArgCount = -1;
	if (proc->name == assignOpString) {
		if (owner.type->IsTypeByRef()) {
			PRINT_MSG(TVR1052, proc->declInfo.line);
			return 1;
		}

		desiredArgCount = 1;

		if (proc->args.size() == 1) {
			// check whether it's default assign operator
			TYPEINFO & argType = proc->args[0].type;
			if (argType.array == 0 && argType.info.type == proc->owner.type) {
				proc->owner.type->specs[TYPE::SPEC_DEF_OP_ASSIGN] = proc;
			}
		}
	} else if (proc->name == castOpString) {
		desiredArgCount = 0;
	} else {
		assert(!"Unknown operator");
	}

	// check operator arg count
	if ((int)proc->args.size() != desiredArgCount) {
		msg["name"] = Type2Str(proc);
		if ((int)proc->args.size() < desiredArgCount) {
			PRINT_MSG(TVR1040, proc->declInfo.line);
		} else {
			assert((int)proc->args.size() > desiredArgCount);
			PRINT_MSG(TVR1041, proc->declInfo.line);
		}
		return 1;
	}

	return 0;
}

/*************************************************************************/
/* PROJECT::AnalyzeOut()                                                 */
/*                                                                       */
/*************************************************************************/
int PROJECT::AnalyzeOut(LOOKUP_CONTEXT & context, const NODE * tree, TYPE * type)
{
	// verify that there is no method with such name
	METHOD method;
	method.name = tree->name;

	TYPEINFO ti;
	PLACEMENT_INFO pi;
	LOOKUP_CONTEXT ctxt;
	ctxt.module  = context.module;
	ctxt.curType = type;
	if (NameLookup(method.name, ctxt, ti, pi, LOOKUP_TYPE_LOCALS_ONLY)) {
		msg["name"] = method.name.str();
		PRINT_MSG(TVR1016, tree->subnode->line);

		if (ti.type == TYPEINFO::TYPE_PROC_NAME) {
			msg["name"] = Type2Str(ti);
		}
		PRINT_MSG_INFO(TVR_SEEDECL_INFO, pi.declInfo);
		return 1;
	}

	EVALUATE eval;
	eval.context = context;
	int res = Evaluate(tree->subnode, eval);
	if (res > 0) {
		return res;
	}

	if (!eval.typeinfo.IsEvent()) {
		msg["name"] = Type2Str(eval.typeinfo);
		PRINT_MSG(TVR1065, tree->subnode->line);
		return 1;
	}

	TYPE * eventType = eval.typeinfo.info.type;

	// create out-method for each event contructor
	method.type = METHOD::TYPE_OUT;
	method.owner.type		= type;
	method.declInfo.module  = context.module;
	method.declInfo.line	= tree->subnode->line;
	method.result.SetType(defType[DEF_TYPE_VOID]);

	for (int i=0; i<(int)eventType->methods.size(); ++i) {
		METHOD * constr = eventType->methods[i];
		if (constr->type != METHOD::TYPE_CONSTRUCTOR) {
			continue;
		}

		METHOD * mt = context.module->NewMethod();
		*mt = method;
		mt->args = constr->args;
		type->methods.push_back(mt);
	}

	type->outTable.insert(make_pair(method.name, eventType));

	return 0;
}

/*************************************************************************/
/* PROJECT::AnalizeSubauto()                                             */
/*                                                                       */
/*************************************************************************/
int PROJECT::AnalizeSubauto(LOOKUP_CONTEXT & lookupContext, const NODE * tree, TYPE * type)
{
	// create subauto proc, data & etc.
	SUBAUTO_INFO si;
	si.id = subautoUIdCntr++;

	ecg::STR_ID name = tree->name;
	if (name == emptyString) {
		// generate unique name
		name = toSTR_ID("auto^" + ToString(si.id));
	} else {
		LOOKUP_CONTEXT ctxt;
		ctxt.module  = lookupContext.module;
		ctxt.curType = type;

		TYPEINFO ti;
		PLACEMENT_INFO pi;
		if (NameLookup(name, ctxt, ti, pi, LOOKUP_TYPE_LOCALS_ONLY)) {
			msg["name"] = name.str();
			PRINT_MSG(TVR1016, tree->subnode->line);

			if (ti.type == TYPEINFO::TYPE_PROC_NAME) {
				msg["name"] = Type2Str(ti);
			}
			PRINT_MSG_INFO(TVR_SEEDECL_INFO, pi.declInfo);
			return 1;
		}
	}

	// create new auto proc
	METHOD * method = lookupContext.module->NewMethod();
	method->name = name;
	method->owner.type = type;
	method->type = METHOD::TYPE_AUTO_PROC;
	method->result.SetType(defType[DEF_TYPE_VOID]);
	method->impl = tree;
	type->methods.push_back(method);

	si.proc = method;
	type->subautos.push_back(si);

	return 0;
}

/************************************************************************/
/* tavr::PROJECT::AnalyzeMethods
/*
/************************************************************************/
int PROJECT::AnalyzeMethods(LOOKUP_CONTEXT & lookupContext, const NODE * tree)
{
	int res = 0;

	std::vector<const NODE *> decls;
	decls.reserve(64);
	for (const NODE * cur = tree; cur != NULL; cur = cur->next) {
		decls.push_back(cur);
	}

	for (int declIndex = (int)decls.size() - 1; declIndex >= 0; --declIndex) {
		tree = decls[declIndex];

		switch (tree->type) {
			case NODE::TYPE_DECL_PROC:
				{
					METHOD * proc = NULL;

					METHOD::OWNER owner;
					owner.names = lookupContext.curNamespace;
					res += AnalyzeMethodsProc(lookupContext, tree, METHOD::TYPE_PROCEDURE, owner, proc);

					assert(res != 0 || proc != NULL);
					if (res != 0) {
						break;
					}

					lookupContext.curNamespace->procs.push_back(proc);
				}
				break;

			case NODE::TYPE_AUTO:
			case NODE::TYPE_STRUCT:
			case NODE::TYPE_EVENT:
				{
					const bool isEvent = tree->type == NODE::TYPE_EVENT;
					const bool isAuto = (tree->type == NODE::TYPE_AUTO || isEvent);

					TYPE * type = TypeLookup(tree->name, lookupContext, LOOKUP_TYPE_LOCALS_ONLY);
					assert(type != NULL);
					assert(type->parentNamespace == lookupContext.curNamespace);

					std::vector<const NODE *> typeDecls;
					typeDecls.reserve(64);
					for (const NODE * structDecl = tree->subnode->next->subnode; structDecl != NULL; structDecl = structDecl->next) {
						typeDecls.push_back(structDecl);
					}

					bool stepDeclarationError = false;
					bool outDeclarationError = false;

					// analyze type arguments
					for (int structDeclIndex = (int)typeDecls.size() - 1; structDeclIndex >= 0; --structDeclIndex) {
						const NODE * structDecl = typeDecls[structDeclIndex];

						switch (structDecl->type) {
							case NODE::TYPE_DECL_MEMBER_VAR:
								res += AnalyzeMemberVar(type, structDecl, lookupContext, isAuto);
								break;

							case NODE::TYPE_DECL_OPERATOR:
							case NODE::TYPE_DECL_CONSTR:
							case NODE::TYPE_DECL_METHOD:
								{
									const bool isConstr = (structDecl->type == NODE::TYPE_DECL_CONSTR);
									const bool isOperator = (structDecl->type == NODE::TYPE_DECL_OPERATOR);

									const NODE * attribList = structDecl->subnode;
									const NODE * procHeader	= attribList->next;
									const NODE * procImpl	= procHeader->next;

									NAMESPACE * oldNames	= NULL;
									TYPE *		oldType		= type;
									Swap(oldNames,	lookupContext.curNamespace);
									Swap(oldType,	lookupContext.curType);

									METHOD * proc = NULL;
									if (!isOperator) {
										METHOD::TYPE procType;
										if (!isConstr) {
											procType = METHOD::TYPE_METHOD;
										} else {
											procType = METHOD::TYPE_CONSTRUCTOR;
										}

										METHOD::OWNER owner;
										owner.type = type;

										res += AnalyzeMethodsProc(lookupContext, structDecl, procType, owner, proc);
									} else {
										res += AnalyzeMethodsOperator(lookupContext, structDecl, proc);
									}

									Swap(oldNames,	lookupContext.curNamespace);
									Swap(oldType,	lookupContext.curType);

									if (res == 0 && isConstr && proc->name != type->name) {
										TYPEINFO ti;
										ti.SetType(type);
										msg["name"]  = Type2Str(ti);
										msg["name2"] = proc->name.str();
										PRINT_MSG(TVR1039, structDecl->subnode->next->subnode->line);
										++res;
										break;
									}

									if (res == 0) {
										type->methods.push_back(proc);
									}
								}
								break;

							case NODE::TYPE_STEP:
							case NODE::TYPE_LOOP:
							case NODE::TYPE_FIXED:
							case NODE::TYPE_SUBAUTO:
								if (isEvent) {
									if (!stepDeclarationError) {
										stepDeclarationError = true;

										TYPEINFO ti;
										ti.SetType(type);
										msg["name"] = Type2Str(ti);
										switch (structDecl->type) {
											case NODE::TYPE_STEP:    msg["name"] = "step"; break;
											case NODE::TYPE_LOOP:    msg["name"] = "loop"; break;
											case NODE::TYPE_FIXED:   msg["name"] = "fixed"; break;
											case NODE::TYPE_SUBAUTO: msg["name"] = "subauto"; break;
											default: assert(!"Unknown block type"); break;
										}
										PRINT_MSG(TVR1057, structDecl->line);
										++res;
									}
									break;
								}

								if (structDecl->type != NODE::TYPE_SUBAUTO) {
									type->steps.push_back(structDecl);
									break;
								}

								res += AnalizeSubauto(lookupContext, structDecl, type); 
								break;

							case NODE::TYPE_OUT:
								if (isEvent) {
									if (!outDeclarationError) {
										outDeclarationError = true;

										TYPEINFO ti;
										ti.SetType(type);
										msg["name"] = Type2Str(ti);
										msg["type"] = "out";
										PRINT_MSG(TVR1057, structDecl->line);
									}
								}

								res += AnalyzeOut(lookupContext, structDecl, type);
								break;

							default:
								assert(!"Unknown struct declaration type");
								break;
						}
					}

					// create special automaton procedure
					if (!type->steps.empty() || !type->subautos.empty() || type->subtype == TYPE::SUBTYPE_AUTO) {
						METHOD * method = lookupContext.module->NewMethod();

						method->type			= METHOD::TYPE_AUTO_PROC;
						method->name			= autoString;
						method->declInfo.module = lookupContext.module;
						method->declInfo.line	= tree->line;
						method->owner.type		= type;
						method->attribs.set(type->attribs.value() & (ATTRIB_EXTERN | ATTRIB_STATIC));
						method->result.SetType(defType[DEF_TYPE_VOID]);

						type->specs[TYPE::SPEC_AUTO_PROC] = method;
						type->methods.push_back(method);
					}

					std::reverse(type->fields.begin(), type->fields.end());

					// check for default and copy constructors
					bool anyConstrFound = false;
					for (int i=(int)type->methods.size() - 1; i >= 0; --i) {
						METHOD * method = type->methods[i];
						if (method->type == METHOD::TYPE_CONSTRUCTOR) {
							anyConstrFound = true;

							if (method->args.empty()) {
								type->specs[TYPE::SPEC_DEF_CONSTR] = method;
							} else if (method->args.size() == 1) {
								METHOD::ARG & arg = method->args[0];

								if (arg.type.array == 0 && arg.type.info.type == type) {
									type->specs[TYPE::SPEC_COPY_CONSTR] = method;

									if (type->IsTypeByValue() && !arg.type.isRef) {
										msg["name"] = Type2Str(method);
										PRINT_MSG(TVR1042, method->declInfo.line);
										++res;
									}
								}
							}
						}
					}

					#define ADD_DEF_METHOD(NAME, TYPE, SPEC_INDEX) \
						METHOD * method = lookupContext.module->NewMethod(); \
						method->name = NAME; \
						method->owner.type = type; \
						method->type = TYPE; \
						method->result.SetType(defType[DEF_TYPE_VOID]); \
						method->attribs.set(type->attribs.value() & ATTRIB_EXTERN); \
						\
						type->methods.push_back(method); \
						type->specs[SPEC_INDEX] = method;

					// add default constructor if needed
					// 1. always create default constructor for all structs
					// 2. add def-constructor when auto&event don't have any constroctor
					if (!isAuto && type->specs[TYPE::SPEC_DEF_CONSTR] == NULL
					|| !anyConstrFound) {
						ADD_DEF_METHOD(type->name, METHOD::TYPE_CONSTRUCTOR, TYPE::SPEC_DEF_CONSTR);

						if (isEvent) {
							assert(isAuto);
							// add events default contructor, it should have the same args as fields
							for (int i=0, len = (int)type->fields.size(); i<len; ++i) {
								TYPE::FIELD & field = type->fields[len - i - 1];

								METHOD::ARG arg;
								arg.name = field.name;
								arg.type = field.type;
								method->args.push_back(arg);
							}

							// event's default constructor could be non-default
							type->specs[TYPE::SPEC_DEF_CONSTR]   = NULL;
							type->specs[TYPE::SPEC_EVENT_CONSTR] = method;
						}
					}

					// add copy constructor if needed (always create copy constructors for structures
					// with auto proc, because there could be relax stack)
					std::set<const TYPE *> processingInfo;
					bool areCopyOpsNeeded = type->DoesContainTypesByRef(processingInfo)
						|| type->specs[TYPE::SPEC_AUTO_PROC] != NULL;

					// add destructor (always add destructor to auto types)
					if ((areCopyOpsNeeded || isAuto) && type->specs[TYPE::SPEC_DESTR] == NULL) {
						ADD_DEF_METHOD(toSTR_ID("~" + type->name.str()), METHOD::TYPE_DESTRUCTOR, TYPE::SPEC_DESTR);
					} else {
						assert(!isAuto || type->specs[TYPE::SPEC_DESTR] == NULL);
					}

					if (!isAuto) {
						#define DEF_COPY_METHOD(NAME, TYPE, SPEC_INDEX) \
							if (areCopyOpsNeeded && type->specs[SPEC_INDEX] == NULL) { \
								METHOD * method = lookupContext.module->NewMethod(); \
								method->name = NAME; \
								method->owner.type = type; \
								method->type = TYPE; \
								method->result.SetType(defType[DEF_TYPE_VOID]); \
								\
								METHOD::ARG arg; \
								/* ignore arg name to handle default copy constructor on code generation level (yep, it's a hack) */ \
								arg.name = emptyString; \
								arg.type.SetType(type); \
								arg.type.isRef = true; \
								method->args.push_back(arg); \
								\
								type->methods.push_back(method); \
								type->specs[SPEC_INDEX] = method; \
							}

						DEF_COPY_METHOD(type->name,				METHOD::TYPE_CONSTRUCTOR,	TYPE::SPEC_COPY_CONSTR);
						DEF_COPY_METHOD(assignOpString,			METHOD::TYPE_OPERATOR,		TYPE::SPEC_DEF_OP_ASSIGN);
					}
				}
				break;
		}
	}

	return res;
}

PROJECT::DEF_TYPE_INFO PROJECT::defTypeInfo[] = {
	{ "void",		-1 },
	{ "null",		-1 },
	{ "bool",		1  },
	{ "char",		2  },
	{ "sbyte",		1  },
	{ "byte",		1  },
	{ "short",		2  },
	{ "ushort",		2  },
	{ "int",		4  },
	{ "uint",		4  },
	{ "long",		8  },
	{ "ulong",		8  },
	{ "float",		4  },
	{ "double",		8  },
	{ "string",		-1 },
};

PROJECT::DEF_TYPE_SUBTYPE_INFO PROJECT::defTypeSubtype[] = {
	{ PROJECT::DEF_TYPE_VOID,		TYPE::SUBTYPE_VOID },
	{ PROJECT::DEF_TYPE_NULL,		TYPE::SUBTYPE_NULL },
	{ PROJECT::DEF_TYPE_BOOL,		TYPE::SUBTYPE_BOOLEAN },
	{ PROJECT::DEF_TYPE_CHAR,		TYPE::SUBTYPE_CHARACTER },
	{ PROJECT::DEF_TYPE_S_BYTE,		TYPE::SUBTYPE_INTEGRAL },
	{ PROJECT::DEF_TYPE_FLOAT,		TYPE::SUBTYPE_FLOATING },
	{ PROJECT::DEF_TYPE_STRING,		TYPE::SUBTYPE_AUTO },
};

/************************************************************************/
/* tavr::PROJECT::PROJECT
/*
/************************************************************************/
PROJECT::PROJECT()
	: error(), eventIdCntr(), subautoUIdCntr()
{
	ECG_STATIC_ASSERT(ECG_NUM_ELEMENTS(defTypeInfo) == DEF_TYPE_MAX);

	target.Init(this);

	emptyString				= toSTR_ID("<empty>");
	autoString				= toSTR_ID("auto");
	castOpString			= toSTR_ID("(T)");
	assignOpString			= toSTR_ID("=");
	nullString				= toSTR_ID("null");
	eventString				= toSTR_ID("event");
	trueString				= toSTR_ID("true");
	falseString				= toSTR_ID("false");

	globalName				= alloc.Alloc<NAMESPACE>();
	globalName->name		= toSTR_ID("global");

	// generate default types
	for (int i=DEF_TYPE_0, nextSubtypeIndex = 0; i < DEF_TYPE_MAX; ++i) {
		TYPE * type			= alloc.Alloc<TYPE>();

		if (defTypeSubtype[nextSubtypeIndex].defType == i) {
			++nextSubtypeIndex;
		}

		assert(nextSubtypeIndex <= ECG_NUM_ELEMENTS(defTypeSubtype));
		type->subtype		= defTypeSubtype[nextSubtypeIndex - 1].subtype;
		type->name			= toSTR_ID(defTypeInfo[i].name);

		if (i != DEF_TYPE_NULL) {
			globalName->types.push_back(type);
		}

		type->size			= defTypeInfo[i].size;

		defType[i]			= type;
	}

	// generate default type casts for arithmetic types
	ecg::STR_ID typeCastName = castOpString;
	for (int i=DEF_TYPE_CONVERTABLE_0; i < DEF_TYPE_CONVERTABLE_MAX; ++i) {
		for (int j=DEF_TYPE_CONVERTABLE_0; j < DEF_TYPE_CONVERTABLE_MAX; ++j) {
			METHOD * method	= alloc.Alloc<METHOD>();
			method->type	= METHOD::TYPE_OPERATOR;
			method->name	= typeCastName;
			method->owner.type = defType[i];
			method->result.SetType(defType[j]);

			defType[i]->methods.push_back(method);
		}
	}

	// STRING TYPE
	TYPE * type = defType[DEF_TYPE_STRING];
	type->attribs.set(ATTRIB_EXTERN);

	// length
	TYPE::FIELD field;
	field.name	= toSTR_ID("length");
	field.type.SetType(defType[DEF_TYPE_INT]);
	type->fields.push_back(field);

	// def constr
	METHOD * method = alloc.Alloc<METHOD>();
	method->name = type->name;
	method->owner.type = type;
	method->type = METHOD::TYPE_CONSTRUCTOR;
	method->result.SetType(defType[DEF_TYPE_VOID]);
	method->attribs.set(ATTRIB_EXTERN);
	type->methods.push_back(method);
	type->specs[TYPE::SPEC_DEF_CONSTR] = method;

	// def destr
	method = alloc.Alloc<METHOD>();
	method->name = toSTR_ID("~" + type->name.str());
	method->owner.type = type;
	method->type = METHOD::TYPE_DESTRUCTOR;
	method->result.SetType(defType[DEF_TYPE_VOID]);
	method->attribs.set(ATTRIB_EXTERN);
	type->methods.push_back(method);
	type->specs[TYPE::SPEC_DESTR] = method;
}

/************************************************************************/
/* tavr::PROJECT::Type2Str
/*
/************************************************************************/
std::string PROJECT::Type2Str(METHOD * method)
{
	TYPEINFO ti;
	ti.type = TYPEINFO::TYPE_PROC_NAME;
	ti.info.procName = method;
	return Type2Str(ti);
}

/************************************************************************/
/* tavr::PROJECT::Type2Str
/*
/************************************************************************/
std::string PROJECT::Type2Str(const TYPEINFO & ti)
{
	// TODO: support namespaces

	string res;
	res.reserve(128);
	switch (ti.type) {
		case TYPEINFO::TYPE_L_VALUE:
		case TYPEINFO::TYPE_R_VALUE:
		case TYPEINFO::TYPE_TYPE:
			assert(ti.info.type != NULL);
			res += ti.info.type->name.str();
			for (int i=0; i<ti.array; ++i) {
				res += "[]";
			}
			return res;

		case TYPEINFO::TYPE_NAMESPACE:
			res += ti.info.names->name.str();
			return res;

		case TYPEINFO::TYPE_PROC_NAME:
		case TYPEINFO::TYPE_MEMBER_PROC_NAME:
			{
				bool needName = true;
				bool needResult = true;

				if (ti.info.procName->type == METHOD::TYPE_CONSTRUCTOR) {
					needResult = false;
				}

				string namePrefix;
				if (ti.info.procName->IsMethod()) {
					TYPEINFO t;
					t.type = TYPEINFO::TYPE_TYPE;
					t.info.type = ti.info.procName->owner.type;
					namePrefix += Type2Str(t);
					namePrefix += ".";

					if (ti.info.procName->type == METHOD::TYPE_OPERATOR) {
						namePrefix += "operator";

						if (ti.info.procName->name == castOpString) {
							namePrefix += " ";
							namePrefix += Type2Str(ti.info.procName->result);
							needName = false;
							needResult = false;
						}
					}
				}

				if (needResult) {
					res += Type2Str(ti.info.procName->result);
					res += " ";
				}

				res += namePrefix;

				if (needName) {
					res += ti.info.procName->name.str();
				}

				res += "(";
				for (int i=0; i<(int)ti.info.procName->args.size(); ++i) {
					METHOD::ARG & arg = ti.info.procName->args[i];

					switch (arg.subtype) {
						case METHOD::ARG::SUBTYPE_INOUT: res += "inout ";	break;
						case METHOD::ARG::SUBTYPE_OUT:	 res += "out ";		break;
					}

					res += Type2Str(arg.type);
					if (i + 1 < (int)ti.info.procName->args.size()) {
						res += ",";
					}
				}
				res += ")";
			}
			break;

		default:
			assert(!"Unknown typeinfo type");
			break;
	}

	return res;
}

/************************************************************************/
/* tavr::PROJECT::PrintMsg
/*
/************************************************************************/
void PROJECT::PrintMsg(int msgId, int line)
{
	const MSG_HOLDER * msgHolder = TAVR_LOADER_GetMessageByIndex(msgId);
	std::string format;
	if (line == -1) {
		format = "{info}: ";
	} else {
		format = "{info}({line}): ";
	}
	format += msgHolder->format;

	msg["line"] = ecg::ToString(line);

	GetError() << ecg::Compose(format.c_str(), msg) << endl;
}

/************************************************************************/
/* tavr::PROJECT::PrintMsgInfo
/*
/************************************************************************/
void PROJECT::PrintMsgInfo(int msgId, const DECL_INFO & info)
{
	msg["info2"] = info.module->name;

	const MSG_HOLDER * msgHolder = TAVR_LOADER_GetMessageByIndex(msgId);
	std::string format;
	if (info.line == -1) {
		format = "\t{info2}: ";
	} else {
		format = "\t{info2}({line}): ";
	}
	format += msgHolder->format;

	msg["line"] = ecg::ToString(info.line);

	GetError() << ecg::Compose(format.c_str(), msg) << endl;
}

#ifdef _DEBUG
/************************************************************************/
/* tavr::PROJECT::VerifyTree
/*
/************************************************************************/
void PROJECT::VerifyTree(const NODE * tree)
{
	if (tree == NULL) {
		return ;
	}

	assert(tree->name.isValid());

	assert(
		tree->type == NODE::TYPE_ARGUMENT ||
		tree->type == NODE::TYPE_CODE_BLOCK ||
		tree->line != -1
	);

	VerifyTree(tree->next);
	VerifyTree(tree->subnode);
}
#endif

} // namespace tavr
