
#ifndef TAVR_H
#define TAVR_H

#pragma once

#include "tavr_target_c.h"

namespace ecg {
	class TAVR_PROJECT_LOADER;
}

namespace tavr {

// PREPARSE DATA
struct MODULE;
struct NAMESPACE;
struct TYPE;
struct METHOD;

typedef TARGET_C TARGET_T;

const int TYPE_BY_REF_SIZE	= 8;
const int TYPE_INT_SIZE		= 4;

struct NODE {
	enum TYPE {
		TYPE_UNKNOWN,
		TYPE_ARGUMENT,				// an argument in an argument list

		// naming
		TYPE_NAME,					// $string
		TYPE_SUBNAME,				// subnode "." $string
		
		// variable and method declarations
		TYPE_DECL_VAR,				// type : sub, arg * : next, var_name : arg, var_init : arg.sub
		TYPE_DECL_VAR_CONSTR,		// type a(arg1, arg2);
		TYPE_DECL_MEMBER_VAR,		// attrib_list type var_name_list
		
		TYPE_DECL_PROC,				// func_header [impl]
		TYPE_DECL_METHOD,			// attrib_list func_header impl
		TYPE_DECL_CONSTR,			// attrib_list constr_header impl
		TYPE_DECL_OPERATOR,			// attrib_list operator_header impl

		TYPE_PROC_HEADER,			// int main(int a)
		TYPE_CODE_BLOCK,			// sub : statements list

		// in/out specifications
		TYPE_EXPR_INOUT,			// "inout"
		TYPE_EXPR_OUT,				// "out"

		// attributes
		TYPE_ATTRIB_PUBLIC,			// "public"
		TYPE_ATTRIB_PRIVATE,		// "private"
		TYPE_ATTRIB_PROTECTED,		// "protected"
		TYPE_ATTRIB_STATIC,			// "static"
		TYPE_ATTRIB_EXTERN,			// "extern"

		// arithmetic
		TYPE_EXPR_ASSIGN,			// a = b
		TYPE_EXPR_ADD,				// a + b
		TYPE_EXPR_SUB,				// a - b
		TYPE_EXPR_MUL,				// a * b
		TYPE_EXPR_DIV,				// a / b
		TYPE_EXPR_MOD,				// a % b

		// unar operators
		TYPE_EXPR_NEG,				// - a
		TYPE_EXPR_POS,				// + a
		TYPE_EXPR_LOG_NOT,			// ! a
		TYPE_EXPR_BIT_NOT,			// ~ a

		// logic
		TYPE_EXPR_LOG_OR,			// '||'
		TYPE_EXPR_LOG_AND,			// '&&'
		TYPE_EXPR_EQ,				// '=='
		TYPE_EXPR_NOT_EQ,			// '!='
		TYPE_EXPR_LESS,				// '<'
		TYPE_EXPR_LESS_EQ,			// '<='
		TYPE_EXPR_GREATER,			// '>'
		TYPE_EXPR_GREATER_EQ,		// '>='
		TYPE_EXPR_IF,				// '?:'

		// increments
		TYPE_EXPR_PRE_INC,			// ++i
		TYPE_EXPR_PRE_DEC,			// --i
		TYPE_EXPR_POST_INC,			// i++
		TYPE_EXPR_POST_DEC,			// i--
		TYPE_EXPR_CALL,				// proc : sub ( args : next )

		// allocations and arrays
		TYPE_EXPR_NEW_AUTO,			// new MyAuto()
		TYPE_EXPR_NEW_ARRAY,		// new int[count], 
		TYPE_EXPR_ARRAY,			// val[index] or type[]

		// constants and types
		//TYPE_EXPR_REF,				// "ref"
		TYPE_EXPR_SIZEOF,			// "sizeof"
		TYPE_EXPR_TYPEOF,			// "typeof"

		TYPE_EXPR_TYPE_CAST,		// "(typename)"

		TYPE_EXPR_GLOBAL,			// "global"
		TYPE_EXPR_THIS,				// "this"
		TYPE_EXPR_EVENT,			// "event"

		TYPE_EXPR_NULL,				// "null"
		TYPE_EXPR_BOOLEAN,			// "true" or "false"
		TYPE_EXPR_CHAR,				// 'c'
		TYPE_EXPR_STRING,			// "hehe"
		TYPE_EXPR_INT,				// $integer
		TYPE_EXPR_FLOAT,			// $float
		TYPE_EXPR_DOUBLE,			// $double

		// statements
		TYPE_RETURN,				// "return" var : sub
		TYPE_BREAK,					// "break"
		TYPE_CONTINUE,				// "continue"
		TYPE_REPEAT,				// "repeat"
		TYPE_GOTO,					// "goto" someStep

		TYPE_SEND,					// "send" event ["to" obj]
		TYPE_POST,					// "post" event ["to" obj]
		TYPE_REG_SEND,				// a.e1 += send b;
		TYPE_REG_POST,				// a.e1 += post b;
		TYPE_UNREG,					// a.e1 -= b;
		TYPE_RELAX,					// "relax" [stepA]
		TYPE_RELAX_STEP,			// "relax" "step"

		TYPE_SUBAUTO,				// "subauto" [name] [code_block]
		TYPE_STEP,					// "step" [name] [code_block]
		TYPE_LOOP,					// "loop" [name] code_block
		TYPE_FIXED,					// "fixed" [name] [code_block]
		TYPE_OUT,					// "out" E1 action_e1;

		TYPE_IF,					// "if" (cond : sub) then : sub.next "else" else : sub.next.next
		TYPE_FOR,					// "for" (init; cond; inc) impl
		TYPE_WHILE,					// "while" (cond) impl
		TYPE_DO_WHILE,				// "do" impl "while" (cond)

		TYPE_ON,					// "on" event { stat* }
		TYPE_OTHERWISE,				// "otherwise" { stat* }

		// type declarations
		TYPE_STRUCT,				// 'struct' name { struct_decl* }
		TYPE_AUTO,					// 'auto' name { auto_decl* }
		TYPE_EVENT,					// 'event' name { event_decl* }
	};

	TYPE					type;

	ecg::STR_ID				name;
	int						line;

	NODE *					next;
	NODE *					subnode;

	NODE(void)
		: type(TYPE_UNKNOWN), name(), line(), next(NULL), subnode(NULL)
	{}
};

// TYPES AND LOOKUP CONTEXTS
struct TYPEINFO {
	enum TYPE {
		TYPE_UNKNOWN,
		TYPE_TYPE,
		TYPE_NAMESPACE,
		TYPE_L_VALUE,
		TYPE_R_VALUE,
		TYPE_PROC_NAME,
		TYPE_MEMBER_PROC_NAME,			// tells that proc is member of some value (i.e. this calls)
		TYPE_NAME_MEMBER_PROC_NAME,		// tells that user has specified proc's namespace
	};

	TYPE					type;
	int						array;
	bool					isRef;

	union INFO {
		tavr::TYPE *		type;		// TYPE_TYPE, TYPE_*_VALUE
		tavr::METHOD *		procName;	// TYPE_PROC_NAME, TYPE_MEMBER_PROC_NAME
		tavr::NAMESPACE *	names;		// TYPE_NAMESPACE
	} info;

	TYPEINFO(void)
		: type(TYPE_UNKNOWN), info(), array(), isRef()
	{}

	inline bool				IsEqual					(const TYPEINFO & ti) const;
	inline bool				IsTypeEqual				(const TYPEINFO & ti) const;

	inline bool				IsLValue				(void) const { return type == TYPE_L_VALUE; }
	inline bool				IsRValue				(void) const { return type == TYPE_R_VALUE; }
	inline bool				IsValue					(void) const { return IsLValue() || IsRValue(); }

	inline bool				IsUserType				(void) const; // user types only (no arrays and so on)
	inline bool				IsTypeByValue			(void) const; // user types only
	inline bool				IsTypeByRef				(void) const; // user types and arrays
	inline bool				IsBoolean				(void) const;
	inline bool				IsNull					(void) const;
	inline bool				IsCharacter				(void) const;
	inline bool				IsIntegral				(void) const;
	inline bool				IsFloating				(void) const;
	inline bool				IsArithmetic			(void) const { return IsIntegral() || IsFloating(); }

	inline bool				IsArray					(void) const { return array > 0; }
	inline bool				IsStruct				(void) const;
	inline bool				IsAuto					(void) const;
	inline bool				IsEvent					(void) const;

	inline void				SetType					(tavr::TYPE * type);
	inline void				SetRValue				(tavr::TYPE * type);
	inline void				SetLValue				(tavr::TYPE * type);

	inline void				Clear					(void);

	inline bool				operator <				(const TYPEINFO & ti) const;
};

struct VARIABLE {
	ecg::STR_ID				name;
	TYPEINFO				type;
	int						line;
	int						cmdIndex;
};
typedef std::vector<VARIABLE>	VAR_SCOPE;
typedef std::stack<VAR_SCOPE>	VAR_SCOPE_STACK;

enum LOOKUP_TYPE {
	LOOKUP_TYPE_DEFAULT					= 0x0000,
	LOOKUP_TYPE_IGNORE_ARGS				= 0x0001,
	LOOKUP_TYPE_IGNORE_CONSTRUCTORS		= 0x0002,
	LOOKUP_TYPE_LOCALS_ONLY				= 0x0004,
	LOOKUP_TYPE_PRINT_AMBIGUITY_INFO	= 0x0008,
};
DECL_FLAGS_S(LOOKUP_TYPE);

struct LOOKUP_CONTEXT {
	MODULE *				module;
	TYPE *					curType;
	NAMESPACE *				curNamespace;
	METHOD *				curMethod;
	VAR_SCOPE_STACK			localVars;
	
	LOOKUP_CONTEXT(void)
		: module(), curType(), curNamespace(), curMethod()
	{}

	inline NAMESPACE *		GetParentNamespace		(void) const;
};

// EVALUATION DATA
struct EVAL_COMMAND {
	enum TYPE {
		TYPE_UNKNOWN,

		TYPE_CODE_BLOCK_BEGIN,
		TYPE_CODE_BLOCK_END,
		TYPE_LABEL,					// unanimous label that is used in goto statements
		TYPE_GOTO,					// 'goto'
		TYPE_RELAX,					// 'relax'

		TYPE_DECL,					// variable declaration
		TYPE_DECL_RELAX,			// tells that variable should be in relax stack

		TYPE_PROC_CALL,				// static proc call
		TYPE_THIS_CALL,				// method call for this-object
		TYPE_CONSTR_CALL,			// S2(x,y)
		TYPE_OUT_CALL,				// output action activization

		TYPE_VARIABLE,
		TYPE_CONSTANT,
		TYPE_THIS,					// this
		TYPE_SUBNAME,				// .member (only variables)

		TYPE_PRE_INC,				// ++i
		TYPE_PRE_DEC,				// ++i
		TYPE_POST_INC,				// i++
		TYPE_POST_DEC,				// i--

		TYPE_NEW_ARRAY,				// new int[count]
		TYPE_NEW_AUTO,				// new MyAuto()

		TYPE_ARRAY_LENGTH,			// arr.length
		TYPE_ARRAY_INDEX,			// val[index1, index2, index3...]
		TYPE_STRING_INDEX,			// str[index1, index2, index3...]
		TYPE_ASSIGN,				// '='

		TYPE_ADD,					// '+'
		TYPE_SUB,					// '-'
		TYPE_MUL,					// '*'
		TYPE_DIV,					// '/'
		TYPE_MOD,					// '%'

		TYPE_NEG,					// '-' (unar)
		TYPE_LOG_NOT,				// '!'
		TYPE_BIT_NOT,				// '~'

		TYPE_LOG_OR,				// '||'
		TYPE_LOG_AND,				// '&&'
		TYPE_EQ,					// '=='
		TYPE_NOT_EQ,				// '!='
		TYPE_LESS,					// '<'
		TYPE_LESS_EQ,				// '<='
		TYPE_GREATER,				// '>'
		TYPE_GREATER_EQ,			// '>='

		TYPE_TYPE_CAST,				// (int)l

		TYPE_IF,					// 'if'
		TYPE_ELSE,					// 'else'
		TYPE_ON,					// 'on'

		TYPE_EXPR_IF_COND,			// '?:' first part
		TYPE_EXPR_IF_ELSE,			// '?:' second part
		TYPE_EXPR_IF_RES,			// '?:' third part

		TYPE_SIZEOF,				// 'sizeof'
		TYPE_TYPEOF,				// 'typeof'
		TYPE_RETURN,				// 'return'

		TYPE_SEND,					// 'send' event 'to' auto
		TYPE_POST,					// 'post' event 'to' auto
		TYPE_EVENT,					// 'event'
		TYPE_REG_POST,				// += 'post'
		TYPE_REG_SEND,				// += 'send'
		TYPE_UNREG,					// -=
	};

	TYPE					type;
	std::vector<int>		args;
	struct INFO {
		METHOD *			proc;
		ecg::STR_ID			name;
	} info;
	TYPEINFO				typeinfo;

	std::vector<int>		availVars;
	bool					isUsed;
	int						line;

	EVAL_COMMAND(void)
		: type(TYPE_UNKNOWN), isUsed(), line(-1)
	{ memset(&info, 0, sizeof(INFO)); }
};
typedef std::vector<EVAL_COMMAND> EVAL_CODE;

struct DECL_INFO {
	MODULE *				module;
	int						line;

	DECL_INFO(void)
		: module(NULL), line(-1)
	{}
};

struct PLACEMENT_INFO {
	enum TYPE {
		TYPE_UNKNOWN,
		TYPE_DEFAULT,
		TYPE_PROC_ARG,			// index = argument index in the procedure
		TYPE_LOCAL_VAR,			// index = eval-command index in the code
		TYPE_TYPE_FIELD,		// index = field index in a type declaration
	};
	TYPE					type;
	DECL_INFO				declInfo;

	int						index;

	PLACEMENT_INFO(void)
		: type(TYPE_UNKNOWN), index(-1)
	{}
};

enum GOTO_TYPE {
	GOTO_TYPE_GOTO		= 0x0001,
	GOTO_TYPE_BREAK		= 0x0002,
	GOTO_TYPE_CONTINUE	= 0x0004,
	GOTO_TYPE_REPEAT	= 0x0008,
};
DECL_FLAGS_S(GOTO_TYPE);

struct GOTO_INFO {
	GOTO_TYPE				type;
	int						commandIndex;
	
	// GOTO_TYPE_GOTO
	ecg::STR_ID				name;				
	int						line;				// declaration line
};
typedef std::vector<GOTO_INFO> GOTO_INFO_ARRAY;

struct STEP_INFO {
	ecg::STR_ID				name;
	int						labelIndex;
	DECL_INFO				declInfo;
};
typedef std::vector<STEP_INFO> STEP_INFO_ARRAY;

struct EVALUATE {
	// input data
	LOOKUP_CONTEXT			context;
	bool					generateCode;
	EVAL_CODE				code;

	GOTO_TYPE_F				gotoAvailable;
	GOTO_INFO_ARRAY			gotoPending;

	std::stack<STEP_INFO_ARRAY>	steps;

	TYPE *					eventType;
	int						eventDeclIdx;

	// output data
	TYPEINFO				typeinfo;

	EVALUATE(void) : generateCode(), eventType(), eventDeclIdx(-1) {}
};

// STRUCTURE DATA
struct MODULE_MEMBER {
	enum TYPE {
		TYPE_UNKNOWN,
		TYPE_NAMESPACE,
		TYPE_METHOD,
		TYPE_TYPE,
	};
	TYPE					type;

	union INFO {
		tavr::TYPE *		type;
		NAMESPACE *			names;
		METHOD *			proc;
	} info;

	MODULE_MEMBER *			next;

	MODULE_MEMBER(void) 
		: info(), next()
	{}
};

struct MODULE {
	enum STATE {
		STATE_NOT_LOADED,				// needs tree parsing
		STATE_TREE_WAS_LOADED,			// tree was parsed, needs context parsing
		STATE_TYPES_WERE_PARSED,		// ready to parse methods declarations
		STATE_METHODS_WERE_PARSED,		// ready to code generation
		STATE_COMPLETED,				// module was successfully parsed
	};
	STATE					state;

	MODULE(void)
		: state(STATE_NOT_LOADED), tree(), members()
	{}

	inline METHOD *			NewMethod				(void);
	inline TYPE *			NewType					(void);

	std::string				name;
	std::string				source;

	ecg::CLASS_BUFFER		treeAlloc;
	NODE *					tree;

	ecg::CLASS_BUFFER		memAlloc;
	MODULE_MEMBER *			members;
};

enum ATTRIBS {
	ATTRIB_PUBLIC			= 0x0001,
	ATTRIB_PROTECTED		= 0x0002,
	ATTRIB_PRIVATE			= 0x0004,
	ATTRIB_STATIC			= 0x0008,
	ATTRIB_EXTERN			= 0x0010,

	ATTRIB_ACCESS_MASK		= ATTRIB_PRIVATE | ATTRIB_PUBLIC | ATTRIB_PROTECTED,
};
DECL_FLAGS_S(ATTRIBS);

struct METHOD { // class methods
	struct ARG {
		TYPEINFO			type;
		ecg::STR_ID			name;

		enum SUBTYPE {
			SUBTYPE_DEFAULT,
			SUBTYPE_INOUT,
			SUBTYPE_OUT,
		};

		SUBTYPE				subtype;

		ARG(void) : type(), subtype(SUBTYPE_DEFAULT) {}
	};
	typedef std::vector<ARG> ARG_ARRAY;

	inline bool 			operator == 			(const METHOD & p) const;
	inline bool 			operator != 			(const METHOD & p) const { return !(*this == p); }

	METHOD(void)
		: type(TYPE_UNKNOWN), result(), impl(), owner(OWNER())
	{}

	enum TYPE {
		TYPE_UNKNOWN,
		TYPE_PROCEDURE,		// global procedure
		TYPE_METHOD,		// auto/struct method (could be statis)
		TYPE_CONSTRUCTOR,	// auto/struct constructor
		TYPE_DESTRUCTOR,
		TYPE_OPERATOR,		
		TYPE_OUT,			// auto/struct output action
		TYPE_AUTO_PROC,		// auto/struct special auto proc
	};
	TYPE					type;

	ecg::STR_ID				name;
	DECL_INFO				declInfo;

	ATTRIBS_F				attribs;

	union OWNER {
		tavr::TYPE *		type;
		NAMESPACE *			names;
	} owner;

	TYPEINFO				result;
	ARG_ARRAY				args;
	const NODE *			impl;
	EVAL_CODE				code;

	// global proc (without this, thus static methods are 'IsProcedure' too)
	inline bool				IsProcedure				(void) const { return type == TYPE_PROCEDURE; }
	inline bool				IsMethod				(void) const { return type != TYPE_PROCEDURE; }

	inline void				Remove					(void);
};
typedef std::vector<METHOD *> METHOD_ARRAY;

struct RELAX_INFO {
	std::map<int, int>		context;
	int						stackSize;

	RELAX_INFO(void) : stackSize() {}
};

struct SUBAUTO_INFO {
	int						id;
	METHOD *				proc;
	RELAX_INFO				relaxInfo;
};

struct TYPE {
	enum SUBTYPE {
		SUBTYPE_UNKNOWN,
		SUBTYPE_VOID,
		SUBTYPE_NULL,
		SUBTYPE_BOOLEAN,
		SUBTYPE_CHARACTER,
		SUBTYPE_INTEGRAL,
		SUBTYPE_FLOATING,
		SUBTYPE_STRUCT,
		SUBTYPE_AUTO,
		SUBTYPE_EVENT,
	};
	SUBTYPE					subtype;

	struct FIELD {
		ecg::STR_ID			name;
		TYPEINFO			type;

		const NODE *		value;
		EVAL_CODE			code;

		int					line;

		FIELD(void)
			: value(), line(-1)
		{}
	};
	typedef std::vector<FIELD> FIELD_ARRAY;

	NAMESPACE *				parentNamespace;

	ecg::STR_ID				name;
	DECL_INFO				declInfo;

	ATTRIBS_F				attribs;

	FIELD_ARRAY				fields;
	METHOD_ARRAY			methods;

	int						size;

	RELAX_INFO				relaxInfo;
	std::vector<const NODE *> steps;				// list of automaton steps

	std::vector<SUBAUTO_INFO> subautos;

	int						eventId;

	typedef std::map<ecg::STR_ID, TYPE *> OUT_TABLE;
	OUT_TABLE				outTable;

	enum SPECIAL_METHOD {
		SPEC_DEF_CONSTR,
		SPEC_COPY_CONSTR,
		SPEC_EVENT_CONSTR,
		SPEC_DESTR,
		SPEC_DEF_OP_ASSIGN,			// 'operator='
		SPEC_AUTO_PROC,
		SPEC_MAX,
	};

	METHOD *				specs[SPEC_MAX];

	TYPE(void)
		: parentNamespace(), subtype(SUBTYPE_UNKNOWN)
		, size(), eventId(-1)
	{ memset(specs, 0, sizeof(specs)); }

	inline bool				IsTypeByValue			(void) const;
	inline bool				IsTypeByRef				(void) const;

	bool					DoesContainTypesByRef	(std::set<const TYPE *> & processingInfo) const;
};
typedef std::vector<TYPE *> TYPE_ARRAY;

struct NAMESPACE {
	ecg::STR_ID				name;

	TYPE_ARRAY				types;
	METHOD_ARRAY			procs;
	std::vector<NAMESPACE *> subNames;
};

struct PROJECT_SETTINGS {
	std::string				name;
	std::string				value;
	std::vector<PROJECT_SETTINGS *> subSettings;

	std::string				GetValue				(const char * path, const std::string & def = std::string()) const;
	const PROJECT_SETTINGS * GetSettings			(const char * path) const;
};

class PROJECT {
	friend TARGET_T;
	friend class ecg::TAVR_PROJECT_LOADER;
public:
	bool					Load					(std::istream & source);

	bool					AddModule				(const std::string & name, const std::string & source);
	bool					Compile					(void);

	void					SetError				(std::ostream & error) { this->error = &error; }
	std::ostream &			GetError				(void) const { return (error != NULL) ? *error : std::cout; }

	const PROJECT_SETTINGS * GetSettings			(const char * value) const;

	PROJECT(void);

private:
	int						AnalyzeTypes			(const LOOKUP_CONTEXT & context, const NODE * tree);
	int						AnalyzeMethods			(LOOKUP_CONTEXT & context, const NODE * tree);
	int						AnalyzeMethodsProc		(LOOKUP_CONTEXT & context, const NODE * procDecl, METHOD::TYPE procType, METHOD::OWNER owner, METHOD *& proc);
	int						AnalyzeMethodsOperator	(LOOKUP_CONTEXT & context, const NODE * procDecl, METHOD *& proc);
	int						AnalyzeMemberVar		(TYPE * type, const NODE * structDecl, const LOOKUP_CONTEXT & context, bool isAuto);
	int						AnalyzeOut				(LOOKUP_CONTEXT & context, const NODE * procDecl, TYPE * type);
	int						AnalizeSubauto			(LOOKUP_CONTEXT & lookupContext, const NODE * tree, TYPE * type);

	int						AnalyzeAttribs			(const NODE * tree, ATTRIBS_F & attribs);

	// generic lookup proc
	bool					NameLookup				(ecg::STR_ID name, const LOOKUP_CONTEXT & context, TYPEINFO & ti, PLACEMENT_INFO & placementInfo, LOOKUP_TYPE_F lookupType = LOOKUP_TYPE_DEFAULT);

	bool					VariableLookup			(ecg::STR_ID name, const LOOKUP_CONTEXT & context, TYPEINFO & ti, PLACEMENT_INFO & placementInfo);

	METHOD *				MethodLookupFind		(METHOD_ARRAY & methods, const METHOD * method, bool & ambiguity, LOOKUP_TYPE_F lookupType);
	METHOD *				MethodLookup			(const METHOD * method, const LOOKUP_CONTEXT & context, bool & ambiguity, LOOKUP_TYPE_F lookupType = LOOKUP_TYPE_DEFAULT);
	TYPE *					TypeLookup				(ecg::STR_ID name, const LOOKUP_CONTEXT & context, LOOKUP_TYPE_F lookupType);

	int						Evaluate				(const NODE * tree, EVALUATE & eval);
	
	int						EvalExprArray			(const NODE * tree, EVALUATE & eval);
	int						EvalDeclVar				(const NODE * tree, EVALUATE & eval);
	int						EvalExprCall			(const NODE * tree, EVALUATE & eval);
	int						EvalExprTypeCast		(const NODE * tree, EVALUATE & eval);
	int						EvalExprNewArray		(const NODE * tree, EVALUATE & eval);
	int						EvalExprNewAuto			(const NODE * tree, EVALUATE & eval);
	int						EvalExprIf				(const NODE * tree, EVALUATE & eval);
	int						EvalExprOrAnd			(const NODE * tree, EVALUATE & eval);
	int						EvalExprBinOp			(const NODE * tree, EVALUATE & eval);
	int						EvalExprUnarOp			(const NODE * tree, EVALUATE & eval);

	int						EvalFor					(const NODE * tree, EVALUATE & eval);
	int						EvalWhile				(const NODE * tree, EVALUATE & eval);
	int						EvalDoWhile				(const NODE * tree, EVALUATE & eval);
	int						EvalOn					(const NODE * tree, EVALUATE & eval);
	int						EvalStep				(const NODE * tree, EVALUATE & eval);
	int						EvalCodeBlock			(const NODE * tree, EVALUATE & eval);
	int						EvalSubName				(const NODE * tree, EVALUATE & eval);
	int						EvalGotoRelax			(const NODE * tree, EVALUATE & eval);
	int						EvalSendPost			(const NODE * tree, EVALUATE & eval);
	int						EvalReg					(const NODE * tree, EVALUATE & eval);

	inline int				EvalProcessMethodArgs	(const NODE * firstArg, EVAL_COMMAND & cmd, METHOD & method, EVALUATE & eval);

	inline std::string		Type2Str				(METHOD * method);
	std::string				Type2Str				(const TYPEINFO & ti);
	int						VerifyBinOpTypeCompatibility(NODE::TYPE type, int line, const char * opVal, const TYPEINFO & l, const TYPEINFO & r, METHOD *& overriddenOp);

	int						VerifyEscapeSequences	(const std::string & value, int line);

	struct GENERATE_INFO {
		std::vector<int>	sizeofCommands;
	};

	struct AUTO_PROC_PAIR {
		TYPE *				type;
		METHOD *			method;

		bool operator < (const AUTO_PROC_PAIR & p) const {
			return type < p.type || type == p.type && method < p.method;
		}
	};
	typedef std::map<AUTO_PROC_PAIR, GENERATE_INFO> GENERATE_INFO_MAP;

	int						Generate				(GENERATE_INFO_MAP & info, MODULE * module);
	int						PostGenerate			(GENERATE_INFO_MAP & info, MODULE * module);

	int						PostAnalyzeCode			(GENERATE_INFO_MAP & info, EVALUATE & eval);

	void					PrintMsg				(int msgId, int line);
	void					PrintMsgInfo			(int msgId, const DECL_INFO & info);

	int						CalcRelaxStackSize		(const GENERATE_INFO_MAP & infoMap, const GENERATE_INFO & info, RELAX_INFO & relaxInfo, TYPE * type, METHOD * method);
	int						CalcSizeOf				(const GENERATE_INFO_MAP & info, TYPE * type);

#ifdef _DEBUG
	void					VerifyTree				(const NODE * tree);
#endif

private:
	std::ostream *			error;
	ecg::CLASS_BUFFER		alloc; // TODO: do we need this?

	typedef std::map<std::string, MODULE *> MODULE_MAP;
	MODULE_MAP				modules;

	TARGET_T				target;

	// LANGUAGE DATA
	NAMESPACE *				globalName;

	ecg::MESSAGE::STRING_MAP msg;
	PROJECT_SETTINGS *		settings;

	int						eventIdCntr;
	int						subautoUIdCntr;

public:
	// public constants
	ecg::STR_ID				emptyString;
	ecg::STR_ID				autoString;
	ecg::STR_ID				castOpString;
	ecg::STR_ID				assignOpString;
	ecg::STR_ID				nullString;
	ecg::STR_ID				eventString;
	ecg::STR_ID				trueString;
	ecg::STR_ID				falseString;

	enum DEF_TYPE {
		DEF_TYPE_0,

		DEF_TYPE_VOID = DEF_TYPE_0,
		DEF_TYPE_NULL,
		DEF_TYPE_BOOL,
		DEF_TYPE_CHAR,
		
		DEF_TYPE_S_BYTE,
		DEF_TYPE_BYTE,
		DEF_TYPE_SHORT,
		DEF_TYPE_U_SHORT,
		DEF_TYPE_INT,
		DEF_TYPE_U_INT,
		DEF_TYPE_LONG,
		DEF_TYPE_U_LONG,
		
		DEF_TYPE_FLOAT,
		DEF_TYPE_DOUBLE,

		DEF_TYPE_STRING,

		DEF_TYPE_MAX,

		DEF_TYPE_CONVERTABLE_0		= DEF_TYPE_CHAR,
		DEF_TYPE_CONVERTABLE_MAX	= DEF_TYPE_DOUBLE + 1,

		DEF_TYPE_ARITHMETIC_0		= DEF_TYPE_S_BYTE,
		DEF_TYPE_ARITHMETIC_MAX		= DEF_TYPE_DOUBLE + 1,
	};

	struct DEF_TYPE_INFO {
		const char *		name;
		int					size;
	};

	static DEF_TYPE_INFO	defTypeInfo[];

	struct DEF_TYPE_SUBTYPE_INFO {
		DEF_TYPE			defType;
		TYPE::SUBTYPE		subtype;
	};
	static DEF_TYPE_SUBTYPE_INFO defTypeSubtype[];

	TYPE *					defType[DEF_TYPE_MAX];
};

#include "tavr.hpp"

#define PRINT_MSG(ID, LINE) \
	PrintMsg(MESSAGE_##ID, LINE)

#define PRINT_MSG_INFO(ID, INFO) \
	PrintMsgInfo(MESSAGE_##ID, INFO)

};

#endif // TAVR_H