
#include "stdafx.h"
#include "tavr.h"
#include "tavr_loader_msg.h"

using namespace std;
using namespace ecg;

namespace tavr {

/************************************************************************/
/* tavr::PROJECT::NameLookup
/*
/************************************************************************/
bool PROJECT::NameLookup(ecg::STR_ID name, const LOOKUP_CONTEXT & context, TYPEINFO & ti, PLACEMENT_INFO & placementInfo, LOOKUP_TYPE_F lookupType)
{
	assert(context.module != NULL);

	if (VariableLookup(name, context, ti, placementInfo)) {
		assert(placementInfo.type != PLACEMENT_INFO::TYPE_DEFAULT
			&& placementInfo.type != PLACEMENT_INFO::TYPE_UNKNOWN);
		return true;
	}

	placementInfo.type = PLACEMENT_INFO::TYPE_DEFAULT;

	METHOD methodName;
	methodName.name = name;
	bool ambiguity;
	METHOD * method = MethodLookup(&methodName, context, ambiguity, lookupType.value() | LOOKUP_TYPE_IGNORE_ARGS | LOOKUP_TYPE_IGNORE_CONSTRUCTORS);

	if (method != NULL) {
		ti.type = TYPEINFO::TYPE_PROC_NAME;
		ti.info.procName = method;

		placementInfo.declInfo = method->declInfo;
		return true;
	}

	ti.info.type = TypeLookup(name, context, lookupType);
	if (ti.info.type != NULL) {
		ti.type = TYPEINFO::TYPE_TYPE;

		placementInfo.declInfo = ti.info.type->declInfo;
		return true;
	}

	// TODO: namespaces

	return false;
}

/************************************************************************/
/* tavr::PROJECT::VariableLookup
/*
/************************************************************************/
bool PROJECT::VariableLookup(ecg::STR_ID name, const LOOKUP_CONTEXT & context, TYPEINFO & ti, PLACEMENT_INFO & placementInfo)
{
	assert(context.module != NULL);

	// procedure args
	if (context.curMethod != NULL) {
		METHOD::ARG_ARRAY & args = context.curMethod->args;
		for (int i=0; i<(int)args.size(); ++i) {
			if (args[i].name == name) {
				ti			= args[i].type;
				assert(ti.type == TYPEINFO::TYPE_TYPE);

				ti.type		= TYPEINFO::TYPE_L_VALUE;

				placementInfo.type		= PLACEMENT_INFO::TYPE_PROC_ARG;
				placementInfo.index		= i;
				placementInfo.declInfo	= context.curMethod->declInfo;
				return true;
			}
		}
	}

	// lookup in local variables
	for (int i=(int)context.localVars.size() - 1; i >= 0; --i) {
		const VAR_SCOPE & scope = context.localVars.c[i];
		for (int k=0; k<(int)scope.size(); ++k) {
			if (scope[k].name == name) {
				ti = scope[k].type;

				ti.type = TYPEINFO::TYPE_L_VALUE;

				placementInfo.type				= PLACEMENT_INFO::TYPE_LOCAL_VAR;
				placementInfo.index				= scope[k].cmdIndex;
				placementInfo.declInfo.module	= context.module;
				placementInfo.declInfo.line		= scope[k].line;
				return true;
			}
		}
	}

	if (context.curType != NULL) {
		for (int i=(int)context.curType->fields.size() - 1; i >= 0; --i) {
			TYPE::FIELD & field = context.curType->fields[i];
			if (field.name == name) {
				ti = field.type;

				ti.type = TYPEINFO::TYPE_L_VALUE;
				placementInfo.type				= PLACEMENT_INFO::TYPE_TYPE_FIELD;
				placementInfo.index				= i;
				placementInfo.declInfo.module	= context.curType->declInfo.module;
				placementInfo.declInfo.line		= field.line;
				return true;
			}
		}
	}

	// TODO: global variables

	return false;
}

/************************************************************************/
/* tavr::PROJECT::MethodLookupFind
/*
/************************************************************************/
METHOD * PROJECT::MethodLookupFind(METHOD_ARRAY & methods, const METHOD * method, bool & ambiguity, LOOKUP_TYPE_F lookupType)
{
	METHOD * res = NULL;
	ambiguity = false;

	for (int i = 0, size = (int)methods.size(); i < size; ++i) {
		METHOD * cur = methods[i];

		if (cur->name != method->name) {
			continue;
		}

		if (lookupType.is(LOOKUP_TYPE_IGNORE_CONSTRUCTORS) && cur->type == METHOD::TYPE_CONSTRUCTOR) {
			continue;
		}

		if (lookupType.is(LOOKUP_TYPE_IGNORE_ARGS)) {
			return cur;
		}

		if (cur->type == METHOD::TYPE_OPERATOR
		&& method->name == castOpString && cur->name == castOpString) {
			if (!cur->result.IsTypeEqual(method->result)) {
				continue;
			}
		}

		if (cur->args.size() != method->args.size()) {
			continue;
		}
		bool equals = true;
		for (int argIndex = (int)cur->args.size() - 1; argIndex >= 0; --argIndex) {
			const METHOD::ARG & argA = cur->args[argIndex];
			const METHOD::ARG & argB = method->args[argIndex];
			if (!argA.type.IsTypeEqual(argB.type)) {
				equals = false;
				break;
			}
		}

		if (!equals) {
			continue;
		}

		if (lookupType.is(LOOKUP_TYPE_PRINT_AMBIGUITY_INFO)) {
			msg["name"] = Type2Str(cur);
			PRINT_MSG_INFO(TVR_AMBIGUITY_INFO, cur->declInfo);
			continue;
		}

		if (res != NULL) {
			ambiguity = true;
			return NULL;
		}

		res = cur;
	}

	return res;
}

/************************************************************************/
/* tavr::PROJECT::MethodLookup
/*
/************************************************************************/
METHOD * PROJECT::MethodLookup(const METHOD * method, const LOOKUP_CONTEXT & context, bool & ambiguity, LOOKUP_TYPE_F lookupType)
{
	NAMESPACE * curNamespace;

	ambiguity = false;

	// check whether there is parent namespace of some type or procedure
	if (!lookupType.is(LOOKUP_TYPE_LOCALS_ONLY)) {
		curNamespace = context.GetParentNamespace();
	} else {
		curNamespace = context.curNamespace;
	}

	if (context.curType != NULL) {
		METHOD * find = MethodLookupFind(context.curType->methods, method, ambiguity, lookupType);
		if (find != NULL || ambiguity) {
			return find;
		}
	}

	if (curNamespace != NULL) {
		METHOD * find = MethodLookupFind(curNamespace->procs, method, ambiguity, lookupType);
		if (find != NULL || ambiguity) {
			return find;
		}
	}

	return NULL;
}

struct TYPE_LOOKUP{
	TYPE_LOOKUP(STR_ID name)
		: name(name)
	{}

	STR_ID name;

	friend bool operator == (const TYPE * a, const TYPE_LOOKUP & p)
	{
		// TODO: support array specification
		return a->name == p.name;
	}
};


/************************************************************************/
/* tavr::PROJECT::TypeLookup
/*
/************************************************************************/
TYPE * PROJECT::TypeLookup(STR_ID name, const LOOKUP_CONTEXT & context, LOOKUP_TYPE_F lookupType)
{
	// TODO: lookup should fail if there is a variable with same name (for this context)
	// TODO: support namespaces

	NAMESPACE * curNamespace;
	if (lookupType.is(LOOKUP_TYPE_LOCALS_ONLY)) {
		curNamespace = context.curNamespace;
	} else {
		curNamespace = context.GetParentNamespace();
	}

	if (curNamespace != NULL) {
		TYPE_ARRAY::iterator iterEnd = curNamespace->types.end();
		TYPE_ARRAY::const_iterator iter = find(curNamespace->types.begin(), iterEnd, TYPE_LOOKUP(name));
		if (iter != iterEnd) {
			return *iter;
		}
	}

	// TODO: support using declarations in the module

	return NULL;
}

/************************************************************************/
/* tavr::PROJECT::EvalProcessMethodArgs
/*
/************************************************************************/
int PROJECT::EvalProcessMethodArgs(const NODE * firstArg, EVAL_COMMAND & cmd, METHOD & method, EVALUATE & eval)
{
	int res = 0;

	cmd.args.reserve(4);
	METHOD::ARG arg;
	for (const NODE * curArg = firstArg; curArg != NULL; curArg = curArg->next) {
		int evalResult = Evaluate(curArg, eval);
		if (evalResult == 0 && !eval.typeinfo.IsValue()) {
			msg["name"] = Type2Str(eval.typeinfo);
			PRINT_MSG(TVR1022, curArg->line);
			++res;
			continue;
		}

		if (evalResult == 0) {
			arg.type		= eval.typeinfo;
			arg.type.type	= TYPEINFO::TYPE_TYPE;
			method.args.push_back(arg);

			cmd.args.push_back((int)eval.code.size() - 1);
		} else {
			res += evalResult;
		}
	}

	std::reverse(cmd.args.begin(),		cmd.args.end());
	std::reverse(method.args.begin(),	method.args.end());

	return res;
}

/************************************************************************/
/* tavr::PROJECT::EvalExprArray
/*
/************************************************************************/
int PROJECT::EvalExprArray(const NODE * tree, EVALUATE & eval)
{
	int res = Evaluate(tree->subnode, eval);

	if (res == 0) {
		// check whether it's an array type
		if (eval.typeinfo.type == TYPEINFO::TYPE_TYPE) {
			assert(!eval.generateCode);

			if (tree->subnode->next != NULL) {
				PRINT_MSG(TVR1038, tree->subnode->line);
				return res + 1;
			}

			++eval.typeinfo.array;
			return res;
		}

		if (!eval.typeinfo.IsValue()) {
			msg["name"] = Type2Str(eval.typeinfo);
			PRINT_MSG(TVR1022, tree->subnode->line);
			++res;
		} 
	}

	EVAL_COMMAND cmd;
	cmd.args.reserve(5);
	cmd.typeinfo	= eval.typeinfo;

	if (res == 0) {
		cmd.args.push_back((int)eval.code.size() - 1);
	}

	std::vector<TYPEINFO> argTypes;
	argTypes.reserve(4);

	// TODO: support array indexes
	for (const NODE * arg = tree->subnode->next; arg != NULL; arg = arg->next) {
		res += Evaluate(arg, eval);
		argTypes.push_back(eval.typeinfo);
		cmd.args.push_back((int)eval.code.size() - 1);
	}

	if (res != 0) {
		return res;
	}

	if (cmd.args.empty()) {
		PRINT_MSG(TVR1018, tree->subnode->line);
		return res + 1;
	}

	// only one index possible for array types
	if (cmd.typeinfo.array > 0 || cmd.typeinfo.info.type == defType[DEF_TYPE_STRING]) {
		// first is array, second is index
		if (cmd.args.size() != 2) {
			assert(cmd.args.size() > 2);
			PRINT_MSG(TVR1019, tree->subnode->next->line);
			return res + 1;
		}

		if (!argTypes[0].IsIntegral()) {
			PRINT_MSG(TVR1017, tree->subnode->next->line);
			return res + 1;
		}

		if (cmd.typeinfo.array > 0) {
			--cmd.typeinfo.array;
			eval.typeinfo	= cmd.typeinfo;
			cmd.type		= EVAL_COMMAND::TYPE_ARRAY_INDEX;
		} else {
			cmd.typeinfo.SetRValue(defType[DEF_TYPE_CHAR]);
			eval.typeinfo	= cmd.typeinfo;
			cmd.type		= EVAL_COMMAND::TYPE_STRING_INDEX;
		}

		if (eval.generateCode) {
			cmd.line	= tree->subnode->line;
			eval.code.push_back(cmd);
		}
		return res;
	}

	// TODO: support overridden indexers
	assert(!"Support operator []");

	return res;
}

/************************************************************************/
/* tavr::PROJECT::EvalDeclVar
/*
/************************************************************************/
int PROJECT::EvalDeclVar(const NODE * tree, EVALUATE & eval)
{
	bool dontGenerateCode = false;
	swap(eval.generateCode, dontGenerateCode);
	TYPE * curType = NULL;
	swap(curType, eval.context.curType);
	int res = Evaluate(tree->subnode, eval);
	swap(curType, eval.context.curType);
	swap(eval.generateCode, dontGenerateCode);

	if (res == 0 && (eval.typeinfo.type != TYPEINFO::TYPE_TYPE || eval.typeinfo.IsNull())) {
		if (eval.typeinfo.IsValue()) {
			PRINT_MSG(TVR1007, tree->subnode->line);
		} else {
			msg["name"] = Type2Str(eval.typeinfo);
			PRINT_MSG(TVR1015, tree->subnode->line);
		}
		++res;
	}

	EVAL_COMMAND cmd;
	cmd.type			= EVAL_COMMAND::TYPE_DECL;
	cmd.typeinfo		= eval.typeinfo;
	cmd.args.reserve(4);

	VARIABLE var;
	var.type			= eval.typeinfo;
	var.line			= tree->subnode->line;
	var.cmdIndex		= -1;

	std::vector<const NODE *> vars;
	vars.reserve(8);
	for (const NODE * cur = tree->subnode->next; cur != NULL; cur = cur->next) {
		vars.push_back(cur);
	}

	// process elements
	assert(tree->subnode->next != NULL);

	LOOKUP_CONTEXT ctxt;
	ctxt.module = eval.context.module;

	PLACEMENT_INFO placementInfo;
	for (int i=(int)vars.size() - 1; i >= 0; --i) {
		const NODE * curVar = vars[i];

		const bool isConstr = (curVar->type == NODE::TYPE_DECL_VAR_CONSTR);
		assert(isConstr || curVar->type == NODE::TYPE_ARGUMENT);

		cmd.info.name	= curVar->name;
		cmd.args.clear();

		var.name = curVar->name;

		swap(ctxt.localVars, eval.context.localVars);
		ctxt.curMethod = eval.context.curMethod;
		if (VariableLookup(var.name, ctxt, var.type, placementInfo)) {
			msg["name"] = var.name.str();
			PRINT_MSG(TVR1016, curVar->line);
			PRINT_MSG_INFO(TVR_SEEDECL_INFO, placementInfo.declInfo);
			++res;
		}
		swap(ctxt.localVars, eval.context.localVars);
		assert(ctxt.localVars.empty());

		// check whether a variable has initialization expression
		if (!isConstr && curVar->subnode != NULL) {
			int evalResult = Evaluate(curVar->subnode, eval);
			
			// check type compatibility

			if (res == 0 && evalResult == 0) {
				TYPEINFO ti = cmd.typeinfo;

				METHOD * overridden = NULL;
				res += VerifyBinOpTypeCompatibility(NODE::TYPE_DECL_VAR, curVar->line, "=", ti, eval.typeinfo, overridden);
				
				// TODO: support overridden copy constructor
				assert(overridden == NULL);

				cmd.args.push_back((int)eval.code.size() - 1);
			} else {
				res += evalResult;
			}
		} else if (isConstr) {
			if (!var.type.IsTypeByValue()) {
				msg["name"] = Type2Str(var.type);
				PRINT_MSG(TVR1031, curVar->line);
				++res;
				continue;
			}

			// process constructor arguments
			METHOD method;
			int errorResult = EvalProcessMethodArgs(curVar->subnode, cmd, method, eval);
			if (errorResult != 0) {
				res += errorResult;
				continue;
			}

			ctxt.curType = var.type.info.type;

			method.type = METHOD::TYPE_CONSTRUCTOR;
			method.name = ctxt.curType->name;

			bool ambiguity;
			METHOD * found = MethodLookup(&method, ctxt, ambiguity, LOOKUP_TYPE_LOCALS_ONLY);

			if (found == NULL) {
				method.owner.type = var.type.info.type;
				msg["name"] = Type2Str(&method);

				if (!ambiguity) {
					PRINT_MSG(TVR1030, curVar->line);
				} else {
					PRINT_MSG(TVR1045, curVar->line);
					MethodLookup(&method, ctxt, ambiguity, LOOKUP_TYPE_LOCALS_ONLY | LOOKUP_TYPE_PRINT_AMBIGUITY_INFO);
				}
				++res;
				continue;
			}

			cmd.info.proc = found;
		}

		if (res == 0) {
			eval.context.localVars.top().push_back(var);

			if (eval.generateCode) {
				// store variable declaration eval-command index
				eval.context.localVars.top().back().cmdIndex = (int)eval.code.size();

				cmd.line = tree->subnode->line;
				eval.code.push_back(cmd);
			}
		}
	}

	eval.typeinfo.SetRValue(defType[DEF_TYPE_VOID]);

	return res;
}

/************************************************************************/
/* tavr::PROJECT::EvalSubName
/*
/************************************************************************/
int PROJECT::EvalSubName(const NODE * tree, EVALUATE & eval)
{
	int res = Evaluate(tree->subnode, eval);

	if (res != 0) {
		return res;
	}

	// TODO: support subtypes

	LOOKUP_CONTEXT context;
	context.module = eval.context.module;

	bool isValueMember = false;
	switch (eval.typeinfo.type) {
		case TYPEINFO::TYPE_NAMESPACE:
			context.curNamespace = eval.typeinfo.info.names;
			break;

		case TYPEINFO::TYPE_TYPE:
			context.curType		 = eval.typeinfo.info.type;
			break;

		default:
			// this is a hack: simulate field 'length' for arrays
			if (eval.typeinfo.array > 0 && eval.typeinfo.IsValue() && tree->name.str() == "length") {
				eval.typeinfo.SetRValue(defType[DEF_TYPE_INT]);

				assert(res == 0);
				if (eval.generateCode) {
					EVAL_COMMAND cmd;
					cmd.type	= EVAL_COMMAND::TYPE_ARRAY_LENGTH;
					cmd.args.push_back((int)eval.code.size() - 1);
					cmd.line	= tree->subnode->line;
					cmd.typeinfo = eval.typeinfo;
					eval.code.push_back(cmd);
				}

				return 0;
			}

			// otherwise it should be value user type
			if (!eval.typeinfo.IsUserType() || !eval.typeinfo.IsValue()) {
				msg["name"] = tree->name.str();
				msg["name2"] = Type2Str(eval.typeinfo);
				PRINT_MSG(TVR1026, tree->line);
				return 1;
			}

			isValueMember = true;
			context.curType = eval.typeinfo.info.type;
			break;
	}

	TYPEINFO ti;
	PLACEMENT_INFO placementInfo;
	if (!NameLookup(tree->name, context, ti, placementInfo, LOOKUP_TYPE_LOCALS_ONLY)) {
		msg["name"] = tree->name.str();
		msg["name2"] = Type2Str(eval.typeinfo);
		PRINT_MSG(TVR1027, tree->line);
		if (eval.typeinfo.type == TYPEINFO::TYPE_TYPE) {
			msg["name"] = Type2Str(eval.typeinfo);
			PRINT_MSG_INFO(TVR_SEEDECL_INFO, eval.typeinfo.info.type->declInfo);
		}
		return 1;
	}

	if (ti.IsValue()) {
		if (eval.generateCode) {
			EVAL_COMMAND cmd;
			cmd.type		= EVAL_COMMAND::TYPE_SUBNAME;
			cmd.info.name	= tree->name;
			cmd.args.push_back((int)eval.code.size() - 1);
			cmd.typeinfo	= ti;
			cmd.line		= tree->line;
			eval.code.push_back(cmd);
		}
	} else if (ti.type == TYPEINFO::TYPE_PROC_NAME) {
		if (isValueMember) {
			ti.type = TYPEINFO::TYPE_MEMBER_PROC_NAME;
		} else {
			ti.type = TYPEINFO::TYPE_NAME_MEMBER_PROC_NAME;
		}
	} else {
		assert(ti.type == TYPEINFO::TYPE_TYPE);
	}

	eval.typeinfo = ti;

	return 0;
}

/************************************************************************/
/* tavr::PROJECT::EvalExprCall
/*
/************************************************************************/
int PROJECT::EvalExprCall(const NODE * tree, EVALUATE & eval)
{
	// retrieve function name
	int res = Evaluate(tree->subnode, eval);

	METHOD * proc = NULL;
	METHOD method;

	bool isConstr = false;

	if (res == 0) {
		switch (eval.typeinfo.type) {
			case TYPEINFO::TYPE_PROC_NAME:
			case TYPEINFO::TYPE_MEMBER_PROC_NAME:
			case TYPEINFO::TYPE_NAME_MEMBER_PROC_NAME:
				proc = eval.typeinfo.info.procName;
				method.name = proc->name;
				break;

			case TYPEINFO::TYPE_TYPE:
				if (!eval.typeinfo.IsTypeByValue()) {
					msg["name"] = Type2Str(eval.typeinfo);
					PRINT_MSG(TVR1031, tree->subnode->line);
					++res;
					break;
				}

				method.type = METHOD::TYPE_CONSTRUCTOR;
				method.owner.type = eval.typeinfo.info.type;
				method.name = method.owner.type->name;
				proc = &method;

				isConstr = true;
				break;

			default:
				msg["name"] = Type2Str(eval.typeinfo);
				PRINT_MSG(TVR1013, tree->subnode->line);
				++res;
				break;
		}
	}

	const bool isMemberProc		= (eval.typeinfo.type == TYPEINFO::TYPE_MEMBER_PROC_NAME);
	const bool isNameMemberProc = (eval.typeinfo.type == TYPEINFO::TYPE_NAME_MEMBER_PROC_NAME);

	// store argument for member proc
	const int thisArgIndex = (int)eval.code.size() - 1;

	// evaluate function args
	EVAL_COMMAND cmd;
	cmd.line = tree->line;

	res += EvalProcessMethodArgs(tree->subnode->next, cmd, method, eval);

	if (res != 0) {
		return res;
	}

	METHOD * found = NULL;

	bool ambiguity;
	LOOKUP_CONTEXT context;
	if (isMemberProc || isNameMemberProc || isConstr) {
		if (proc->type == METHOD::TYPE_PROCEDURE) {
			context.curNamespace = proc->owner.names;
		} else {
			assert(proc->IsMethod());
			context.curType	= proc->owner.type;
		}

		found = MethodLookup(&method, context, ambiguity, LOOKUP_TYPE_LOCALS_ONLY);

		if (isNameMemberProc && found != NULL && found->IsMethod()) {
			TYPE * curTypeContext = NULL;
			if (eval.context.curMethod != NULL && eval.context.curMethod->IsMethod()) {
				curTypeContext = eval.context.curMethod->owner.type;
			}

			if (!found->attribs.is(ATTRIB_STATIC) && curTypeContext != found->owner.type) {
				msg["name"] = Type2Str(found);
				PRINT_MSG(TVR1050, tree->subnode->line);
				++res;
			}
		}
	} else {
		found = MethodLookup(&method, eval.context, ambiguity);
	}

	if (res == 0 && found == NULL) {
		method.result.type = TYPEINFO::TYPE_TYPE;
		method.result.info.type = defType[DEF_TYPE_VOID];
		method.type = proc->type;
		method.owner = proc->owner;

		msg["name"] = Type2Str(&method);

		if (!ambiguity) {
			PRINT_MSG(TVR1030, tree->line);
		} else {
			PRINT_MSG(TVR1045, tree->line);
			if (isMemberProc || isConstr) {
				MethodLookup(&method, context, ambiguity, LOOKUP_TYPE_LOCALS_ONLY | LOOKUP_TYPE_PRINT_AMBIGUITY_INFO);
			} else {
				MethodLookup(&method, eval.context, ambiguity, LOOKUP_TYPE_LOCALS_ONLY | LOOKUP_TYPE_PRINT_AMBIGUITY_INFO);
			}
		}
		return 1;
	}

	if (!isConstr) {
		eval.typeinfo = found->result;
		eval.typeinfo.type = TYPEINFO::TYPE_R_VALUE;
		// TODO: could be l-value
	} else {
		eval.typeinfo.SetRValue(method.owner.type);
	}

	assert(isConstr == (found->type == METHOD::TYPE_CONSTRUCTOR));

	if (eval.generateCode) {
		// support this call
		switch (found->type) {
			case METHOD::TYPE_PROCEDURE:
				cmd.type = EVAL_COMMAND::TYPE_PROC_CALL;
				break;

			case METHOD::TYPE_METHOD:
			case METHOD::TYPE_AUTO_PROC:
				if (found->attribs.is(ATTRIB_STATIC)) {
					cmd.type = EVAL_COMMAND::TYPE_PROC_CALL;
					break;
				}

				if (!isMemberProc) {
					EVAL_COMMAND thisCmd;
					thisCmd.type = EVAL_COMMAND::TYPE_THIS;
					thisCmd.line = tree->subnode->line;
					eval.code.push_back(thisCmd);

					cmd.args.insert(cmd.args.begin(), (int)eval.code.size() - 1);
				} else {
					cmd.args.insert(cmd.args.begin(), thisArgIndex);
				}

				cmd.type = EVAL_COMMAND::TYPE_THIS_CALL;
				break;

			case METHOD::TYPE_CONSTRUCTOR:
				cmd.type = EVAL_COMMAND::TYPE_CONSTR_CALL;
				break;

			case METHOD::TYPE_OUT:
				{
					TYPE * curType = found->owner.type;
					assert(curType->outTable.find(found->name) != curType->outTable.end());
					TYPE * eventType = curType->outTable[found->name];

					// retrieve event constructor
					METHOD mt = *found;
					mt.type = METHOD::TYPE_CONSTRUCTOR;
					mt.name = eventType->name;

					swap(eventType, eval.context.curType);
					METHOD * constr = MethodLookup(&mt, eval.context, ambiguity, LOOKUP_TYPE_LOCALS_ONLY);
					assert(constr != NULL);
					swap(eventType, eval.context.curType);

					cmd.type = EVAL_COMMAND::TYPE_NEW_AUTO;
					assert(cmd.args.size() == constr->args.size());
					cmd.typeinfo.SetType(eventType);
					cmd.info.proc = constr;
					eval.code.push_back(cmd);

					cmd.type = EVAL_COMMAND::TYPE_OUT_CALL;
					cmd.args.clear();
					cmd.args.push_back((int)eval.code.size() - 1);
					cmd.info.name = found->name;
					eval.code.push_back(cmd);
				}
				break;

			default:
				assert(!"Unknown method type");
				break;
		}

		if (found->type != METHOD::TYPE_OUT) {
			cmd.info.proc	= found;
			cmd.typeinfo	= eval.typeinfo;

			// make sure that proc pointer is valid
			assert(cmd.info.proc->owner.names != NULL);
			assert(cmd.info.proc->name.c_str());

			eval.code.push_back(cmd);
		}
	}

	return res;
}

/************************************************************************/
/* tavr::PROJECT::EvalFor
/*
/************************************************************************/
int PROJECT::EvalFor(const NODE * tree, EVALUATE & eval)
{
	int res = 0;

	EVAL_COMMAND cmd;
	cmd.line = tree->line;

	if (eval.generateCode) {
		cmd.type = EVAL_COMMAND::TYPE_CODE_BLOCK_BEGIN;
		eval.code.push_back(cmd);
	}

	const int localVarsStackSize = (int)eval.context.localVars.size();
	eval.context.localVars.push(VAR_SCOPE());

	// generate for INITIALIZATION code
	const NODE * forInitArg = tree->subnode;
	if (forInitArg->subnode != NULL) {
		res += Evaluate(forInitArg->subnode, eval);
		if (res != 0) {
			eval.context.localVars.pop();
			return res;
		}
	}

	// generate REPEAT LABEL
	assert(res == 0);
	if (eval.generateCode) {
		cmd.type = EVAL_COMMAND::TYPE_LABEL;
		eval.code.push_back(cmd);
	}

	EVAL_COMMAND gotoCmd;
	gotoCmd.type = EVAL_COMMAND::TYPE_GOTO;
	gotoCmd.args.push_back((int)eval.code.size() - 1);

	// generate for CONDITION code
	const NODE * forCondArg = forInitArg->next;
	if (forCondArg->subnode != NULL) {
		res += Evaluate(forCondArg->subnode, eval);

		if (res == 0 && (!eval.typeinfo.IsValue() || !eval.typeinfo.IsBoolean())) {
			msg["name"] = "for";
			msg["name2"] = Type2Str(eval.typeinfo);
			PRINT_MSG(TVR1032, forCondArg->subnode->line);
			++res;
		}

		if (res == 0 && eval.generateCode) {
			EVAL_COMMAND ifCmd;
			ifCmd.type = EVAL_COMMAND::TYPE_IF;
			ifCmd.args.push_back((int)eval.code.size() - 1);
			ifCmd.line = forCondArg->subnode->line;
			eval.code.push_back(ifCmd);
		}
	}

	// generate for BODY
	const NODE * forIncArg = forCondArg->next;
	const NODE * forBodyArg = forIncArg->next;

	// make break and continue available
	GOTO_TYPE_F oldGotoAvailable = eval.gotoAvailable;
	eval.gotoAvailable.set(GOTO_TYPE_BREAK | GOTO_TYPE_CONTINUE);
	eval.gotoAvailable.clear(GOTO_TYPE_REPEAT);

	GOTO_INFO_ARRAY gotoPending;
	gotoPending.reserve(4);
	swap(eval.gotoPending, gotoPending);

	res += Evaluate(forBodyArg->subnode, eval);

	// restore goto availability
	eval.gotoAvailable = oldGotoAvailable;
	swap(eval.gotoPending, gotoPending);

	// EXIT GOTO & CONTINUE LABEL
	int continueLabelIndex = -1;
	if (res == 0 && eval.generateCode) {
		if (forCondArg->subnode != NULL) {
			cmd.type	= EVAL_COMMAND::TYPE_ELSE;
			cmd.line	= tree->line;
			eval.code.push_back(cmd);

			cmd.type		= EVAL_COMMAND::TYPE_GOTO;
			eval.code.push_back(cmd);

			GOTO_INFO gotoInfo;
			gotoInfo.type = GOTO_TYPE_BREAK;
			gotoInfo.commandIndex = (int)eval.code.size() - 1;
			gotoPending.push_back(gotoInfo);
		}

		// CONTINUE LABEL
		cmd.type = EVAL_COMMAND::TYPE_LABEL;
		eval.code.push_back(cmd);
		continueLabelIndex = (int)eval.code.size() - 1;
	}

	// INCREMENT
	if (forIncArg->subnode != NULL) {
		if (res == 0 && eval.generateCode) {
			cmd.type = EVAL_COMMAND::TYPE_CODE_BLOCK_BEGIN;
			assert(cmd.args.empty() && cmd.line != -1);
			eval.code.push_back(cmd);
		}

		res += Evaluate(forIncArg->subnode, eval);

		if (res == 0 && eval.generateCode) {
			cmd.type = EVAL_COMMAND::TYPE_CODE_BLOCK_END;
			eval.code.push_back(cmd);
		}
	}

	// BREAK LABEL
	if (res == 0 && eval.generateCode) {
		gotoCmd.line = tree->line;
		eval.code.push_back(gotoCmd);

		cmd.type = EVAL_COMMAND::TYPE_LABEL;
		eval.code.push_back(cmd);

		int breakLabelIndex = (int)eval.code.size() - 1;
		for (int i=(int)gotoPending.size() - 1; i >= 0; --i) {
			GOTO_INFO & gt = gotoPending[i];

			assert(eval.code[gt.commandIndex].type == EVAL_COMMAND::TYPE_GOTO);
			assert(eval.code[gt.commandIndex].args.empty());
			switch (gt.type) {
				case GOTO_TYPE_CONTINUE:	eval.code[gt.commandIndex].args.push_back(continueLabelIndex);	break;
				case GOTO_TYPE_BREAK:		eval.code[gt.commandIndex].args.push_back(breakLabelIndex);		break;
				default:					eval.gotoPending.push_back(gt);									break;
			}
		}
	}

	assert(eval.context.localVars.size() == localVarsStackSize + 1);
	eval.context.localVars.pop();

	if (res == 0 && eval.generateCode) {
		cmd.type = EVAL_COMMAND::TYPE_CODE_BLOCK_END;
		eval.code.push_back(cmd);
	}

	eval.typeinfo.SetRValue(defType[DEF_TYPE_VOID]);

	return res;
}

/************************************************************************/
/* tavr::PROJECT::EvalWhile
/*
/************************************************************************/
int PROJECT::EvalWhile(const NODE * tree, EVALUATE & eval)
{
	EVAL_COMMAND cmd;
	cmd.line = tree->line;

	// REPEAT LABEL
	int repeatLabelIndex = -1;
	if (eval.generateCode) {
		cmd.type = EVAL_COMMAND::TYPE_LABEL;
		eval.code.push_back(cmd);
		repeatLabelIndex = (int)eval.code.size() - 1;
	}

	// CONDITION
	int res = Evaluate(tree->subnode, eval);

	if (res == 0 && (!eval.typeinfo.IsValue() || !eval.typeinfo.IsBoolean())) {
		msg["name"] = "while";
		msg["name2"] = Type2Str(eval.typeinfo);
		PRINT_MSG(TVR1032, tree->subnode->line);
		++res;
	}

	if (res == 0 && eval.generateCode) {
		cmd.type = EVAL_COMMAND::TYPE_IF;
		cmd.args.push_back((int)eval.code.size() - 1);
		eval.code.push_back(cmd);
	}

	// WHILE BODY
	GOTO_TYPE_F oldGotoAvailable = eval.gotoAvailable;
	eval.gotoAvailable.set(GOTO_TYPE_BREAK | GOTO_TYPE_CONTINUE);
	eval.gotoAvailable.clear(GOTO_TYPE_REPEAT);

	GOTO_INFO_ARRAY gotoPending;
	swap(gotoPending, eval.gotoPending);

	assert(tree->subnode->next->next == NULL);
	res += Evaluate(tree->subnode->next, eval);

	eval.gotoAvailable = oldGotoAvailable;
	swap(gotoPending, eval.gotoPending);

	// BREAK GOTO	
	if (res == 0 && eval.generateCode) {
		assert(eval.code.back().type == EVAL_COMMAND::TYPE_CODE_BLOCK_END);

		cmd.type = EVAL_COMMAND::TYPE_GOTO;
		cmd.args.clear();
		cmd.args.push_back(repeatLabelIndex);
		eval.code.insert(eval.code.begin() + eval.code.size() - 1, cmd);

		cmd.type = EVAL_COMMAND::TYPE_ELSE;
		cmd.args.clear();
		eval.code.push_back(cmd);

		cmd.type = EVAL_COMMAND::TYPE_CODE_BLOCK_BEGIN;
		eval.code.push_back(cmd);

		cmd.type = EVAL_COMMAND::TYPE_CODE_BLOCK_END;
		eval.code.push_back(cmd);

		cmd.type = EVAL_COMMAND::TYPE_LABEL;
		eval.code.push_back(cmd);

		int breakLabelIndex = (int)eval.code.size() - 1;

		for (int i=(int)gotoPending.size() - 1; i >= 0; --i) {
			GOTO_INFO & gt = gotoPending[i];

			assert(eval.code[gt.commandIndex].type == EVAL_COMMAND::TYPE_GOTO);
			assert(eval.code[gt.commandIndex].args.empty());
			switch (gt.type) {
				case GOTO_TYPE_BREAK:		eval.code[gt.commandIndex].args.push_back(breakLabelIndex);		break;
				case GOTO_TYPE_CONTINUE:	eval.code[gt.commandIndex].args.push_back(repeatLabelIndex);	break;
				default:					eval.gotoPending.push_back(gt);									break;
			}
		}
	}

	return res;
}

/*************************************************************************/
/* PROJECT::EvalDoWhile()                                                */
/*                                                                       */
/*************************************************************************/
int PROJECT::EvalDoWhile(const NODE * tree, EVALUATE & eval)
{
	EVAL_COMMAND cmd;
	cmd.line = tree->line;

	// REPEAT LABEL
	int repeatLabelIndex = -1;
	if (eval.generateCode) {
		cmd.type = EVAL_COMMAND::TYPE_LABEL;
		eval.code.push_back(cmd);
		repeatLabelIndex = (int)eval.code.size() - 1;
	}

	// DO-WHILE BODY
	GOTO_TYPE_F oldGotoAvailable = eval.gotoAvailable;
	eval.gotoAvailable.set(GOTO_TYPE_BREAK | GOTO_TYPE_CONTINUE);
	eval.gotoAvailable.clear(GOTO_TYPE_REPEAT);

	GOTO_INFO_ARRAY gotoPending;
	swap(gotoPending, eval.gotoPending);

	assert(tree->subnode->next->next == NULL);
	int res = Evaluate(tree->subnode, eval);

	eval.gotoAvailable = oldGotoAvailable;
	swap(gotoPending, eval.gotoPending);

	assert(res == 0 || !eval.generateCode || eval.code.back().type == EVAL_COMMAND::TYPE_CODE_BLOCK_END);

	// CONDITION
	res += Evaluate(tree->subnode->next, eval);

	if (res == 0 && (!eval.typeinfo.IsValue() || !eval.typeinfo.IsBoolean())) {
		msg["name"] = "while";
		msg["name2"] = Type2Str(eval.typeinfo);
		PRINT_MSG(TVR1032, tree->subnode->line);
		++res;
	}

	if (res == 0 && eval.generateCode) {
		// IF-REPEAT BLOCK
		cmd.type = EVAL_COMMAND::TYPE_IF;
		cmd.args.push_back((int)eval.code.size() - 1);
		eval.code.push_back(cmd);

		cmd.type = EVAL_COMMAND::TYPE_GOTO;
		cmd.args[0] = repeatLabelIndex;
		eval.code.push_back(cmd);

		cmd.type = EVAL_COMMAND::TYPE_ELSE;
		cmd.args.clear();
		eval.code.push_back(cmd);

		cmd.type = EVAL_COMMAND::TYPE_CODE_BLOCK_BEGIN;
		eval.code.push_back(cmd);

		cmd.type = EVAL_COMMAND::TYPE_CODE_BLOCK_END;
		eval.code.push_back(cmd);

		// BREAK GOTO	
		assert(cmd.args.empty());
		cmd.type = EVAL_COMMAND::TYPE_LABEL;
		eval.code.push_back(cmd);

		const int breakLabelIndex = (int)eval.code.size() - 1;

		for (int i=(int)gotoPending.size() - 1; i >= 0; --i) {
			GOTO_INFO & gt = gotoPending[i];

			assert(eval.code[gt.commandIndex].type == EVAL_COMMAND::TYPE_GOTO);
			assert(eval.code[gt.commandIndex].args.empty());
			switch (gt.type) {
				case GOTO_TYPE_BREAK:		eval.code[gt.commandIndex].args.push_back(breakLabelIndex);		break;
				case GOTO_TYPE_CONTINUE:	eval.code[gt.commandIndex].args.push_back(repeatLabelIndex);	break;
				default:					eval.gotoPending.push_back(gt);									break;
			}
		}
	}

	return res;
}

/************************************************************************/
/* tavr::PROJECT::EvalStep
/*
/************************************************************************/
int PROJECT::EvalStep(const NODE * tree, EVALUATE & eval)
{
	assert(tree->type == NODE::TYPE_STEP
		|| tree->type == NODE::TYPE_LOOP
		|| tree->type == NODE::TYPE_FIXED
		|| tree->type == NODE::TYPE_SUBAUTO);

	int res = 0;

	// check for redefinition
	STEP_INFO_ARRAY & stepInfo = eval.steps.top();
	for (int i=(int)stepInfo.size() - 1; i >= 0; --i) {
		if (stepInfo[i].name == tree->name) {
			msg["name"] = tree->name.str();
			PRINT_MSG(TVR1016, tree->line);
			PRINT_MSG_INFO(TVR_SEEDECL_INFO, stepInfo[i].declInfo);
			++res;
			break;
		}
	}

	STEP_INFO si;
	si.name				= tree->name;
	si.labelIndex		= -1;
	si.declInfo.module	= eval.context.module;
	si.declInfo.line	= tree->line;

	GOTO_TYPE_F oldGotoAvailable = eval.gotoAvailable;
	eval.gotoAvailable.set(GOTO_TYPE_BREAK | GOTO_TYPE_REPEAT);
	eval.gotoAvailable.clear(GOTO_TYPE_CONTINUE);

	// REPEAT LABEL
	int repeatLabelIndex = -1;
	EVAL_COMMAND cmd;
	cmd.line = tree->line;
	if (res == 0 && eval.generateCode) {
		cmd.type = EVAL_COMMAND::TYPE_LABEL;
		eval.code.push_back(cmd);

		si.labelIndex = repeatLabelIndex = (int)eval.code.size() - 1;
	}

	// check for pending 'goto'
	if (si.name != emptyString) {
		stepInfo.push_back(si);

		GOTO_INFO_ARRAY gotoPending;
		gotoPending.reserve(eval.gotoPending.size());
		for (int i=(int)eval.gotoPending.size() - 1; i >= 0; --i) {
			GOTO_INFO & gt = eval.gotoPending[i];
			if (gt.type == GOTO_TYPE_GOTO && gt.name == si.name) {
				if (eval.generateCode) {
					assert(eval.code[gt.commandIndex].type == EVAL_COMMAND::TYPE_GOTO
						|| eval.code[gt.commandIndex].type == EVAL_COMMAND::TYPE_RELAX);
					assert(eval.code[gt.commandIndex].args.empty());
					eval.code[gt.commandIndex].args.push_back(si.labelIndex);
				}
			} else {
				gotoPending.push_back(gt);
			}
		}
		swap(gotoPending, eval.gotoPending);
	}

	// could be empty step
	if (tree->subnode == NULL) {
		eval.typeinfo.SetRValue(defType[DEF_TYPE_VOID]);
		return res;
	}

	// STEP BODY
	GOTO_INFO_ARRAY gotoPending;
	swap(eval.gotoPending, gotoPending);

	const int stepsStackSize = (int)eval.steps.size();
	eval.steps.push(STEP_INFO_ARRAY());

	res += Evaluate(tree->subnode, eval);

	assert(stepsStackSize + 1 == eval.steps.size());
	eval.steps.pop();

	eval.gotoAvailable = oldGotoAvailable;
	swap(eval.gotoPending, gotoPending);

	if (eval.generateCode) {
		switch (tree->type) {
			case NODE::TYPE_LOOP:
				cmd.type = EVAL_COMMAND::TYPE_GOTO;
				cmd.args.push_back(repeatLabelIndex);
				eval.code.push_back(cmd);
				break;
			case NODE::TYPE_FIXED:
				cmd.type = EVAL_COMMAND::TYPE_RELAX;
				cmd.args.push_back(repeatLabelIndex);
				eval.code.push_back(cmd);
				break;
			default:
				assert(tree->type == NODE::TYPE_STEP || tree->type == NODE::TYPE_SUBAUTO);
				break;
		}

		// BREAK LABEL
		cmd.type = EVAL_COMMAND::TYPE_LABEL;
		cmd.args.clear();
		eval.code.push_back(cmd);

		int breakLabelIndex = (int)eval.code.size() - 1;

		for (int i=(int)gotoPending.size() - 1; i >= 0; --i) {
			GOTO_INFO & gt = gotoPending[i];

			assert(eval.code[gt.commandIndex].type == EVAL_COMMAND::TYPE_GOTO
				|| eval.code[gt.commandIndex].type == EVAL_COMMAND::TYPE_RELAX);
			assert(eval.code[gt.commandIndex].args.empty());
			switch (gt.type) {
				case GOTO_TYPE_REPEAT:	eval.code[gt.commandIndex].args.push_back(repeatLabelIndex);	break;
				case GOTO_TYPE_BREAK:	eval.code[gt.commandIndex].args.push_back(breakLabelIndex);		break;
				default:				eval.gotoPending.push_back(gt);									break;
			}
		}
	}

	eval.typeinfo.SetRValue(defType[DEF_TYPE_VOID]);

	return res;
}

/************************************************************************/
/* tavr::PROJECT::EvalCodeBlock
/*
/************************************************************************/
int PROJECT::EvalCodeBlock(const NODE * tree, EVALUATE & eval)
{
	int res = 0;

	EVAL_COMMAND cmd;
	if (eval.generateCode) {
		cmd.type = EVAL_COMMAND::TYPE_CODE_BLOCK_BEGIN;
		cmd.line = tree->line;
		eval.code.push_back(cmd);
	}

	std::vector<const NODE *> statements;
	statements.reserve(256);

	const NODE * cur = tree->subnode;
	while (cur != NULL) {
		statements.push_back(cur);
		cur = cur->next;
	}

	const int localVarsStackSize = (int)eval.context.localVars.size();
	eval.context.localVars.push(VAR_SCOPE());

	const int stepsStackSize = (int)eval.steps.size();
	eval.steps.push(STEP_INFO_ARRAY());

	for (int i=(int)statements.size() - 1; i >= 0; --i) {
		res += Evaluate(statements[i], eval);
	}

	assert(stepsStackSize + 1 == eval.steps.size());
	eval.steps.pop();

	assert(eval.context.localVars.size() == localVarsStackSize + 1);
	eval.context.localVars.pop();

	if (res == 0 && eval.generateCode) {
		cmd.type = EVAL_COMMAND::TYPE_CODE_BLOCK_END;
		eval.code.push_back(cmd);
	}

	eval.typeinfo.SetRValue(defType[DEF_TYPE_VOID]);

	return res;
}

/************************************************************************/
/* tavr::PROJECT::VerifyBinOpTypeCompatibility
/*
/************************************************************************/
int PROJECT::VerifyBinOpTypeCompatibility(const NODE::TYPE type, int line, const char * opVal, const TYPEINFO & l, const TYPEINFO & r, METHOD *& overriddenOp)
{
	int res = 0;

	if (type != NODE::TYPE_DECL_VAR && type != NODE::TYPE_RETURN && !l.IsValue()) {
		msg["name"] = Type2Str(l);
		if (l.type == TYPEINFO::TYPE_TYPE) {
			PRINT_MSG(TVR1022, line);
		} else {
			msg["op"] = opVal;
			PRINT_MSG(TVR1023, line);
		}
		++res;
	}

	if (!r.IsValue()) {
		msg["name"] = Type2Str(r);
		if (r.type == TYPEINFO::TYPE_TYPE) {
			PRINT_MSG(TVR1022, line);
		} else {
			msg["op"] = opVal;
			PRINT_MSG(TVR1023, line);
		}
		++res;
	}

	if (res != 0) {
		return res;
	}

	switch (type) {
		case NODE::TYPE_EXPR_ASSIGN:
			if (!l.IsLValue()) {
				PRINT_MSG(TVR1014, line);
				return 1;
			}

			// check for overridden operator =
			if (l.IsUserType() && l.info.type != r.info.type) {
				TYPE * type = l.info.type;
				for (int i=(int)type->methods.size() - 1; i >= 0; --i) {
					METHOD * method = type->methods[i];
					if (method->type != METHOD::TYPE_OPERATOR || method->name.str() != "=") {
						continue;
					}

					if (method->args[0].type.IsTypeEqual(r)) {
						overriddenOp = method;
						return 0;
					}
				}
			}
			break;

		case NODE::TYPE_EXPR_LESS:
		case NODE::TYPE_EXPR_LESS_EQ:
		case NODE::TYPE_EXPR_GREATER:
		case NODE::TYPE_EXPR_GREATER_EQ:
			// TODO: support overridden operators
			if (l.IsTypeByRef()) {
				msg["op"] = opVal;
				msg["name"] = Type2Str(l);
				PRINT_MSG(TVR1023, line);
				return 1;
			}
			assert(!l.IsTypeByRef() && !r.IsTypeByRef());
			break;
	}

	if (!l.IsTypeEqual(r)) {
		msg["op"]		= opVal;
		msg["name"]		= Type2Str(l);
		msg["name2"]	= Type2Str(r);
		PRINT_MSG(TVR1024, line);
		return 1;
	}

	return 0;
}

/************************************************************************/
/* tavr::PROJECT::EvalExprNewArray
/*
/************************************************************************/
int PROJECT::EvalExprNewArray(const NODE * tree, EVALUATE & eval)
{
	int res = 0;

	// retrieve array type
	bool dontGenerateCode = false;
	swap(dontGenerateCode, eval.generateCode);
	res += Evaluate(tree->subnode, eval);
	swap(dontGenerateCode, eval.generateCode);

	if (res == 0 && eval.typeinfo.type != TYPEINFO::TYPE_TYPE) {
		msg["name"] = "new";
		PRINT_MSG(TVR1025, tree->subnode->line);
		++res;
	}

	EVAL_COMMAND cmd;
	cmd.type		= EVAL_COMMAND::TYPE_NEW_ARRAY;
	cmd.typeinfo	= eval.typeinfo;
	++cmd.typeinfo.array;

	// elements count
	assert(tree->subnode->next != NULL);
	assert(tree->subnode->next->next == NULL);

	res += Evaluate(tree->subnode->next, eval);

	// array item count should be an integral value
	if (res == 0 && !eval.typeinfo.IsIntegral()) {
		PRINT_MSG(TVR1017, tree->line);
		++res;
	}

	if (res == 0 && eval.generateCode) {
		cmd.args.push_back((int)eval.code.size() - 1);
		cmd.line = tree->subnode->line;
		eval.code.push_back(cmd);
	}

	eval.typeinfo = cmd.typeinfo;
	eval.typeinfo.type = TYPEINFO::TYPE_R_VALUE;

	return res;
}

/************************************************************************/
/* tavr::PROJECT::EvalExprNewAuto
/*
/************************************************************************/
int PROJECT::EvalExprNewAuto(const NODE * tree, EVALUATE & eval)
{
	int res = 0;

	bool dontGenerateCode = false;
	swap(eval.generateCode, dontGenerateCode);
	TYPE * curType = NULL;
	swap(curType,		eval.context.curType);
	res += Evaluate(tree->subnode, eval);
	swap(curType,		eval.context.curType);
	swap(eval.generateCode, dontGenerateCode);

	if (res == 0) {
		if (eval.typeinfo.type != TYPEINFO::TYPE_TYPE) {
			msg["name"] = "new";
			PRINT_MSG(TVR1025, tree->subnode->line);
			++res;
		} else if (!eval.typeinfo.IsAuto() && !eval.typeinfo.IsEvent()) {
			msg["name"] = Type2Str(eval.typeinfo);
			PRINT_MSG(TVR1035, tree->subnode->line);
			++res;
		} else if (eval.typeinfo.info.type->attribs.is(ATTRIB_STATIC)) {
			msg["name"] = Type2Str(eval.typeinfo);
			PRINT_MSG(TVR1056, tree->subnode->line);
			++res;
		}
	}

	TYPE * type = eval.typeinfo.info.type;

	EVAL_COMMAND cmd;
	METHOD method;
	res += EvalProcessMethodArgs(tree->subnode->next, cmd, method, eval);

	if (res != 0) {
		return res;
	}

	method.type = METHOD::TYPE_CONSTRUCTOR;
	method.name = type->name;

	LOOKUP_CONTEXT ctxt;
	ctxt.curType = type;
	bool ambiguity;
	METHOD * found = MethodLookup(&method, ctxt, ambiguity, LOOKUP_TYPE_LOCALS_ONLY);

	if (found == NULL) {
		method.owner.type = type;
		msg["name"] = Type2Str(&method);
		if (!ambiguity) {
			PRINT_MSG(TVR1030, tree->subnode->line);
		} else {
			PRINT_MSG(TVR1045, tree->subnode->line);
			MethodLookup(&method, ctxt, ambiguity, LOOKUP_TYPE_LOCALS_ONLY | LOOKUP_TYPE_PRINT_AMBIGUITY_INFO);
		}
		return 1;
	}

	eval.typeinfo.SetRValue(type);

	if (eval.generateCode) {
		cmd.type = EVAL_COMMAND::TYPE_NEW_AUTO;
		cmd.info.proc = found;
		cmd.typeinfo = eval.typeinfo;
		cmd.line = tree->subnode->line;
		eval.code.push_back(cmd);
	}

	return 0;
}

/************************************************************************/
/* tavr::PROJECT::EvalExprTypeCast
/*
/************************************************************************/
int PROJECT::EvalExprTypeCast(const NODE * tree, EVALUATE & eval)
{
	// cast type
	int res = Evaluate(tree->subnode, eval);
	if (res != 0) {
		return res;
	}

	if (res == 0 && eval.typeinfo.type != TYPEINFO::TYPE_TYPE) {
		msg["name"] = "()";
		PRINT_MSG(TVR1025, tree->subnode->line);
		return 1;
	}

	METHOD method;
	method.name = castOpString;
	method.result = eval.typeinfo;

	// cast expression
	const NODE * castExpr = tree->subnode->next;
	res += Evaluate(castExpr, eval);
	if (res != 0) {
		return res;
	}

	if (res == 0 && !eval.typeinfo.IsValue()) {
		msg["name"] = Type2Str(eval.typeinfo);
		PRINT_MSG(TVR1022, castExpr->line);
		return 1;
	}

	// method lookup is not needed if it is a '(Ref)null'
	if (method.result.IsTypeByRef() && eval.typeinfo.IsNull()) {
		// do not generate code for null, because it is already generated

		eval.typeinfo = method.result;
		eval.typeinfo.type = TYPEINFO::TYPE_R_VALUE;
		return 0;
	}

	NAMESPACE * curNamespace = NULL;
	TYPE * curType = eval.typeinfo.info.type;
	swap(eval.context.curNamespace, curNamespace);
	swap(eval.context.curType,		curType);

	bool ambiguity;
	METHOD * found = MethodLookup(&method, eval.context, ambiguity, LOOKUP_TYPE_LOCALS_ONLY);
	if (found == NULL) {
		assert(!ambiguity);

		method.type = METHOD::TYPE_OPERATOR;
		method.owner.type = eval.context.curType;
		msg["name"] = Type2Str(&method);
		PRINT_MSG(TVR1030, tree->subnode->line);
		++res;
	}

	swap(eval.context.curNamespace, curNamespace);
	swap(eval.context.curType,		curType);

	if (res != 0) {
		return res;
	}

	if (eval.generateCode) {
		EVAL_COMMAND cmd;
		cmd.args.push_back((int)eval.code.size() - 1);

		if (found->impl != NULL) {
			cmd.type		= EVAL_COMMAND::TYPE_THIS_CALL;
			cmd.info.proc	= found;
		} else {
			// standard type cast
			cmd.type		= EVAL_COMMAND::TYPE_TYPE_CAST;
		}

		cmd.typeinfo	= found->result;
		cmd.line		= tree->subnode->line;
		eval.code.push_back(cmd);
	}

	// TODO: could be l-value in some cases
	eval.typeinfo = found->result;
	eval.typeinfo.type = TYPEINFO::TYPE_R_VALUE;

	return res;
}

/************************************************************************/
/* tavr::PROJECT::VerifyEscapeSequences
/*
/************************************************************************/
int PROJECT::VerifyEscapeSequences(const std::string & value, int line)
{
	int res = 0;
	bool isInEscape = false;
	for (int i=0; i<(int)value.size(); ++i) {
		if (!isInEscape) {
			if (value[i] == '\\') {
				isInEscape = true;
				continue;
			}
			continue;
		}

		switch (value[i]) {
			case 'n':	case 't':
			case '\\':	case '\'':
			case '"':
				break;

			default:
				msg["name"] = value.substr(i - 1, 2);
				PRINT_MSG(TVR1051, line);
				++res;
				break;
		}

		isInEscape = false;
	}

	return res;
}

/************************************************************************/
/* tavr::PROJECT::EvalGotoRelax
/*
/************************************************************************/
int PROJECT::EvalGotoRelax(const NODE * tree, EVALUATE & eval)
{
	const bool isRelax = (tree->type == NODE::TYPE_RELAX);

	// 'relax' 'step'
	if (tree->type == NODE::TYPE_RELAX_STEP) {
		assert(eval.context.curMethod != NULL);
		if (eval.context.curMethod->type != METHOD::TYPE_AUTO_PROC) {
			msg["name"] = "relax";
			PRINT_MSG(TVR1033, tree->line);
			return 1;
		}

		if (eval.generateCode) {
			EVAL_COMMAND cmd;
			cmd.line = tree->line;
			cmd.type = EVAL_COMMAND::TYPE_RELAX;
			eval.code.push_back(cmd);

			cmd.type = EVAL_COMMAND::TYPE_LABEL;
			eval.code.push_back(cmd);

			eval.code[eval.code.size() - 2].args.push_back((int)eval.code.size() - 1);
		}

		return 0;
	}

	assert(isRelax || tree->type == NODE::TYPE_GOTO);

	// 'relax' without step name works like 'repeat'
	if (isRelax && tree->name == emptyString) {
		assert(eval.context.curMethod != NULL);

		// TODO: support named 'relax'
		assert(tree->name == emptyString);

		// 'relax' statement could be used only in 'auto' method
		if (!eval.gotoAvailable.is(GOTO_TYPE_REPEAT)
			|| eval.context.curMethod->name != autoString)
		{
			msg["name"] = "relax";
			PRINT_MSG(TVR1033, tree->line);
			return 1;
		}

		if (eval.generateCode) {
			EVAL_COMMAND cmd;
			cmd.line		= tree->line;
			cmd.type		= EVAL_COMMAND::TYPE_RELAX;
			eval.code.push_back(cmd);

			GOTO_INFO gt;
			gt.commandIndex	= (int)eval.code.size() - 1;
			gt.type			= GOTO_TYPE_REPEAT;
			eval.gotoPending.push_back(gt);
		}

		return 0;
	}

	int gotoIndex = -1;
	if (eval.generateCode) {
		EVAL_COMMAND cmd;
		if (!isRelax) {
			cmd.type	= EVAL_COMMAND::TYPE_GOTO;
		} else {
			cmd.type	= EVAL_COMMAND::TYPE_RELAX;
		}
		cmd.line	= tree->line;
		eval.code.push_back(cmd);

		gotoIndex = (int)eval.code.size() - 1;
	}

	// look for declared step
	bool found = false;
	for (int stackIndex = (int)eval.steps.size() - 1; stackIndex >= 0; --stackIndex) {
		STEP_INFO_ARRAY & stepInfo = eval.steps.c[stackIndex];
		for (int i=(int)stepInfo.size() - 1; i >= 0; --i) {
			if (stepInfo[i].name == tree->name) {
				found = true;
				if (eval.generateCode) {
					assert(gotoIndex == (int)eval.code.size() - 1);
					eval.code[gotoIndex].args.push_back(stepInfo[i].labelIndex);
				}
			}
		}
	}

	if (!found) {
		GOTO_INFO gt;
		gt.type 		= GOTO_TYPE_GOTO;
		gt.name 		= tree->name;
		gt.line			= tree->line;
		gt.commandIndex	= gotoIndex;

		eval.gotoPending.push_back(gt);
	}

	return 0;
}

/*************************************************************************/
/* PROJECT::EvalSendPost()                                               */
/*                                                                       */
/*************************************************************************/
int PROJECT::EvalSendPost(const NODE * tree, EVALUATE & eval)
{
	const bool isSend = (tree->type == NODE::TYPE_SEND);
	assert(isSend || tree->type == NODE::TYPE_POST);

	int res = 0;

	const NODE * eventTree = tree->subnode->subnode;
	const NODE * nameTree = eventTree ? eventTree->subnode : NULL;
	const NODE * nameArgsTree = nameTree ? nameTree->next : NULL;
	const NODE * destTree = tree->subnode->next;
	assert((!eventTree || nameTree && nameArgsTree) && destTree && destTree->next == NULL);

	// ensure correct post context
	if (!isSend && (eval.context.curType == NULL
	||  eval.context.curMethod->type != METHOD::TYPE_AUTO_PROC)) {
		msg["name"] = "post";
		PRINT_MSG(TVR1033, nameTree ? nameTree->subnode->line : tree->line);
		return 1;
	}

	EVAL_COMMAND cmd;
	cmd.line = tree->line;

	// process event name
	TYPE * eventType = NULL;
	int eventCmdIndex = -1;
	if (nameTree != NULL) {
		// post E2(...) [to ...]

		bool dontGenerateCode = false;
		swap(dontGenerateCode, eval.generateCode);
		res += Evaluate(nameTree->subnode, eval);
		swap(dontGenerateCode, eval.generateCode);

		if (res > 0) {
			return res;
		}

		if (!eval.typeinfo.IsEvent()) {
			msg["name"] = nameTree->subnode->name.str();
			PRINT_MSG(TVR1058, nameTree->subnode->line);
			return 1;
		}

		assert(eval.typeinfo.type == TYPEINFO::TYPE_TYPE);
		eventType = eval.typeinfo.info.type;

		cmd.type = EVAL_COMMAND::TYPE_NEW_AUTO;
		cmd.line = nameTree->subnode->line;

		// process event constructor args
		METHOD method;
		method.type = METHOD::TYPE_CONSTRUCTOR;
		method.name = eventType->name;
		METHOD::ARG tmpArg;
		for (const NODE * arg = nameArgsTree->subnode; arg != NULL; arg = arg->next) {
			res += Evaluate(arg, eval);
			tmpArg.type = eval.typeinfo;
			method.args.push_back(tmpArg);

			if (res == 0 && !eval.typeinfo.IsValue()) {
				msg["name"] = Type2Str(eval.typeinfo);
				PRINT_MSG(TVR1022, arg->line);
				++res;
				break;
			}

			cmd.args.push_back((int)eval.code.size() - 1);
		}

		if (res > 0) {
			return res;
		}

		// lookup event constructor
		std::reverse(method.args.begin(), method.args.end());
		std::reverse(cmd.args.begin(), cmd.args.end());

		swap(eventType, eval.context.curType);
		bool ambiguity = false;
		METHOD * eventConstr = MethodLookup(&method, eval.context, ambiguity, LOOKUP_TYPE_LOCALS_ONLY);
		swap(eventType, eval.context.curType);

		cmd.info.proc = eventConstr;

		if (eventConstr == NULL) {
			method.owner.type = eventType;
			msg["name"] = Type2Str(&method);

			if (!ambiguity) {
				PRINT_MSG(TVR1030, nameTree->subnode->line);
			} else {
				PRINT_MSG(TVR1045, nameTree->subnode->line);
				swap(eventType, eval.context.curType);
				MethodLookup(&method, eval.context, ambiguity, LOOKUP_TYPE_LOCALS_ONLY | LOOKUP_TYPE_PRINT_AMBIGUITY_INFO);
				swap(eventType, eval.context.curType);
			}
			return 1;
		}

		if (eval.generateCode) {
			cmd.typeinfo.SetRValue(eventType);
			eval.code.push_back(cmd);
		}

		eventCmdIndex = (int)eval.code.size() - 1;
	} else {
		// post [to ...]

		if (eval.eventType == NULL) {
			msg["name"] = "post";
			PRINT_MSG(TVR1033, tree->line);
			return 1;
		}

		eventType = eval.eventType;

		if (eval.generateCode) {
			cmd.type = EVAL_COMMAND::TYPE_VARIABLE;
			cmd.info.name = eval.code[eval.eventDeclIdx].info.name;
			cmd.typeinfo.SetRValue(eventType);
			cmd.args.push_back(eval.eventDeclIdx);
			assert(cmd.args.size() == 1);
			eval.code.push_back(cmd);
		}

		eventCmdIndex = (int)eval.code.size() - 1;
	}

	assert(eventCmdIndex != -1 || !eval.generateCode);

	// process destination value
	if (destTree->subnode != NULL) {
		res += Evaluate(destTree->subnode, eval);
		if (res > 0) {
			return res;
		}

		if ((eval.typeinfo.type == TYPEINFO::TYPE_PROC_NAME
		||  eval.typeinfo.type == TYPEINFO::TYPE_MEMBER_PROC_NAME)
		&&  eval.typeinfo.info.procName->type == METHOD::TYPE_AUTO_PROC) {
			const bool needThis = eval.typeinfo.type == TYPEINFO::TYPE_PROC_NAME;

			cmd.info.proc = eval.typeinfo.info.procName;
			eval.typeinfo.SetRValue(cmd.info.proc->owner.type);

			if (eval.generateCode && needThis) {
				EVAL_COMMAND c = cmd;
				c.type = EVAL_COMMAND::TYPE_THIS;
				c.args.clear();
				c.typeinfo = eval.typeinfo;
				eval.code.push_back(c);
			}

			assert(cmd.info.proc->owner.type->specs[TYPE::SPEC_AUTO_PROC] != NULL);
		} else {
			if (!eval.typeinfo.IsValue()) {
				msg["name"] = Type2Str(eval.typeinfo);
				PRINT_MSG(TVR1022, destTree->subnode->line);
				return 1;
			}

			// check for 'auto' type
			cmd.info.proc = eval.typeinfo.info.type->specs[TYPE::SPEC_AUTO_PROC];
			if (cmd.info.proc == NULL) {
				msg["name"] = Type2Str(eval.typeinfo);
				PRINT_MSG(TVR1060, destTree->subnode->line);
				return 1;
			}
		}

		if (!eval.typeinfo.IsAuto() && (!eval.typeinfo.IsStruct() || !isSend)) {
			msg["name"] = Type2Str(eval.typeinfo);
			PRINT_MSG(TVR1059, destTree->subnode->line);
			return 1;
		}
	} else {
		if (isSend) {
			PRINT_MSG(TVR1061, nameTree->subnode->line);
			return 1;
		}

		if (eval.context.curType == NULL || eval.context.curMethod->attribs.is(ATTRIB_STATIC)) {
			PRINT_MSG(TVR1062, nameTree->subnode->line);
			return 1;
		}

		cmd.type = EVAL_COMMAND::TYPE_THIS;
		cmd.args.clear();
		cmd.typeinfo.SetRValue(eval.context.curType);

		if (eval.generateCode) {
			eval.code.push_back(cmd);
		}

		TYPE * curType = eval.context.curType;
		assert(curType->specs[TYPE::SPEC_AUTO_PROC] != NULL);
		cmd.info.proc = curType->specs[TYPE::SPEC_AUTO_PROC];
	}

	eval.typeinfo.SetRValue(defType[DEF_TYPE_VOID]);

	// add send command
	if (eval.generateCode) {
		assert(res == 0);

		cmd.type = isSend ? EVAL_COMMAND::TYPE_SEND : EVAL_COMMAND::TYPE_POST;

		cmd.args.clear();
		cmd.args.push_back(eventCmdIndex);				// event
		cmd.args.push_back((int)eval.code.size() - 1);	// destination
		cmd.typeinfo = eval.typeinfo;

		eval.code.push_back(cmd);
	}

	return 0;
}

/*************************************************************************/
/* PROJECT::EvalOn()                                                     */
/*                                                                       */
/*************************************************************************/
int PROJECT::EvalOn(const NODE * tree, EVALUATE & eval)
{
	const bool isOn = (tree->type == NODE::TYPE_ON);

	const NODE * eventTree = isOn ? tree->subnode : NULL;
	const NODE * codeTree  = isOn ? eventTree->next : tree;

	// verify on-block context
	if (eval.context.curType == NULL || eval.context.curMethod->attribs.is(ATTRIB_STATIC)
	|| (eval.context.curType->subtype != TYPE::SUBTYPE_AUTO && eval.context.curType->subtype != TYPE::SUBTYPE_STRUCT)) {
		msg["name"] = isOn ? "on" : "otherwise";
		PRINT_MSG(TVR1033, eventTree->subnode->line);
		return 1;
	}

	if (isOn) {
		// process event
		bool dontGenerateCode = false;
		swap(dontGenerateCode, eval.generateCode);
		int res = Evaluate(eventTree->subnode, eval);
		swap(dontGenerateCode, eval.generateCode);

		if (res > 0) {
			return res;
		}

		assert(eval.typeinfo.type == TYPEINFO::TYPE_TYPE);
		if (!eval.typeinfo.IsEvent()) {
			msg["name"] = Type2Str(eval.typeinfo);
			PRINT_MSG(TVR1063, eventTree->subnode->line);
			return 1;
		}
	} else {
		// null means any event (i.e. 'otherwise')
		eval.typeinfo.SetType(defType[DEF_TYPE_NULL]);
	}

	// process code
	EVAL_COMMAND cmd;
	cmd.type = EVAL_COMMAND::TYPE_ON;
	cmd.line = isOn ? eventTree->subnode->line : tree->line;
	cmd.typeinfo = eval.typeinfo;

	int eventIdx = 0; // event declaration index in the code
	if (eval.generateCode) {
		eval.code.push_back(cmd);

		cmd.type = EVAL_COMMAND::TYPE_CODE_BLOCK_BEGIN;
		eval.code.push_back(cmd);

		cmd.type = EVAL_COMMAND::TYPE_EVENT;
		eval.code.push_back(cmd);

		cmd.type = EVAL_COMMAND::TYPE_DECL;
		cmd.info.name = eventString;
		cmd.args.push_back((int)eval.code.size() - 1);
		eval.code.push_back(cmd);

		// store 'event' variable declaration
		eventIdx = (int)eval.code.size() - 1;

		cmd.type = EVAL_COMMAND::TYPE_CONSTANT;
		cmd.info.name = nullString;
		cmd.args.clear();
		cmd.typeinfo.SetRValue(defType[DEF_TYPE_NULL]);
		eval.code.push_back(cmd);

		cmd.type = EVAL_COMMAND::TYPE_EVENT;
		cmd.typeinfo = eval.typeinfo;
		eval.code.push_back(cmd);

		cmd.type = EVAL_COMMAND::TYPE_ASSIGN;
		cmd.args.push_back((int)eval.code.size() - 1);
		cmd.args.push_back((int)eval.code.size() - 2);
		eval.code.push_back(cmd);
	}

	TYPE * eventType = eval.typeinfo.info.type;
	assert(eventType->subtype == TYPE::SUBTYPE_EVENT
		|| eventType->subtype == TYPE::SUBTYPE_NULL);
	swap(eventIdx, eval.eventDeclIdx);
	assert(!eval.generateCode || eval.eventDeclIdx >= 0);
	swap(eventType, eval.eventType);

	int res = Evaluate(codeTree->subnode, eval);

	swap(eventIdx, eval.eventDeclIdx);
	swap(eventType, eval.eventType);

	if (res > 0) {
		return res;
	}

	if (eval.generateCode) {
		cmd.args.clear();
		cmd.type = EVAL_COMMAND::TYPE_CODE_BLOCK_END;
		eval.code.push_back(cmd);
		cmd.type = EVAL_COMMAND::TYPE_ELSE;
		eval.code.push_back(cmd);
		cmd.type = EVAL_COMMAND::TYPE_CODE_BLOCK_BEGIN;
		eval.code.push_back(cmd);
		cmd.type = EVAL_COMMAND::TYPE_CODE_BLOCK_END;
		eval.code.push_back(cmd);
	}

	return 0;
}

/*************************************************************************/
/* PROJECT::EvalReg()                                                    */
/*                                                                       */
/*************************************************************************/
int PROJECT::EvalReg(const NODE * tree, EVALUATE & eval)
{
	// process subscribe auto
	int res = Evaluate(tree->subnode, eval);
	if (res > 0) {
		return res;
	}

	// it should be any 'out' action proc
	if (eval.typeinfo.type != TYPEINFO::TYPE_MEMBER_PROC_NAME
	&&  eval.typeinfo.type != TYPEINFO::TYPE_PROC_NAME
	||  eval.typeinfo.info.procName->type != METHOD::TYPE_OUT) {
		msg["name"] = Type2Str(eval.typeinfo);
		PRINT_MSG(TVR1066, tree->subnode->line);
		return 1;
	}

	const bool isThisSubscribe = (eval.typeinfo.type == TYPEINFO::TYPE_PROC_NAME);

	// prepare eval command
	TYPE * autoType = isThisSubscribe ? eval.context.curType : eval.typeinfo.info.procName->owner.type;

	EVAL_COMMAND cmd;
	cmd.line = tree->subnode->line;

	if (isThisSubscribe && eval.generateCode) {
		cmd.type = EVAL_COMMAND::TYPE_THIS;
		cmd.typeinfo.SetRValue(autoType);
		eval.code.push_back(cmd);
	}

	switch (tree->type) {
		case NODE::TYPE_REG_POST: cmd.type = EVAL_COMMAND::TYPE_REG_POST; break;
		case NODE::TYPE_REG_SEND: cmd.type = EVAL_COMMAND::TYPE_REG_SEND; break;
		case NODE::TYPE_UNREG:    cmd.type = EVAL_COMMAND::TYPE_UNREG; break;
		default: assert(!"Unknown registration type"); break;
	}
	cmd.args.push_back((int)eval.code.size() - 1);
	assert(!eval.generateCode || eval.code.back().typeinfo.info.type == autoType);
	cmd.info.name = eval.typeinfo.info.procName->name;

	// process subscriber
	res += Evaluate(tree->subnode->next, eval);
	if (res > 0) {
		return res;
	}

	if ((eval.typeinfo.type == TYPEINFO::TYPE_MEMBER_PROC_NAME
	||  eval.typeinfo.type == TYPEINFO::TYPE_PROC_NAME)
	&&  eval.typeinfo.info.procName->type == METHOD::TYPE_AUTO_PROC
	&&  tree->type != NODE::TYPE_UNREG) {
		const bool needThis = (eval.typeinfo.type == TYPEINFO::TYPE_PROC_NAME);
		
		cmd.info.proc = eval.typeinfo.info.procName;

		if (needThis && eval.generateCode) {
			EVAL_COMMAND c = cmd;
			c.type = EVAL_COMMAND::TYPE_THIS;
			c.args.clear();
			c.typeinfo.SetRValue(cmd.info.proc->owner.type);
			eval.code.push_back(c);
		}
	} else {
		if (!eval.typeinfo.IsValue()) {
			msg["name"] = Type2Str(eval.typeinfo);
			PRINT_MSG(TVR1022, tree->subnode->next->line);
			return 1;
		}

		if (!eval.typeinfo.IsAuto()) {
			msg["name"] = Type2Str(eval.typeinfo);
			PRINT_MSG(TVR1067, tree->subnode->next->line);
			return 1;
		}

		// auto should have auto proc
		cmd.info.proc = eval.typeinfo.info.type->specs[TYPE::SPEC_AUTO_PROC];
		if (cmd.info.proc == NULL) {
			msg["name"] = Type2Str(eval.typeinfo);
			PRINT_MSG(TVR1060, tree->subnode->line);
			return 1;
		}
	}

	cmd.args.push_back((int)eval.code.size() - 1);

	if (eval.generateCode) {
		eval.code.push_back(cmd);
	}

	return 0;
}

/*************************************************************************/
/* PROJECT::EvalExprIf()                                                 */
/*                                                                       */
/*************************************************************************/
int PROJECT::EvalExprIf(const NODE * tree, EVALUATE & eval)
{
	// process condition
	int res = Evaluate(tree->subnode, eval);
	if (res > 0) {
		return res;
	}

	if (!eval.typeinfo.IsBoolean()) {
		msg["name"] = "?:";
		msg["name2"] = Type2Str(eval.typeinfo);
		PRINT_MSG(TVR1032, tree->subnode->line);
		return 1;
	}

	EVAL_COMMAND cmd;
	cmd.line = tree->subnode->line;
	int condCmdIndex = 0;
	if (eval.generateCode) {
		cmd.type = EVAL_COMMAND::TYPE_EXPR_IF_COND;
		cmd.args.push_back((int)eval.code.size() - 1);
		eval.code.push_back(cmd);
		condCmdIndex = (int)eval.code.size() - 1;
	}

	// process left expression
	res += Evaluate(tree->subnode->next, eval);
	if (res > 0) {
		return res;
	}

	TYPEINFO resultType = eval.typeinfo;

	int elseCmdIndex = 0;
	if (eval.generateCode) {
		cmd.type = EVAL_COMMAND::TYPE_EXPR_IF_ELSE;
		cmd.args.clear();
		cmd.args.push_back(condCmdIndex);
		cmd.args.push_back((int)eval.code.size() - 1);
		eval.code.push_back(cmd);
		elseCmdIndex = (int)eval.code.size() - 1;
	}

	// process right expression
	res += Evaluate(tree->subnode->next->next, eval);
	if (res > 0) {
		return res;
	}

	if (eval.generateCode) {
		cmd.type = EVAL_COMMAND::TYPE_EXPR_IF_RES;
		cmd.args.clear();
		cmd.args.push_back(elseCmdIndex);
		cmd.args.push_back((int)eval.code.size() - 1);
		eval.code.push_back(cmd);
	}

	// analyze result type
	if (!resultType.IsEqual(eval.typeinfo)) {
		if (resultType.IsNull() && eval.typeinfo.IsTypeByRef()) {
			resultType = eval.typeinfo;
		} else if (resultType.IsTypeByRef() && eval.typeinfo.IsNull()) {
			// that's ok
		} else {
			msg["op"] = "?:";
			msg["name"] = Type2Str(resultType);
			msg["name2"] = Type2Str(eval.typeinfo);
			PRINT_MSG(TVR1024, tree->subnode->next->line);
			return 1;
		}
	}

	if (eval.generateCode) {
		eval.code[condCmdIndex].typeinfo = resultType;
		eval.code.back().typeinfo		 = resultType;
	}
	eval.typeinfo = resultType;

	return 0;
}

/*************************************************************************/
/* PROJECT::EvalExprOrAnd()                                              */
/*                                                                       */
/*************************************************************************/
int PROJECT::EvalExprOrAnd(const NODE * tree, EVALUATE & eval)
{
	const bool isOr = (tree->type == NODE::TYPE_EXPR_LOG_OR);
	assert(isOr || tree->type == NODE::TYPE_EXPR_LOG_AND);

	// left value
	int res = Evaluate(tree->subnode, eval);
	if (res > 0) {
		return res;
	}

	if (!eval.typeinfo.IsBoolean()) {
		msg["name"] = isOr ? "||" : "&&";
		msg["name2"] = Type2Str(eval.typeinfo);
		PRINT_MSG(TVR1032, tree->subnode->line);
		return 1;
	}

	EVAL_COMMAND cmd;
	cmd.type = EVAL_COMMAND::TYPE_EXPR_IF_COND;
	cmd.line = tree->subnode->line;
	cmd.args.push_back((int)eval.code.size() - 1);
	cmd.typeinfo.SetRValue(defType[DEF_TYPE_BOOL]);

	EVAL_COMMAND elseCmd;
	elseCmd.line = cmd.line;
	elseCmd.type = EVAL_COMMAND::TYPE_EXPR_IF_ELSE;

	int elseCmdIndex = 0;
	if (eval.generateCode) {
		eval.code.push_back(cmd);
		elseCmd.args.push_back((int)eval.code.size() - 1);
		
		if (isOr) {
			cmd.type = EVAL_COMMAND::TYPE_CONSTANT;
			cmd.info.name = trueString;
			cmd.args.clear();
			eval.code.push_back(cmd);

			elseCmd.args.push_back((int)eval.code.size() - 1);
			eval.code.push_back(elseCmd);
			elseCmdIndex = (int)eval.code.size() - 1;
		}
	}

	// right value
	res += Evaluate(tree->subnode->next, eval);

	if (!eval.typeinfo.IsBoolean()) {
		msg["name"] = isOr ? "||" : "&&";
		msg["name2"] = Type2Str(eval.typeinfo);
		PRINT_MSG(TVR1032, tree->subnode->line);
		return 1;
	}

	if (eval.generateCode) {
		if (!isOr) {
			elseCmd.args.push_back((int)eval.code.size() - 1);
			eval.code.push_back(elseCmd);
			elseCmdIndex = (int)eval.code.size() - 1;

			cmd.type = EVAL_COMMAND::TYPE_CONSTANT;
			cmd.info.name = falseString;
			cmd.args.clear();
			eval.code.push_back(cmd);
		}

		cmd.type = EVAL_COMMAND::TYPE_EXPR_IF_RES;
		cmd.args.clear();
		cmd.args.push_back(elseCmdIndex);
		cmd.args.push_back((int)eval.code.size() - 1);
		eval.code.push_back(cmd);
	}

	return 0;
}

/*************************************************************************/
/* PROJECT::EvalExprBinOp()                                              */
/*                                                                       */
/*************************************************************************/
int PROJECT::EvalExprBinOp(const NODE * tree, EVALUATE & eval)
{
	// left value
	int res = Evaluate(tree->subnode, eval);
	TYPEINFO leftType = eval.typeinfo;

	EVAL_COMMAND cmd;
	cmd.args.reserve(2);
	if (res == 0 && eval.generateCode) {
		assert(!eval.code.empty());
		cmd.args.push_back((int)eval.code.size() - 1);
	}

	// right value
	res += Evaluate(tree->subnode->next, eval);
	TYPEINFO rightType = eval.typeinfo;

	METHOD * overridden = NULL;
	if (res == 0) {
		const char * opVal = NULL;
#define CASE_OPVAL(TYPE, VAL) case NODE::TYPE_EXPR_##TYPE: opVal = VAL; break;
		switch (tree->type) {
			CASE_OPVAL(ASSIGN,		"=");

			CASE_OPVAL(ADD, 		"+");
			CASE_OPVAL(SUB, 		"-");
			CASE_OPVAL(MUL, 		"*");
			CASE_OPVAL(DIV, 		"/");
			CASE_OPVAL(MOD, 		"%");

			CASE_OPVAL(EQ,			"==");
			CASE_OPVAL(NOT_EQ,		"!=");
			CASE_OPVAL(LESS,		"<" );
			CASE_OPVAL(LESS_EQ,		"<=");
			CASE_OPVAL(GREATER,		">" );
			CASE_OPVAL(GREATER_EQ,	">=");

			default: assert(!"Unknown operator type"); break;
		}
#undef  CASE_OPVAL

		res += VerifyBinOpTypeCompatibility(tree->type, tree->line, opVal, leftType, rightType, overridden);
	}

	assert(res > 0 || leftType.IsTypeByRef() == leftType.IsTypeByRef());

	switch (tree->type) {
		case NODE::TYPE_EXPR_EQ:
		case NODE::TYPE_EXPR_NOT_EQ:
		case NODE::TYPE_EXPR_LESS:
		case NODE::TYPE_EXPR_LESS_EQ:
		case NODE::TYPE_EXPR_GREATER:
		case NODE::TYPE_EXPR_GREATER_EQ:
			assert(overridden == NULL);
			eval.typeinfo.SetRValue(defType[DEF_TYPE_BOOL]);
			break;

		case NODE::TYPE_EXPR_ASSIGN:
		case NODE::TYPE_EXPR_ADD:
		case NODE::TYPE_EXPR_SUB:
		case NODE::TYPE_EXPR_MUL:
		case NODE::TYPE_EXPR_DIV:
		case NODE::TYPE_EXPR_MOD:
			if (tree->type == NODE::TYPE_EXPR_ASSIGN || overridden == NULL) {
				eval.typeinfo = leftType;
			} else {
				eval.typeinfo = overridden->result;
			}
			break;

		default:
			assert(!"Unknown operator type");
			break;
	}

	if (res == 0 && eval.generateCode) {
		assert(!eval.code.empty());
		cmd.args.push_back((int)eval.code.size() - 1);

#define CASE_CMDTYPE(TYPE) case NODE::TYPE_EXPR_##TYPE: cmd.type = EVAL_COMMAND::TYPE_##TYPE; break;
		switch (tree->type) {
			CASE_CMDTYPE(ASSIGN);

			CASE_CMDTYPE(ADD);
			CASE_CMDTYPE(SUB);
			CASE_CMDTYPE(MUL);
			CASE_CMDTYPE(DIV);
			CASE_CMDTYPE(MOD);

			CASE_CMDTYPE(EQ);
			CASE_CMDTYPE(NOT_EQ);
			CASE_CMDTYPE(LESS);
			CASE_CMDTYPE(LESS_EQ);
			CASE_CMDTYPE(GREATER);
			CASE_CMDTYPE(GREATER_EQ);

			default: assert(!"Unknown operator type"); break;
		}
#undef  CASE_CMDTYPE

		cmd.typeinfo	= eval.typeinfo;
		cmd.info.proc	= overridden;
		cmd.line		= tree->subnode->line;
		eval.code.push_back(cmd);
	}

	return res;
}

/*************************************************************************/
/* PROJECT::EvalExprUnarOp()                                             */
/*                                                                       */
/*************************************************************************/
int PROJECT::EvalExprUnarOp(const NODE * tree, EVALUATE & eval)
{
	// process operand
	int res = Evaluate(tree->subnode, eval);
	if (res > 0) {
		return res;
	}

	if (!eval.typeinfo.IsValue()) {
		msg["name"] = Type2Str(eval.typeinfo);
		PRINT_MSG(TVR1022, tree->subnode->line);
		return 1;
	}

	EVAL_COMMAND cmd;
	cmd.line = tree->subnode->line;
	cmd.args.push_back((int)eval.code.size() - 1);

	switch (tree->type) {
		case NODE::TYPE_EXPR_NEG:
		case NODE::TYPE_EXPR_POS:
			if (!eval.typeinfo.IsArithmetic()) {
				msg["op"]   = (tree->type == NODE::TYPE_EXPR_NEG) ? "-" : "+";
				msg["name"] = Type2Str(eval.typeinfo);
				PRINT_MSG(TVR1023, tree->subnode->line);
				return 1;
			}

			if (tree->type == NODE::TYPE_EXPR_NEG && eval.generateCode) {
				cmd.type = EVAL_COMMAND::TYPE_NEG;
				eval.code.push_back(cmd);
			}
			break;

		case NODE::TYPE_EXPR_BIT_NOT:
			if (!eval.typeinfo.IsIntegral()) {
				msg["name"] = Type2Str(eval.typeinfo);
				PRINT_MSG(TVR1020, tree->subnode->line);
				return 1;
			}

			if (eval.generateCode) {
				cmd.type = EVAL_COMMAND::TYPE_BIT_NOT;
				eval.code.push_back(cmd);
			}
			break;

		case NODE::TYPE_EXPR_LOG_NOT:
			if (!eval.typeinfo.IsBoolean()) {
				msg["name2"] = "!";
				msg["name"] = Type2Str(eval.typeinfo);
				PRINT_MSG(TVR1032, tree->subnode->line);
				return 1;
			}

			if (eval.generateCode) {
				cmd.type = EVAL_COMMAND::TYPE_LOG_NOT;
				eval.code.push_back(cmd);
			}
			break;

		default:
			assert(!"Unknown unar operator");
			break;
	}

	return res;
}

/************************************************************************/
/* tavr::PROJECT::Evaluate
/*
/************************************************************************/
int PROJECT::Evaluate(const NODE * tree, EVALUATE & eval)
{
	assert(tree != NULL);

	int res = 0;
	bool dontGenerateCode = false;

	eval.typeinfo.Clear();

	switch (tree->type) {
		case NODE::TYPE_CODE_BLOCK:
			res += EvalCodeBlock(tree, eval);
			break;

		case NODE::TYPE_NAME:
			{
				assert(tree->subnode == NULL);

				PLACEMENT_INFO placementInfo;
				if (!NameLookup(tree->name, eval.context, eval.typeinfo, placementInfo)) {
					msg["name"] = tree->name.str();
					PRINT_MSG(TVR1008, tree->line);
					++res;
					break;
				}

				if (eval.generateCode && eval.typeinfo.IsValue()) {
					EVAL_COMMAND cmd;
					cmd.line = tree->line;

					// check whether this variable is type field
					switch (placementInfo.type) {
						case PLACEMENT_INFO::TYPE_TYPE_FIELD:
							// first get this value
							cmd.type = EVAL_COMMAND::TYPE_THIS;
							cmd.typeinfo.type		= TYPEINFO::TYPE_R_VALUE;
							assert(eval.context.curType != NULL);
							cmd.typeinfo.info.type	= eval.context.curType;
							eval.code.push_back(cmd);

							// get subname
							cmd.type = EVAL_COMMAND::TYPE_SUBNAME;
							cmd.args.push_back((int)eval.code.size() - 1);
							break;

						case PLACEMENT_INFO::TYPE_LOCAL_VAR:
							cmd.type = EVAL_COMMAND::TYPE_VARIABLE;
							cmd.args.push_back(placementInfo.index);
							break;

						case PLACEMENT_INFO::TYPE_PROC_ARG:
							cmd.type = EVAL_COMMAND::TYPE_VARIABLE;
							assert(eval.context.curMethod != NULL);
							cmd.info.proc = eval.context.curMethod;
							cmd.args.push_back(placementInfo.index);
							break;

						default:
							assert(placementInfo.type != PLACEMENT_INFO::TYPE_UNKNOWN);
							break;
					}

					cmd.typeinfo	= eval.typeinfo;
					cmd.info.name	= tree->name;
					eval.code.push_back(cmd);
				}
			}
			break;

		case NODE::TYPE_SUBNAME:
			res += EvalSubName(tree, eval);
			break;

		case NODE::TYPE_EXPR_GLOBAL:
			eval.typeinfo.type = TYPEINFO::TYPE_NAMESPACE;
			eval.typeinfo.info.names = globalName;
			break;

		case NODE::TYPE_EXPR_THIS:
			{
				if (eval.context.curType == NULL) {
					assert(eval.context.curMethod != NULL);
					TYPEINFO ti;
					ti.type = TYPEINFO::TYPE_PROC_NAME;
					ti.info.procName = eval.context.curMethod;
					msg["name"] = Type2Str(ti);
					PRINT_MSG(TVR1029, tree->line);
					++res;
					break;
				}

				eval.typeinfo.type = eval.context.curType->IsTypeByRef()
					? TYPEINFO::TYPE_R_VALUE : TYPEINFO::TYPE_L_VALUE;

				eval.typeinfo.info.type = eval.context.curType;

				if (eval.generateCode) {
					EVAL_COMMAND cmd;
					cmd.type		= EVAL_COMMAND::TYPE_THIS;
					cmd.typeinfo	= eval.typeinfo;
					cmd.line		= tree->line;
					eval.code.push_back(cmd);
				}
			}
			break;

		case NODE::TYPE_EXPR_EVENT:
			if (eval.eventType == NULL || eval.eventType->subtype != TYPE::SUBTYPE_EVENT) {
				PRINT_MSG(TVR1064, tree->line);
				++res;
				break;
			}

			eval.typeinfo.SetRValue(eval.eventType);
			
			if (eval.generateCode) {
				EVAL_COMMAND cmd;
				cmd.type = EVAL_COMMAND::TYPE_VARIABLE;
				cmd.line = tree->line;
				cmd.info.name = eval.code[eval.eventDeclIdx].info.name;
				cmd.typeinfo = eval.typeinfo;
				cmd.args.push_back(eval.eventDeclIdx);
				eval.code.push_back(cmd);
			}
			break;

		case NODE::TYPE_EXPR_ARRAY:
			res += EvalExprArray(tree, eval);
			break;

		case NODE::TYPE_DECL_VAR:
			res += EvalDeclVar(tree, eval);
			break;

		case NODE::TYPE_EXPR_NEW_ARRAY:
			res += EvalExprNewArray(tree, eval);
			break;

		case NODE::TYPE_EXPR_NEW_AUTO:
			res += EvalExprNewAuto(tree, eval);
			break;

		case NODE::TYPE_EXPR_NULL:
			eval.typeinfo.SetRValue(defType[DEF_TYPE_NULL]);

			if (eval.generateCode) {
				EVAL_COMMAND cmd;
				cmd.type		= EVAL_COMMAND::TYPE_CONSTANT;
				cmd.typeinfo	= eval.typeinfo;
				cmd.info.name	= defType[DEF_TYPE_NULL]->name;
				cmd.line		= tree->line;
				eval.code.push_back(cmd);
			}
			break;

		case NODE::TYPE_EXPR_BOOLEAN:
			assert(tree->name.str() == "true" || tree->name.str() == "false");
			eval.typeinfo.SetRValue(defType[DEF_TYPE_BOOL]);

			if (eval.generateCode) {
				EVAL_COMMAND cmd;
				cmd.type		= EVAL_COMMAND::TYPE_CONSTANT;
				cmd.typeinfo	= eval.typeinfo;
				cmd.info.name	= tree->name;
				cmd.line		= tree->line;
				eval.code.push_back(cmd);
			}
			break;

	#define CASE_CONSTANT(TYPE) \
		case NODE::TYPE_EXPR_##TYPE: \
			eval.typeinfo.SetRValue(defType[DEF_TYPE_##TYPE]); \
			\
			if (eval.generateCode) { \
				EVAL_COMMAND cmd; \
				cmd.type		= EVAL_COMMAND::TYPE_CONSTANT; \
				cmd.typeinfo	= eval.typeinfo; \
				cmd.info.name	= tree->name; \
				cmd.line		= tree->line; \
				eval.code.push_back(cmd); \
			} \
			break;

		CASE_CONSTANT(INT);
		CASE_CONSTANT(FLOAT);
		CASE_CONSTANT(DOUBLE);

		case NODE::TYPE_EXPR_CHAR:
			eval.typeinfo.SetRValue(defType[DEF_TYPE_CHAR]);

			res += VerifyEscapeSequences(tree->name.str(), tree->line);

			if (res == 0 && eval.generateCode) {
				EVAL_COMMAND cmd;
				cmd.type		= EVAL_COMMAND::TYPE_CONSTANT;
				cmd.typeinfo	= eval.typeinfo;
				cmd.info.name	= tree->name;
				cmd.line		= tree->line;
				eval.code.push_back(cmd);
			}
			break;

		case NODE::TYPE_EXPR_STRING:
			eval.typeinfo.SetRValue(defType[DEF_TYPE_STRING]);

			res += VerifyEscapeSequences(tree->name.str(), tree->line);

			if (res == 0 && eval.generateCode) {
				EVAL_COMMAND cmd;
				cmd.type		= EVAL_COMMAND::TYPE_CONSTANT;
				cmd.typeinfo	= eval.typeinfo;
				cmd.info.name	= tree->name;
				cmd.line		= tree->line;
				eval.code.push_back(cmd);
			}
			break;

		case NODE::TYPE_EXPR_IF:
			res += EvalExprIf(tree, eval);
			break;

		case NODE::TYPE_REG_POST:
		case NODE::TYPE_REG_SEND:
		case NODE::TYPE_UNREG:
			res += EvalReg(tree, eval);
			break;

#define CASE_EXPR_BREAK(NAME, STR_NAME) \
		case NODE::TYPE_##NAME: \
			if (eval.gotoAvailable.is(GOTO_TYPE_##NAME)) { \
				if (eval.generateCode) { \
					EVAL_COMMAND cmd; \
					cmd.type = EVAL_COMMAND::TYPE_GOTO; \
					cmd.line		= tree->line; \
					eval.code.push_back(cmd); \
					\
					GOTO_INFO gt; \
					gt.commandIndex = (int)eval.code.size() - 1; \
					gt.type			= GOTO_TYPE_##NAME; \
					eval.gotoPending.push_back(gt); \
				} \
			} else { \
				msg["name"] = STR_NAME; \
				PRINT_MSG(TVR1033, tree->line); \
				++res; \
			} \
			break

		CASE_EXPR_BREAK(BREAK,		"break");
		CASE_EXPR_BREAK(CONTINUE,	"continue");
		CASE_EXPR_BREAK(REPEAT,		"repeat");

		case NODE::TYPE_GOTO:
		case NODE::TYPE_RELAX:
		case NODE::TYPE_RELAX_STEP:
			res += EvalGotoRelax(tree, eval);
			break;

		case NODE::TYPE_RETURN:
			// return statements are not allowed in 'auto' methods
			assert(eval.context.curMethod != NULL);
			if (eval.context.curMethod->name == autoString) {
				PRINT_MSG(TVR1054, tree->line);
				++res;
				break;
			}

			// if tree->subnode is null, it's an empty 'return'
			if (tree->subnode != NULL) {
				res += Evaluate(tree->subnode, eval);
				if (res != 0) {
					break;
				}

				TYPEINFO left = eval.context.curMethod->result;

				METHOD * overridden = NULL;
				res += VerifyBinOpTypeCompatibility(NODE::TYPE_RETURN, tree->subnode->line,
					"return", left, eval.typeinfo, overridden);

				// TODO: support overriden copy constructor
				assert(overridden == NULL);

				if (res != 0) {
					break;
				}
			}

			if (eval.generateCode) {
				EVAL_COMMAND cmd;
				cmd.type = EVAL_COMMAND::TYPE_RETURN;

				assert(eval.context.curMethod != NULL);
				cmd.typeinfo = eval.context.curMethod->result;
				if (tree->subnode != NULL) {
				   cmd.args.push_back((int)eval.code.size() - 1);
				   cmd.line		= tree->subnode->line;
				} else {
					cmd.line	= tree->line;
				}
				eval.code.push_back(cmd);
			}
			break;

		case NODE::TYPE_SEND:
		case NODE::TYPE_POST:
			res += EvalSendPost(tree, eval);
			break;

		case NODE::TYPE_EXPR_SIZEOF:
			swap(dontGenerateCode, eval.generateCode);
			res += Evaluate(tree->subnode, eval);
			swap(dontGenerateCode, eval.generateCode);
			
			if (res != 0) {
				break;
			}

			if (eval.generateCode) {
				EVAL_COMMAND cmd;
				cmd.type		= EVAL_COMMAND::TYPE_SIZEOF;
				cmd.typeinfo	= eval.typeinfo;
				cmd.line		= tree->subnode->line;
				eval.code.push_back(cmd);
			}

			eval.typeinfo.SetRValue(defType[DEF_TYPE_INT]);
			break;

		case NODE::TYPE_EXPR_TYPEOF:
			swap(dontGenerateCode, eval.generateCode);
			res += Evaluate(tree->subnode, eval);
			swap(dontGenerateCode, eval.generateCode);
			if (res != 0) {
				break;
			}

			if (eval.typeinfo.IsValue()) {
				if (eval.typeinfo.info.type == defType[DEF_TYPE_NULL]) {
					msg["name"] = Type2Str(eval.typeinfo);
					PRINT_MSG(TVR1053, tree->subnode->line);
					++res;
					break;
				}
				eval.typeinfo.type = TYPEINFO::TYPE_TYPE;
			}
			break;

		case NODE::TYPE_EXPR_TYPE_CAST:
			res += EvalExprTypeCast(tree, eval);
			break;

		case NODE::TYPE_EXPR_CALL:
			res += EvalExprCall(tree, eval);
			break;

		case NODE::TYPE_IF:
			{
				// CONDITION
				res += Evaluate(tree->subnode, eval);

				if (res == 0 && (!eval.typeinfo.IsValue() || !eval.typeinfo.IsBoolean())) {
					msg["name"]  = "if";
					msg["name2"] = Type2Str(eval.typeinfo);
					PRINT_MSG(TVR1032, tree->subnode->line);
					++res;
				}

				if (res == 0 && eval.generateCode) {
					EVAL_COMMAND cmd;
					cmd.type	= EVAL_COMMAND::TYPE_IF;
					cmd.args.push_back((int)eval.code.size() - 1);
					cmd.line	= tree->subnode->line;
					eval.code.push_back(cmd);
				}

				const NODE * thenArg = tree->subnode->next;

				// THEN BODY
				res += Evaluate(thenArg->subnode, eval);

				const NODE * elseArg = thenArg->next;

				if (elseArg == NULL) {
					break;
				}

				// ELSE BODY
				if (res == 0 && eval.generateCode) {
					EVAL_COMMAND cmd;
					cmd.type	= EVAL_COMMAND::TYPE_ELSE;
					cmd.line	= tree->line;
					eval.code.push_back(cmd);
				}

				res += Evaluate(elseArg->subnode, eval);
			}
			break;

		case NODE::TYPE_ON:
		case NODE::TYPE_OTHERWISE:
			res += EvalOn(tree, eval);
			break;

		case NODE::TYPE_FOR:
			res += EvalFor(tree, eval);
			break;

		case NODE::TYPE_WHILE:
			res += EvalWhile(tree, eval);
			break;

		case NODE::TYPE_DO_WHILE:
			res += EvalDoWhile(tree, eval);
			break;

		case NODE::TYPE_STEP:
		case NODE::TYPE_LOOP:
		case NODE::TYPE_FIXED:
		case NODE::TYPE_SUBAUTO:
			res += EvalStep(tree, eval);
			break;

		case NODE::TYPE_EXPR_LOG_OR:
		case NODE::TYPE_EXPR_LOG_AND:
			res += EvalExprOrAnd(tree, eval);
			break;

#define CASE_EXPR_INCREMENET(NAME, VAL) \
		case NODE::TYPE_EXPR_##NAME: \
			res += Evaluate(tree->subnode, eval); \
			if (res != 0) { \
				break; \
			} \
			\
			if (!eval.typeinfo.IsIntegral()) { \
				msg["name"] = VAL; \
				PRINT_MSG(TVR1020, tree->subnode->line); \
				++res; \
				break; \
			} \
			\
			if (!eval.typeinfo.IsLValue()) { \
				msg["name"] = VAL; \
				PRINT_MSG(TVR1021, tree->subnode->line); \
				++res; \
				break; \
			} \
			\
			if (res == 0 && eval.generateCode) { \
				EVAL_COMMAND cmd; \
				cmd.type = EVAL_COMMAND::TYPE_##NAME; \
				cmd.args.push_back((int)eval.code.size() - 1); \
				cmd.line		= tree->subnode->line; \
				eval.code.push_back(cmd); \
			} \
			break

		CASE_EXPR_INCREMENET(PRE_INC,	"++");
		CASE_EXPR_INCREMENET(PRE_DEC,	"--");
		CASE_EXPR_INCREMENET(POST_INC,	"++");
		CASE_EXPR_INCREMENET(POST_DEC,	"--");

		case NODE::TYPE_EXPR_EQ:
		case NODE::TYPE_EXPR_NOT_EQ:
		case NODE::TYPE_EXPR_LESS:
		case NODE::TYPE_EXPR_LESS_EQ:
		case NODE::TYPE_EXPR_GREATER:
		case NODE::TYPE_EXPR_GREATER_EQ:

		case NODE::TYPE_EXPR_ASSIGN:
		case NODE::TYPE_EXPR_ADD:
		case NODE::TYPE_EXPR_SUB:
		case NODE::TYPE_EXPR_MUL:
		case NODE::TYPE_EXPR_DIV:
		case NODE::TYPE_EXPR_MOD:
			res += EvalExprBinOp(tree, eval);
			break;

		case NODE::TYPE_EXPR_NEG:
		case NODE::TYPE_EXPR_POS:
		case NODE::TYPE_EXPR_LOG_NOT:
		case NODE::TYPE_EXPR_BIT_NOT:
			res += EvalExprUnarOp(tree, eval);
			break;

		default:
			assert(!"Unknown node type");
			break;
	}

	return res;
}

/************************************************************************/
/* tavr::PROJECT::Generate
/*
/************************************************************************/
int PROJECT::Generate(GENERATE_INFO_MAP & generateInfo, MODULE * module)
{
	int res = 0;

	std::vector<MODULE_MEMBER *> memArr;
	memArr.reserve(64);

	for (MODULE_MEMBER * mem = module->members; mem != NULL; mem = mem->next) {
		memArr.push_back(mem);
	}

	for (int memIndex = (int)memArr.size() - 1; memIndex >= 0; --memIndex) {
		MODULE_MEMBER * mem = memArr[memIndex];

		switch (mem->type) {
			case MODULE_MEMBER::TYPE_METHOD:
				// generate procedures
				{
					METHOD * proc = mem->info.proc;

					const bool isAutoProc = (proc->type == METHOD::TYPE_AUTO_PROC);
					bool isMainAutoProc = false;

					EVALUATE eval;
					if (proc->IsProcedure()) {
						eval.context.curNamespace	= proc->owner.names;
					} else {
						assert(proc->IsMethod());
						eval.context.curType		= proc->owner.type;
						if (proc == eval.context.curType->specs[TYPE::SPEC_AUTO_PROC]) {
							isMainAutoProc = true;
						}
					}

					eval.context.curMethod		= proc;
					eval.context.module			= module;
					eval.generateCode			= (res == 0);

					if (proc->impl == NULL && !isMainAutoProc) {
						break;
					}

					// steps names info
					eval.steps.push(STEP_INFO_ARRAY());

					int evalResult = 0;
					if (!isAutoProc && proc->impl != NULL) {
						evalResult = Evaluate(proc->impl, eval);
					} else if (isAutoProc) {
						EVAL_COMMAND cmd;
						cmd.type = EVAL_COMMAND::TYPE_CODE_BLOCK_BEGIN;
						cmd.line = proc->declInfo.line;
						eval.code.push_back(cmd);

						if (isMainAutoProc) {
							std::vector<const NODE *> & steps = eval.context.curType->steps;
							for (int i=0; i<(int)steps.size(); ++i) {
								evalResult += Evaluate(steps[i], eval);
							}
						} else {
							evalResult = Evaluate(proc->impl, eval);
						}

						cmd.type = EVAL_COMMAND::TYPE_CODE_BLOCK_END;
						eval.code.push_back(cmd);
					}

					eval.steps.pop();
					assert(eval.steps.empty());

					// handle unreferenced goto statements
					if (evalResult == 0) {
						for (int i=0; i<(int)eval.gotoPending.size(); ++i) {
							GOTO_INFO & gt = eval.gotoPending[i];
							assert(gt.type == GOTO_TYPE_GOTO);

							msg["name"] = gt.name.str();
							PRINT_MSG(TVR1034, gt.line);
							++evalResult;
						}
					}

					if (evalResult != 0) {
						res += evalResult;
					} else {
						assert(proc->impl == NULL
							|| eval.typeinfo.type == TYPEINFO::TYPE_R_VALUE
							&& eval.typeinfo.info.type == defType[DEF_TYPE_VOID]);

						res += PostAnalyzeCode(generateInfo, eval);
					}

					swap(proc->code, eval.code);
				}
				break;

			case MODULE_MEMBER::TYPE_TYPE:
				// generate types
				{
					TYPE * type = mem->info.type;
					assert(type->declInfo.module	== module);

					EVALUATE eval;
					eval.context.curNamespace	= type->parentNamespace;
					eval.context.curType		= type;
					eval.context.module			= module;
					eval.generateCode			= (res == 0);

					for (int i=0; i<(int)type->fields.size(); ++i) {
						TYPE::FIELD & field = type->fields[i];
						if (field.value != NULL) {
							int evalResult = Evaluate(field.value, eval);

							if (evalResult == 0) {
								assert(field.line != -1);
								assert(field.type.type == TYPEINFO::TYPE_TYPE);

								METHOD * overridden = NULL;
								res += VerifyBinOpTypeCompatibility(NODE::TYPE_DECL_VAR, field.line, "=", field.type, eval.typeinfo, overridden);

								// TODO: support overridden copy constructors
								assert(overridden == NULL);

								field.type.type = TYPEINFO::TYPE_TYPE;
							} else {
								res += evalResult;
							}

							if (res == 0) {
								res += PostAnalyzeCode(generateInfo, eval);
								eval.code.back().isUsed = true;

								swap(field.code, eval.code);
							}
						}
					}
				}
				break;

			default:
				assert(!"Generation: Unknown module member type");
				break;
		}
	}

	return res;
}

/*************************************************************************/
/* PROJECT::CalcRelaxStackSize()                                         */
/*                                                                       */
/*************************************************************************/
int PROJECT::CalcRelaxStackSize(const GENERATE_INFO_MAP & infoMap, const GENERATE_INFO & info, RELAX_INFO & relaxInfo, TYPE * type, METHOD * method)
{
	assert(method->type == METHOD::TYPE_AUTO_PROC);

	EVAL_CODE & code = method->code;

	const std::vector<int> & sizeofCommands = info.sizeofCommands;
	std::vector<int> curAvail;
	curAvail.reserve(sizeofCommands.size());
	std::stack<int> prevStackSize;
	int curRelaxStackSize = 0;

	for (int index = 0, length = (int)sizeofCommands.size(); index < length; ++index) {
		int cmdIndex = sizeofCommands[index];
		EVAL_COMMAND & cmd = code[cmdIndex];

		switch (cmd.type) {
			case EVAL_COMMAND::TYPE_DECL_RELAX:
				{
					if (cmd.typeinfo.IsTypeByRef()) {
						curRelaxStackSize += TYPE_BY_REF_SIZE;
						break;
					}

					TYPE * varType = cmd.typeinfo.info.type;

					switch (varType->size) {
						case -1: {
							string info = type->declInfo.module->name;
							swap(info, msg["info"]);
							msg["name"] = cmd.info.name.str();
							PRINT_MSG(TVR1055, cmd.line);
							msg["name"] = varType->name.str();
							PRINT_MSG_INFO(TVR_SEEDECL_INFO, type->declInfo);
							swap(info, msg["info"]);
						}
						return 1;

					case 0: {
						assert(type->subtype == TYPE::SUBTYPE_STRUCT);
						int errors = CalcSizeOf(infoMap, varType);
						if (errors != 0) {
							return errors;
						}
					}
					break;

					default: {
						assert(varType->size > 0);
						curRelaxStackSize += varType->size;
						}
						break;
					}
				}
				break;

			case EVAL_COMMAND::TYPE_CODE_BLOCK_BEGIN:
				prevStackSize.push(curRelaxStackSize);
				break;

			case EVAL_COMMAND::TYPE_CODE_BLOCK_END:
				relaxInfo.stackSize = max(relaxInfo.stackSize, curRelaxStackSize);

				curRelaxStackSize = prevStackSize.top();
				prevStackSize.pop();
				break;

			default:
				assert(!"Unknown sizeof command type");
				break;
		}
	}

	assert(curAvail.empty());
	assert(prevStackSize.empty());

	return 0;
}

/************************************************************************/
/* tavr::PROJECT::CalcSizeOf
/*
/************************************************************************/
int PROJECT::CalcSizeOf(const GENERATE_INFO_MAP & info, TYPE * type)
{
	assert(type->size == 0);

	type->size = -1; // just a hack, mark this type as in processing

	int size = 0;
	for (int i=(int)type->fields.size() - 1; i >= 0; --i) {
		TYPE::FIELD & field = type->fields[i];

		if (field.type.IsTypeByRef()) {
			size += TYPE_BY_REF_SIZE;
			continue;
		}

		switch (field.type.info.type->size) {
			case -1:
				{
					string info = type->declInfo.module->name;
					swap(info, msg["info"]);
					msg["name"] = field.name.str();
					PRINT_MSG(TVR1028, field.line);
					msg["name"] = type->name.str();
					PRINT_MSG_INFO(TVR_SEEDECL_INFO, type->declInfo);
					swap(info, msg["info"]);
				}
				return 1;

			case 0:
				{
					assert(field.type.info.type->subtype == TYPE::SUBTYPE_STRUCT);
					int errors = CalcSizeOf(info, field.type.info.type);
					if (errors != 0) {
						return errors;
					}

					size += field.type.info.type->size;
				}
				break;

			default:
				assert(field.type.info.type->size > 0);
				size += field.type.info.type->size;
				break;
		}
	}

	// out table
	ECG_STATIC_ASSERT(TYPE_INT_SIZE == sizeof(void *));
	size += TYPE_INT_SIZE * (int)type->outTable.size();

	// relax_context value
	if (!type->relaxInfo.context.empty()) {
		size += TYPE_INT_SIZE;
	}
	for (int i=0; i<(int)type->subautos.size(); ++i) {
		SUBAUTO_INFO & si = type->subautos[i];
		if (!si.relaxInfo.context.empty()) {
			size += TYPE_INT_SIZE;
		}
	}

	METHOD * mainAuto = type->specs[TYPE::SPEC_AUTO_PROC];
	if (mainAuto != NULL) {
		// calc relax stack size
		AUTO_PROC_PAIR pair = { type, mainAuto };
		GENERATE_INFO_MAP::const_iterator infoIter = info.find(pair);
		assert(type->relaxInfo.stackSize == 0);
		if (infoIter != info.end()) {
			int res = CalcRelaxStackSize(info, infoIter->second, type->relaxInfo, type, mainAuto);
			if (res > 0) {
				return res;
			}

			size += type->relaxInfo.stackSize;
		}

		// process subautos
		for (int i=0; i<(int)type->subautos.size(); ++i) {
			SUBAUTO_INFO & si = type->subautos[i];

			assert(si.proc != NULL && si.proc->type == METHOD::TYPE_AUTO_PROC);
			AUTO_PROC_PAIR pair = { type, si.proc };

			GENERATE_INFO_MAP::const_iterator infoIter = info.find(pair);
			assert(si.relaxInfo.stackSize == 0);

			if (infoIter != info.end()) {
				int res = CalcRelaxStackSize(info, infoIter->second, si.relaxInfo, type, si.proc);
				if (res > 0) {
					return res;
				}

				size += si.relaxInfo.stackSize;
			}
		}
	}

	type->size = max(size, TYPE_INT_SIZE);

	return 0;
}

/************************************************************************/
/* tavr::PROJECT::PostGenerate
/*
/************************************************************************/
int PROJECT::PostGenerate(GENERATE_INFO_MAP & info, MODULE * module)
{
	// calc type sizes
	for (MODULE_MEMBER * mem = module->members; mem != NULL; mem = mem->next) {
		if (mem->type != MODULE_MEMBER::TYPE_TYPE) {
			continue;
		}

		TYPE * type = mem->info.type;
		if (type->size != 0) {
			assert(type->size > 0);
			continue;
		}

		int errors = CalcSizeOf(info, type);
		if (errors != 0) {
			return errors;
		}
	}

	return 0;
}

/************************************************************************/
/* tavr::PROJECT::PostAnalyzeCode
/*
/************************************************************************/
int PROJECT::PostAnalyzeCode(GENERATE_INFO_MAP & generateInfo, EVALUATE & eval)
{
	int res = 0;

	EVAL_CODE & code = eval.code;

	RELAX_INFO * relaxInfo = NULL;
	if (eval.context.curMethod != NULL && eval.context.curMethod->type == METHOD::TYPE_AUTO_PROC) {
		TYPE * curType = eval.context.curType;
	
		if (eval.context.curMethod == curType->specs[TYPE::SPEC_AUTO_PROC]) {
			relaxInfo = &curType->relaxInfo;
		} else {
			for (int i=0; i<(int)curType->subautos.size(); ++i) {
				if (curType->subautos[i].proc == eval.context.curMethod) {
					relaxInfo = &curType->subautos[i].relaxInfo;
				}
			}
		}
		assert(relaxInfo != NULL);
	}

	//////////////////////////////////////////////////////////////////////////
	// FIRST PASS

	const int codeSize = (int)code.size();

	std::vector<int> curVars;
	curVars.reserve(32);
	std::stack<int> blockVars;

	std::vector<int> relaxCommands;
	relaxCommands.reserve(32);

	std::vector<int> secondPassCommands;
	secondPassCommands.reserve(codeSize);

	for (int cmdIndex = 0; cmdIndex < codeSize; ++cmdIndex) {
		EVAL_COMMAND & cmd = code[cmdIndex];

		// don't mark dependencies for variables (its args are not a command index)
		if (cmd.type != EVAL_COMMAND::TYPE_VARIABLE) {
			for (int j=(int)cmd.args.size() - 1; j >= 0; --j) {
				code[cmd.args[j]].isUsed = true;
			}
		}
	
		// constraint: all statements should have corresponding line in source file
		assert(cmd.type == EVAL_COMMAND::TYPE_CODE_BLOCK_BEGIN
			|| cmd.type == EVAL_COMMAND::TYPE_CODE_BLOCK_END
			|| cmd.line != -1);

		switch (cmd.type) {
			case EVAL_COMMAND::TYPE_DECL:
			case EVAL_COMMAND::TYPE_DECL_RELAX:
				assert(cmd.type != EVAL_COMMAND::TYPE_DECL_RELAX);

				curVars.push_back(cmdIndex);
				secondPassCommands.push_back(cmdIndex);
				break;

			case EVAL_COMMAND::TYPE_LABEL:
				cmd.availVars = curVars;
				break;

			case EVAL_COMMAND::TYPE_CODE_BLOCK_BEGIN:
				blockVars.push((int)curVars.size());
				secondPassCommands.push_back(cmdIndex);
				break;

			case EVAL_COMMAND::TYPE_CODE_BLOCK_END:
				assert(!blockVars.empty());
				curVars.resize(blockVars.top());
				blockVars.pop();
				secondPassCommands.push_back(cmdIndex);
				break;

			case EVAL_COMMAND::TYPE_GOTO:
				secondPassCommands.push_back(cmdIndex);
				break;

			case EVAL_COMMAND::TYPE_RELAX:
				{
					assert(cmd.args.size() == 1);
					assert(eval.context.curMethod->type == METHOD::TYPE_AUTO_PROC);
					assert(relaxInfo != NULL);
					assert(eval.context.curType != NULL);

					// find label with such id to restore context
					TYPE * type = eval.context.curType;

					if (relaxInfo->context.find(cmd.args[0]) == relaxInfo->context.end()) {
						// relax context id should start from 1, because 0 it's default context - 'auto' function beginning
						relaxInfo->context.insert(make_pair(cmd.args[0], 1 + (int)relaxInfo->context.size()));
					}

					// store command index to change placement type
					relaxCommands.push_back(cmdIndex);

					secondPassCommands.push_back(cmdIndex);
				}
				break;
		}
	}

	// update variables placement type in destination step
	for (int i=0; i<(int)relaxCommands.size(); ++i) {
		EVAL_COMMAND & cmd = code[relaxCommands[i]];
		assert(cmd.type == EVAL_COMMAND::TYPE_RELAX);

		assert(cmd.args.size() == 1);
		EVAL_COMMAND & destCmd = code[cmd.args[0]];

		for (int i=(int)destCmd.availVars.size() - 1; i >= 0; --i) {
			EVAL_COMMAND & decl = code[destCmd.availVars[i]];
			assert(decl.type == EVAL_COMMAND::TYPE_DECL
				|| decl.type == EVAL_COMMAND::TYPE_DECL_RELAX);

			decl.type = EVAL_COMMAND::TYPE_DECL_RELAX;
		}
	}

	assert(blockVars.empty());
	assert(curVars.empty());

	//////////////////////////////////////////////////////////////////////////
	// SECOND PASS

	// calculate relax stack size
	TYPE * curType = NULL;
	GENERATE_INFO_MAP::iterator info = generateInfo.end();
	if (eval.context.curMethod != NULL && eval.context.curMethod->type == METHOD::TYPE_AUTO_PROC) {
		curType = eval.context.curMethod->owner.type;
#ifdef _DEBUG
		if (curType->specs[TYPE::SPEC_AUTO_PROC] == eval.context.curMethod) {
			assert(curType->relaxInfo.stackSize == 0);
		} else {
			bool found = false;
			for (int i=0; i<(int)curType->subautos.size(); ++i) {
				if (curType->subautos[i].proc == eval.context.curMethod) {
					assert(curType->subautos[i].relaxInfo.stackSize == 0);
					found = true;
					break;
				}
			}
			assert(found);
		}
#endif

		AUTO_PROC_PAIR pair = { curType, eval.context.curMethod };

		std::pair<GENERATE_INFO_MAP::iterator, bool> infoPair
			= generateInfo.insert(make_pair(pair, GENERATE_INFO()));
		info = infoPair.first;

		// we should process each method only once
		assert(infoPair.second);

		info->second.sizeofCommands.reserve(secondPassCommands.size());
	}

	for (int index = 0; index < (int)secondPassCommands.size(); ++index) {
		int cmdIndex = secondPassCommands[index];

		EVAL_COMMAND & cmd = code[cmdIndex];

		switch (cmd.type) {
			case EVAL_COMMAND::TYPE_DECL_RELAX:
				assert(curType != NULL);
				assert(info != generateInfo.end());
				info->second.sizeofCommands.push_back(cmdIndex);
				// no break
			case EVAL_COMMAND::TYPE_DECL:
				curVars.push_back(cmdIndex);
				break;

			case EVAL_COMMAND::TYPE_CODE_BLOCK_BEGIN:
				blockVars.push((int)curVars.size());

				if (curType != NULL) {
					info->second.sizeofCommands.push_back(cmdIndex);
				}
				break;

			case EVAL_COMMAND::TYPE_CODE_BLOCK_END:
				assert(!blockVars.empty());

#ifdef _DEBUG
				for (int i=blockVars.top(); i<(int)curVars.size(); ++i) {
					EVAL_COMMAND & declCmd = code[curVars[i]];
					if (declCmd.type != EVAL_COMMAND::TYPE_DECL_RELAX) {
						assert(declCmd.type == EVAL_COMMAND::TYPE_DECL);
					}
				}
#endif

				curVars.resize(blockVars.top());
				blockVars.pop();

				if (curType != NULL) {
					info->second.sizeofCommands.push_back(cmdIndex);
				}
				break;

			case EVAL_COMMAND::TYPE_GOTO:
			case EVAL_COMMAND::TYPE_RELAX:
				{
					assert(cmd.args.size() == 1);

					int labelIndex = cmd.args[0];
					EVAL_COMMAND & labelCmd = code[labelIndex];
					if (labelCmd.availVars.empty()) {
						cmd.availVars = curVars;
						break;
					}

					assert(labelCmd.availVars.size() <= curVars.size()
						|| labelIndex > cmdIndex);

					bool skippedError = false;
					int  skippedVarIndex = -1;
					if (curVars.empty()) {
						skippedError = true;
						skippedVarIndex = labelCmd.availVars[0];
					} else if (curVars.back() != labelCmd.availVars.back()) {
						if (cmdIndex > labelIndex) {
#ifdef _DEBUG
							for (int i=0; i<(int)labelCmd.availVars.size(); ++i) {
								assert(labelCmd.availVars[i] == curVars[i]);
							}
#endif
							assert(curVars.size() >= labelCmd.availVars.size());
							assert(cmd.availVars.empty());
							copy(curVars.begin() + labelCmd.availVars.size(), curVars.end(), back_inserter(cmd.availVars));
						} else {
							const int minSize = (int)min(labelCmd.availVars.size(), curVars.size());
							for (int i=0; i<minSize; ++i) {
								if (labelCmd.availVars[i] != curVars[i]) {
									skippedError = true;
									skippedVarIndex = labelCmd.availVars[i];
									break;
								}
							}
							if (!skippedError) {
								copy(curVars.begin() + minSize, curVars.end(), back_inserter(cmd.availVars));
							}
						}
					} else {
						assert(curVars.size() == labelCmd.availVars.size());
					}

					if (skippedError) {
						msg["name"] = code[skippedVarIndex].info.name.str();
						PRINT_MSG(TVR1046, code[skippedVarIndex].line);
						DECL_INFO declInfo;
						declInfo.module = eval.context.module;
						declInfo.line	= cmd.line;
						PRINT_MSG_INFO(TVR_GOTO_INFO, declInfo);
						++res;
						break;
					}
				}
				break;
		}
	}

	return res;
}

} // namespace tavr
