
#ifndef TAVR_LOADER_H
#define TAVR_LOADER_H

#pragma once

#include "tavr.h"
#include "tavr_loader_impl.h"
#include "tavr_loader_lex.h"
#include "tavr_loader_msg.h"

namespace ecg {

#define NT(T) tavr::NODE::TYPE_##T
#define ET(T) tavr::NODE::TYPE_EXPR_##T
#define ERR(T) MESSAGE_##T

class TAVR_LOADER : public HELPER<TAVR_LOADER, TAVR_LOADER_IMPL, TAVR_LOADER_LEX>
{
	friend class TAVR_LOADER_IMPL;
public:
	TAVR_LOADER			(tavr::PROJECT * project)
		: project(project), module(), errorCount()
		, strStep(toSTR_ID("step"))
		, strLoop(toSTR_ID("loop"))
		, strFixed(toSTR_ID("fixed"))
		, strSubauto(toSTR_ID("subauto"))
	{}

	void				SetModule			(tavr::MODULE * module_)
	{
		module		= module_;
		msg["info"] = module->name;
	}

	int					GetErrorCount		(void) const
	{ return errorCount; }

private:
	int					errorCount;
	MESSAGE::STRING_MAP	msg;
	tavr::PROJECT *		project;
	tavr::MODULE *		module;

	const ecg::STR_ID strStep, strLoop, strFixed, strSubauto;

public:
	typedef tavr::NODE *		PNODE;
	typedef tavr::NODE::TYPE	NODE_TYPE;

	struct STRVAL {
		STR_ID			strId;
		int				line;
	};

private:
	inline PNODE		NewNode				(NODE_TYPE type)
	{
		PNODE res	= module->treeAlloc.Alloc<tavr::NODE> ();
		res->name	= project->emptyString;
		res->line	= -1;
		res->type	= type;
		return res;
	}

public:
	inline void			Node				(PNODE & res)
	{
		res = 0;
	}

	inline void			Node				(NODE_TYPE type, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		res = NewNode(type);

		res->line = Line();
	}

	inline void			Node				(PNODE val, PNODE & res)
	{
		// just a transition
		res = val;
	}

	inline void			Node				(NODE_TYPE type, const std::string & value, PNODE & res)
	{ 
		if (errorCount > 0) {
			return ;
		}

		res = NewNode(type);
		res->name = ecg::toSTR_ID(value);

		res->line = Line();
	}

	inline void			Node				(NODE_TYPE type, const std::string & value, PNODE subnode, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		Node(type, value, res);
		res->subnode = subnode;
	}

	inline void			Node				(NODE_TYPE type, STRVAL value, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		res = NewNode(type);
		res->name = value.strId;
		res->line = value.line;
	}

	inline void			Node				(NODE_TYPE type, STRVAL value, PNODE subnode, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		res = NewNode(type);
		res->name = value.strId;
		res->line = value.line;
		res->subnode = subnode;
	}

	inline void			Node				(NODE_TYPE type, PNODE subnode, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		res = NewNode(type);
		res->subnode = subnode;
		res->line	 = Line();
	}

	inline void			Node				(NODE_TYPE type, PNODE subnode, STRVAL value, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		res = NewNode(type);
		res->name = value.strId;
		res->line = value.line;
		res->subnode = subnode;
	}

	inline void			Node				(NODE_TYPE type, PNODE subnode, PNODE next, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		res = NewNode(type);
		res->subnode = subnode;
		res->line	 = Line();
		assert(subnode->next == NULL);
		subnode->next = next;
	}

	inline void			Node				(NODE_TYPE type, PNODE subnode, PNODE next, PNODE next2, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		res = NewNode(type);
		res->subnode = subnode;
		res->line	 = Line();
		assert(subnode->next == NULL);
		subnode->next = next;
		assert(next->next == NULL);
		next->next   = next2;
	}

	inline void			NodeList			(PNODE & res)
	{
		res = NULL;
	}

	inline void			NodeList			(PNODE listInit, PNODE & res)
	{ 
		res = listInit;
	}

	inline void			NodeList			(PNODE nodeList, PNODE item, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		// skip empty items
		if (item == NULL) {
			res = nodeList;
			return ;
		}

		res = item;
		assert(item->next == NULL);
		item->next = nodeList;
	}

	inline void			StrVal				(STRVAL & res)
	{
		if (errorCount > 0) {
			return ;
		}

		res.strId	= toSTR_ID(TokenValue());
		res.line	= Line();
	}

	inline void			Program				(PNODE tree)
	{ module->tree = tree; }

	inline void			FuncHeader			(PNODE resType, STRVAL name, PNODE argList, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		res = NewNode(tavr::NODE::TYPE_PROC_HEADER);
		res->name = name.strId;
		res->line = name.line;
		res->subnode = resType;

		assert(resType->next == NULL);
		resType->next = argList;
	}

	inline void			FuncHeader			(STRVAL name, PNODE argList, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		res = NewNode(tavr::NODE::TYPE_PROC_HEADER);
		res->name = name.strId;
		res->line = name.line;
		
		res->subnode		= NewNode(tavr::NODE::TYPE_ARGUMENT);
		res->subnode->next	= argList;
	}

	inline void			OperatorHeader		(const char * opVal, PNODE argList, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		STRVAL name;
		name.strId = toSTR_ID(opVal);
		name.line  = Line();

		FuncHeader(name, argList, res);
	}

	inline void			OperatorHeader		(const char * opVal, PNODE resType, PNODE argList, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		STRVAL name;
		name.strId = toSTR_ID(opVal);
		name.line  = Line();

		FuncHeader(resType, name, argList, res);
	}

	#define TAVR_LOADER_PUSH_BACK_ARG_BEGIN(ROOT, TYPE) \
		(ROOT) = NewNode(TYPE); \
		(ROOT)->line = Line(); \
		PNODE * __parent__ = &(ROOT)->subnode;

	#define TAVR_LOADER_PUSH_BACK_ARG(ARG) \
		PNODE ARG##Arg = NewNode(tavr::NODE::TYPE_ARGUMENT); \
		ARG##Arg->subnode = ARG; \
		assert(*__parent__ == NULL); \
		*__parent__ = ARG##Arg; \
		__parent__ = &ARG##Arg->next;

	inline void			NodeArg				(NODE_TYPE type, PNODE a, PNODE b, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		TAVR_LOADER_PUSH_BACK_ARG_BEGIN(res, type);
		TAVR_LOADER_PUSH_BACK_ARG(a);
		TAVR_LOADER_PUSH_BACK_ARG(b);
	}

	inline void			NodeArg				(NODE_TYPE type, PNODE a, STRVAL val, PNODE b, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		TAVR_LOADER_PUSH_BACK_ARG_BEGIN(res, type);
		TAVR_LOADER_PUSH_BACK_ARG(a);
		TAVR_LOADER_PUSH_BACK_ARG(b);

		res->name = val.strId;
		res->line = val.line;
	}

	inline void			NodeArg				(NODE_TYPE type, PNODE a, PNODE b, PNODE c, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		TAVR_LOADER_PUSH_BACK_ARG_BEGIN(res, type);
		TAVR_LOADER_PUSH_BACK_ARG(a);
		TAVR_LOADER_PUSH_BACK_ARG(b);
		TAVR_LOADER_PUSH_BACK_ARG(c);
	}

	inline void			NodeArg				(NODE_TYPE type, PNODE a, PNODE b, PNODE c, PNODE d, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		TAVR_LOADER_PUSH_BACK_ARG_BEGIN(res, tavr::NODE::TYPE_FOR);

		TAVR_LOADER_PUSH_BACK_ARG(a);
		TAVR_LOADER_PUSH_BACK_ARG(b);
		TAVR_LOADER_PUSH_BACK_ARG(c);
		TAVR_LOADER_PUSH_BACK_ARG(d);
	}

	inline void			Step				(STRVAL stepType, STRVAL name, PNODE codeblock, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		tavr::NODE::TYPE type;
		if (stepType.strId == strLoop) {
			type = tavr::NODE::TYPE_LOOP;
		} else if (stepType.strId == strFixed) {
			type = tavr::NODE::TYPE_FIXED;
		} else if (stepType.strId == strSubauto) {
			type = tavr::NODE::TYPE_SUBAUTO;
		} else {
			assert(stepType.strId == strStep);
			type = tavr::NODE::TYPE_STEP;
		}

		Node(type, name, codeblock, res);
	}

	inline void			Step				(STRVAL stepType, STRVAL name, PNODE & res)
	{
		Step(stepType, name, NULL, res);
	}

	inline void			StatIf				(PNODE cond, PNODE thenList, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		res = NewNode(tavr::NODE::TYPE_IF);
		res->subnode		= cond;
		res->line			= Line();

		PNODE thenArg		= NewNode(tavr::NODE::TYPE_ARGUMENT);
		thenArg->subnode	= thenList;

		assert(cond->next == NULL);
		cond->next			= thenArg;
	}
	
	inline void			StatIf				(PNODE cond, PNODE thenList, PNODE elseList, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		res = NewNode(tavr::NODE::TYPE_IF);
		res->subnode		= cond;
		res->line			= Line();

		PNODE thenArg = NewNode(tavr::NODE::TYPE_ARGUMENT);
		thenArg->subnode	= thenList;
		assert(cond->next == NULL);
		cond->next			= thenArg;

		PNODE elseArg = NewNode(tavr::NODE::TYPE_ARGUMENT);
		elseArg->subnode	= elseList;
		assert(thenArg->next == NULL);
		thenArg->next		= elseArg;
	}

	inline void			ArgEmpty			(STRVAL & res)
	{
		res.strId	= project->emptyString;
		res.line	= Line();
	}

	inline void			ArgList				(PNODE type, STRVAL name, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		res = NewNode(tavr::NODE::TYPE_ARGUMENT);
		res->name = name.strId;
		res->line = name.line;
		res->subnode = type;
	}

	inline void			ArgList				(PNODE argList, PNODE type, STRVAL arg, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		res = NewNode(tavr::NODE::TYPE_ARGUMENT);
		res->name = arg.strId;
		res->line = arg.line;
		res->subnode = type;
		res->next = argList;
	}

	inline void			CodeBlock			(PNODE statList, PNODE & res)
	{
		if (errorCount > 0) {
			return ;
		}

		if (statList != NULL && statList->type == tavr::NODE::TYPE_CODE_BLOCK && statList->next == NULL) {
			res = statList;
			return ;
		}

		res = NewNode(tavr::NODE::TYPE_CODE_BLOCK);
		res->subnode	= statList;
	}

	//////////////////////////////////////////////////////////////////////////
	// ERROR PROCESSING

	inline void			PrintMessage		(int messageId, int line = -1)
	{
		const MSG_HOLDER * msgHolder = TAVR_LOADER_GetMessageByIndex(messageId);
		std::string format = "{info}({line}): ";
		format += msgHolder->format;
		msg["err"]  = msgHolder->id;
		msg["line"] = ToString(line == -1 ? Line() : line);

		msg["token"] = TokenValue();
		if (Token() != TOKEN_EOF) {
			msg["token"] = Compose(TAVR_LOADER_GetMessageByIndex(MESSAGE_TOKEN_INFO), msg);
		} else {
			msg["token"] = Compose(TAVR_LOADER_GetMessageByIndex(MESSAGE_TOKEN_INFO_EOF), msg);
		}

		GetError() << Compose(format, msg) << std::endl;
	}

	inline void			Error				(void)
	{ Error(MESSAGE_TVR1001); }

	template <class T1>
	inline void			Error				(int msgId, T1 & res)
	{ Error(msgId); }

	template <class T1>
	inline void			ErrorOK				(int msgId, T1 & res)
	{ Error(msgId); /*StateOK();*/ }

	template <class T1, class T2>
	inline void			Error				(int msgId, T1 a, T2 & b)
	{ Error(msgId); }

	template <class T1, class T2, class T3>
	inline void			Error				(int msgId, T1 a, T2 b, T3 & res)
	{ Error(msgId); }

	template <class T1, class T2, class T3, class T4>
	inline void			Error				(int msgId, T1 a, T2 b, T3 c, T4 & res)
	{ Error(msgId); }

	inline void			Error				(int msgId)
	{
		if (IsCleanUp()) {
			return ;
		}

		PrintMessage(msgId);
		++errorCount;

		StateCleanUp();
	}

	template <class T1, class T2>
	inline void			NodeError			(T1 a, T2 & b)
	{ NodeError(b); }

	inline void			NodeError			(PNODE & result)
	{ 
		if (!IsCleanUp()) {
			Error();
			return ;
		}

		StateOK();
	}
};

} // namespace ecg

#endif // TAVR_LOADER_H