
#ifndef TAVR_TARGET_C_H
#define TAVR_TARGET_C_H

#pragma once

namespace tavr {

class PROJECT;
struct PROJECT_SETTINGS;
struct MODULE;
struct TYPE;
struct METHOD;
struct TYPEINFO;
struct EVAL_COMMAND;
struct RELAX_INFO;

typedef std::vector<EVAL_COMMAND> EVAL_CODE;

struct PENDING_TMP_VAR {
	int		index;
	int		argSet;
	bool	needDestructor;
};
typedef std::vector<PENDING_TMP_VAR> PENDING_TMP_VAR_ARRAY;

class TARGET_C {
	typedef std::vector<int> INT_ARRAY;

	struct COMMAND_INFO {
		std::string		value;
		INT_ARRAY		tmpVars;
		int				argSet;

		PENDING_TMP_VAR_ARRAY pendingTmpVars;

		COMMAND_INFO(void)
			: argSet(-1)
		{}
	};

	struct TMP_VAR;

	typedef std::vector<TMP_VAR>	TMP_VAR_ARRAY;
	typedef std::set<TMP_VAR>		TMP_VAR_SET;

	struct INFO {
		std::vector<COMMAND_INFO>	cmdInfo;
		std::vector<INT_ARRAY>		argSets;
		std::vector<TMP_VAR_ARRAY>	tmpVarSets;

		std::vector<INFO>			fieldInfo;
		std::vector<INFO>			constrInfo;
		COMMAND_INFO				constrCmdInfo;
		COMMAND_INFO				copyConstrCmdInfo;
		COMMAND_INFO				eventConstrCmdInfo;
		COMMAND_INFO				defAssignOpInfo;
		INT_ARRAY					fieldTmpVars;			// is needed for def assign op generation (it's a hack)
		std::vector<INT_ARRAY>		relaxCopyTmpVars;

		METHOD *					curProc;
		int							blocksCounter;
		TMP_VAR_SET					unusedTmpVars;

		std::map<int, PENDING_TMP_VAR_ARRAY> forceReleaseTmpVars;

		inline INFO(void);
	};

	struct GLOBAL_INFO {
		std::set<METHOD *>			forwardProcs;
		std::set<TYPEINFO>			forwardTypes;
		std::stack<int>				argSetIdx;
		int							labelCounter;
		int							tmpVarCounter;
		int							dummyCounter;

		GLOBAL_INFO(void)
			: labelCounter(), tmpVarCounter(), dummyCounter()
		{}
	};

public:
	TARGET_C				(void)
		: project(), settings()
	{}

	void					Init					(PROJECT * proj) { project = proj; }

	bool					PreGenerate				(void);
	bool					Generate				(MODULE * module, std::string & outputName);
	bool					GenerateTypeinfo		(std::string & outputName);
	bool					PostGenerate			(void);

private:
	void					AddTab					(void) { tabs.push_back('\t'); }
	void					RemoveTab				(void) { if (!tabs.empty()) tabs.resize(tabs.size() - 1); }

	void					PreProcessCode			(const EVAL_CODE & code, INFO & localInfo, GLOBAL_INFO & globalInfo);
	void					PreProcessProc			(METHOD * proc, INFO & info, GLOBAL_INFO & globalInfo);
	inline void				PreProcessProcRelax		(METHOD * proc, RELAX_INFO & relaxInfo, COMMAND_INFO & cmdInfo, INFO & info, GLOBAL_INFO & globalInfo);
	void					ProcessProc				(METHOD * proc, INFO & info, GLOBAL_INFO & globalInfo);

	void					PreProcessType			(TYPE * type, INFO & info, GLOBAL_INFO & globalInfo);
	void					ProcessType				(TYPE * type, INFO & info, GLOBAL_INFO & globalInfo);

	inline void				PreProcessProcPrintHeader(METHOD * proc, GLOBAL_INFO & globalInfo);
	inline void				PreProcessTypePrintDecl	(const TYPEINFO & ti, GLOBAL_INFO & globalInfo);

	void					PrintCode				(const EVAL_CODE & code, INFO & localInfo);

	void					PreProcessFieldCopy		(const TYPEINFO & ti, COMMAND_INFO & cmdInfo, int & tmpVar, INFO & localInfo, GLOBAL_INFO & globalInfo);

	void					PrintProcHeader			(METHOD * proc, GLOBAL_INFO & globalInfo);
	inline void				PrintProcHeaderArrDestr	(const TYPEINFO & ti);
	inline void				PrintVarDefInit			(const TYPEINFO & type, const std::string & value);

	inline void				PreProcessCodeBlockBegin(INFO & localInfo, GLOBAL_INFO & globalInfo, COMMAND_INFO & cmdInfo);
	inline void				PreProcessCodeBlockEnd	(INFO & localInfo, GLOBAL_INFO & globalInfo, COMMAND_INFO & cmdInfo);

	inline void				ProcessCodeBlockBegin	(INFO & localInfo, COMMAND_INFO & cmdInfo, const EVAL_CODE * code = NULL, int index = -1);
	inline void				ProcessCodeBlockEnd		(INFO & localInfo, COMMAND_INFO & cmdInfo, const EVAL_CODE * code = NULL, int index = -1);

	inline void				PrintArrDestrBody		(const TYPEINFO & ti, const std::string & value);
	inline void				PrintDestructor			(const TYPEINFO & ti, const std::string & value);
	inline void				PrintAssignOp			(bool thisCall, const TYPEINFO & leftType, const std::string & left, COMMAND_INFO & cmdInfo, INFO & localInfo, const EVAL_COMMAND * cmd, const EVAL_CODE * code, std::string * right = NULL);
	inline void				PrintCopyConstr			(const TYPEINFO & leftType, const std::string & left, const std::string & right);

	enum PRINT_RELAX_COPY_TYPE {
		PRCT_COPY_CONSTR,
		PRCT_OP_EQ,
		PRCT_DESTR,
	};

	void					PrintRelaxCopy			(PRINT_RELAX_COPY_TYPE copyType, INFO & localInfo, TYPE * type, const std::string & right = std::string());
	void					PrintRelaxCopy			(PRINT_RELAX_COPY_TYPE copyType, RELAX_INFO & relaxInfo, METHOD * method, INFO & localInfo, TYPE * type, const std::string & right);

	inline void				PreProcessProcArgs		(METHOD * proc, COMMAND_INFO & cmdInfo, INFO & localInfo, GLOBAL_INFO & globalInfo);
	inline std::string		ProcessProcArgs			(bool thisCall, METHOD * proc, COMMAND_INFO & cmdInfo, INFO & info, const EVAL_COMMAND * cmd, const EVAL_CODE * code, std::string * right = NULL);

	inline std::string		GetTypeName				(const TYPEINFO & ti, bool decoratedName = false);
	inline std::string		GetDecoratedName		(const std::string & name);
	inline std::string		GetProcName				(METHOD * proc);
	inline std::string		GetRefName				(const TYPEINFO & ti);

	inline int				GetTmpVar				(const TYPEINFO & ti, bool needDestructor, COMMAND_INFO & cmdInfo, INFO & localInfo, GLOBAL_INFO & globalInfo);
	inline void				PreProcessReleaseTmpVars(PENDING_TMP_VAR_ARRAY & pendingTmpVars, COMMAND_INFO & cmdInfo, INFO & localInfo);

	inline void				FinalizeTmpVars			(PENDING_TMP_VAR_ARRAY & arr, INFO & localInfo);
	inline void				FinalizeLocalVars		(COMMAND_INFO & cmdInfo, INFO & localInfo, const EVAL_CODE & code, int returnCmdIndex = -1);
	inline void				FinalizeLocalVarsReturn	(COMMAND_INFO & cmdInfo, INFO & localInfo, const EVAL_CODE & code, int returnCmdIndex);
	inline void				FinalizeProcArgs		(METHOD * proc);
	inline void				FinalizeGotoVars		(const std::vector<int> & availVars, const EVAL_CODE & code, INFO & localInfo);

private:
	PROJECT *				project;

	const PROJECT_SETTINGS * settings;

	// output settings
	std::vector<MODULE *>	moduleList;
	std::string				tmp;
	std::string				out;

	std::string				name;
	std::string				tavrRuntime;

	// generation data
	std::string				tabs;
	std::ofstream			output;

	// typeinfo data
	std::set<TYPEINFO>		typeinfoTypes;
};

struct TAVR_TYPE_SPECIAL_NAMES_T {
	const char * name;
	const char * decoration;
};

extern TAVR_TYPE_SPECIAL_NAMES_T TAVR_TYPE_SPECIAL_NAMES[];

#define TAVR_PREFIX					"__tavr_"
#define TAVR_REF_POSTFIX			"r"
#define TAVR_ARRAY_POSTFIX			"a"
#define TAVR_NEW					"new"
#define TAVR_DESTRUCTOR				"destr"

#define TAVR_THIS					"__tavr_this"
#define TAVR_RELAX_CONTEXT			"__tavr_relax"
#define TAVR_RELAX_STACK			"__tavr_stack"
#define TAVR_OUT_PREFIX				"__tavr_out_"

#define TAVR_TYPEINFO				"__tavr_ti_"
#define TAVR_SRC					"__tavr_src"
#define TAVR_REF					"__tavr_ref_"
#define TAVR_REF_ARRAY				"__tavr_array_"
#define TAVR_TMP					"__tavr_tmp_"
#define TAVR_DUMMY					"__tavr_dummy_"
#define TAVR_DUMMY_ARG				"__tavr_dummy_arg_"
#define TAVR_LABEL					"__tavr_label_"
#define TAVR_ALLOC					"__tavr_alloc"
#define TAVR_ALLOC_ARRAY			"__tavr_alloc_array"
#define TAVR_TYPEINFO_OFFSET		"1"
#define TAVR_OFFSET					"2"
#define TAVR_ARRAY_OFFSET			"3"
#define TAVR_EVENT_ID_OFFSET		"2"

#define TAVR_NULL_TYPE_DEC			"null"
#define TAVR_NULL_TYPE				"__tavr_ref_t"
#define TAVR_ALLOC_STRING			"__tavr_alloc_string"

#define TAVR_INC_REF				"__TAVR_INC_REF"
#define TAVR_DEC_REF_RAW			"__TAVR_DEC_REF_RAW"
#define TAVR_DEC_REF				"__TAVR_DEC_REF"
#define TAVR_DEC_ARRAY				"__TAVR_DEC_ARRAY"
#define TAVR_DEC_ARRAY_REF			"__TAVR_DEC_ARRAY_REF"
#define TAVR_DEC_ARRAY_REF_RAW		"__TAVR_DEC_ARRAY_REF_RAW"
#define TAVR_DEC_ARRAY_BEGIN		"__TAVR_DEC_ARRAY_BEGIN"
#define TAVR_DEC_ARRAY_END			"__TAVR_DEC_ARRAY_END"

#define TAVR_DESTR_ARRAY_REF		"__TAVR_DESTR_ARRAY_REF"
#define TAVR_DESTR_ARRAY_VALUE		"__TAVR_DESTR_ARRAY_VALUE"

#define TAVR_ARRAY_LENGTH			"__TAVR_ARRAY_LENGTH"
#define TAVR_ARRAY_DESTR			"__tavr_arr_destr_"

#define TAVR_ASSERT					"__TAVR_ASSERT"
#define TAVR_ASSERT_NULL_PTR		"__TAVR_ASSERT_NULL_PTR"
#define TAVR_ASSERT_ARRAY_INDEX		"__TAVR_ASSERT_ARRAY_INDEX"
#define TAVR_ASSERT_STRING_INDEX	"__TAVR_ASSERT_STRING_INDEX"
#define TAVR_UNKNOWN_RELAX_CONTEXT	"__TAVR_UNKNOWN_RELAX_CONTEXT"

#define TAVR_RELAX_BUSY_BIT			"__TAVR_RELAX_BUSY_BIT"

#define TAVR_RELAX_CASE_DEFAULT		"__TAVR_RELAX_CASE_DEFAULT"
#define TAVR_RELAX_CASE				"__TAVR_RELAX_CASE"
#define TAVR_RELAX_CASE_DEFAULT2	"__TAVR_RELAX_CASE_DEFAULT2"
#define TAVR_RELAX_CASE2			"__TAVR_RELAX_CASE2"
#define TAVR_RELAX_CASE_DEFAULT_S	"__TAVR_RELAX_CASE_DEFAULT_S"
#define TAVR_RELAX_CASE_S			"__TAVR_RELAX_CASE_S"
#define TAVR_RELAX_CASE_DEFAULT2_S	"__TAVR_RELAX_CASE_DEFAULT2_S"
#define TAVR_RELAX_CASE2_S			"__TAVR_RELAX_CASE2_S"

#define TAVR_CUR_EVENT				"__tavr_cur_event"
#define TAVR_TMP_EVENT				"__tavr_tmp_event"
#define TAVR_POST_EVENT				"__tavr_post_event"
#define TAVR_AUTO_PROC				"__tavr_auto_proc_t"
#define TAVR_ADD_POST_LAYER			"__tavr_add_post_layer"
#define TAVR_REMOVE_POST_LAYER		"__tavr_remove_post_layer"

#define TAVR_OUT_CALL				"__tavr_out_call"
#define TAVR_RELEASE_OUT_LIST		"__tavr_release_out_list"
#define TAVR_ADD_OUT_LIST			"__tavr_add_out_list"
#define TAVR_REMOVE_OUT_LIST		"__tavr_remove_out_list"

}; // namespace tavr

#endif // TAVR_TARGET_C_H
